
import bpy
import os
import functools
import difflib
import time
from subprocess import run
import random
import mathutils
import threading
import sys
import socket
import warnings
from requests import Session
import certifi
import random

headers = {'User-Agent': f'TrueAssets/4.1 Blender/{bpy.app.version_string}'}
s = Session()
s.headers.update(headers)
ADDONS_DIR = os.path.dirname(os.path.dirname(__file__))
THUMBNAIL_RESOLUTION=256
supports_thumbnails=['MESH','LIGHT','VOLUME']
inputs_to_skip_compare=[9,]
default_values_bsdf=[[0.8, 0.8, 0.8, 1.0], 0.0, [1.0, 0.2, 0.1], [0.8, 0.8, 0.8, 1.0], 1.4, 0.0, 0.0, 0.5, 0.0, 0.4, 0.0, 0.0, 0.0, 0.5, 0.0, 0.03, 1.45, 0.0, 0.0, [0.0, 0.0, 0.0, 1.0], 1.0, 1.0, [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]
import uuid
import re
def is_sequence(obj):
    t = type(obj)
    return hasattr(t, '__len__') and hasattr(t, '__getitem__')
def get_catalog_from_directory(dir_path,only_top_level=True):
    catalogs=[]
    if os.path.isfile(os.path.join(dir_path,"blender_assets.cats.txt")):
            with open(os.path.join(dir_path,"blender_assets.cats.txt"), mode='r',newline="\n",encoding="UTF-8") as catalog_file:
                for l in catalog_file.readlines():
                    if not l.startswith("#") and not l.startswith("VERSION") and ":" in l and ('/' not in l or not only_top_level) :
                        catalogs.append((l[:l.index(":")],l[l.index(":")+1:l.rindex(":")]))
    catalogs=sorted(catalogs,key=lambda x:get_name_to_display(x[1]).lower())
    return catalogs
def get_catalogs_to_move_to_for_mats(self, context):
    catalogs=[]
    if os.path.isfile(os.path.join(self.dir_path,"blender_assets.cats.txt")):
        with open(os.path.join(self.dir_path,"blender_assets.cats.txt"), mode='r',newline="\n",encoding="UTF-8") as catalog_file:
            for l in catalog_file.readlines():
                if not l.startswith("#") and not l.startswith("VERSION") and ":" in l:
                    catalogs.append((l[:l.index(":")],l[l.index(":")+1:l.rindex(":")]))
    catalogs=sorted(catalogs,key=lambda x:get_name_to_display(x[1]).lower())
    get_catalogs_to_move_to_for_mats.catalogs= [("00000000-0000-0000-0000-000000000000","None","None"),("NEW","New/Search","New/Search"),("AUTO",'Auto','Auto'),("Directory","Directory Structure","Directory Structure"),("Directory_PH","Directory Structure (PolyHaven)","Directory Structure (PolyHaven)")]+[(a,get_name_to_display(b),b) for a,b in catalogs]
get_catalogs_to_move_to_for_mats.catalogs=[]
def get_catalogs_to_move_to_for_worlds(self, context):
    catalogs=[]
    if os.path.isfile(os.path.join(self.dir_path,"blender_assets.cats.txt")):
        with open(os.path.join(self.dir_path,"blender_assets.cats.txt"), mode='r',newline="\n",encoding="UTF-8") as catalog_file:
            for l in catalog_file.readlines():
                if not l.startswith("#") and not l.startswith("VERSION") and ":" in l:
                    catalogs.append((l[:l.index(":")],l[l.index(":")+1:l.rindex(":")]))
    catalogs=sorted(catalogs,key=lambda x:get_name_to_display(x[1]).lower())
    get_catalogs_to_move_to_for_worlds.catalogs= [("00000000-0000-0000-0000-000000000000","None","None",0),("NEW","New/Search","New/Search",1),("Directory","Directory Structure","Directory Structure",2)]+[(a,get_name_to_display(b),b,i+3) for i,(a,b) in enumerate(catalogs)]
get_catalogs_to_move_to_for_worlds.catalogs=[]
def get_catalogs_to_move_to(self, context):
    catalogs=[]
    if os.path.isfile(os.path.join(self.dir_path,"blender_assets.cats.txt")):
        with open(os.path.join(self.dir_path,"blender_assets.cats.txt"), mode='r',newline="\n",encoding="UTF-8") as catalog_file:
            for l in catalog_file.readlines():
                if not l.startswith("#") and not l.startswith("VERSION") and ":" in l:
                    catalogs.append((l[:l.index(":")],l[l.index(":")+1:l.rindex(":")]))
    catalogs=sorted(catalogs,key=lambda x:get_name_to_display(x[1]).lower())
    get_catalogs_to_move_to.catalogs= [("00000000-0000-0000-0000-000000000000","None","None",0),("NEW","New/Search","New/Search",1)]+[(a,get_name_to_display(b),b,i+2) for i,(a,b) in enumerate(catalogs)]
get_catalogs_to_move_to.catalogs=[]
def join_meshes_with_same_parent(scene):
    object_sets={}
    for o in scene.objects:
        if o.type=='EMPTY':
            object_sets[o.name]=[a for a in o.children if a.type=='MESH']
            o.name=o.name+"_parent"
    for parent,objects in object_sets.items():
        if objects:
            objects[0].name=parent
            if len(objects)>1:
                override = bpy.context.copy()
                override["active_object"] = objects[0]
                override["object"] = objects[0]
                override["selected_objects"] = objects
                override["selected_editable_objects"] = objects
                with bpy.context.temp_override(**override):
                    bpy.ops.object.join()
def get_random_uuid():
    return str(uuid.uuid4())
def get_name_to_display(name):
    if "/" in name:
        #substr=name[name.rindex("/")+1:]
        return name[name.rindex("/")+1:] +f" ({name[:name.rindex('/')]})"
    else:
        return name
def get_catalog_by_name(path,name):
    if os.path.isfile(os.path.join(path,"blender_assets.cats.txt")):
        with open(os.path.join(path,"blender_assets.cats.txt"), mode='r',newline="\n",encoding="UTF-8") as catalog_file:
            for l in catalog_file.readlines():
                if not l.startswith("#") and not l.startswith("VERSION") and ":" in l:
                    if l[l.index(":")+1:l.rindex(":")]==name:
                        return l[:l.index(":")]
    return None
def get_all_catalogs(self, context):
    catalogs=[]
    if os.path.isfile(os.path.join(preferences().objects_directory,"blender_assets.cats.txt")):
        with open(os.path.join(preferences().objects_directory,"blender_assets.cats.txt"), mode='r',newline="\n",encoding="UTF-8") as catalog_file:
            for l in catalog_file.readlines():
                if not l.startswith("#") and not l.startswith("VERSION") and ":" in l:
                    catalogs.append((l[:l.index(":")],l[l.index(":")+1:l.rindex(":")]))
    if os.path.isfile(os.path.join(preferences().materials_directory,"blender_assets.cats.txt")):
        with open(os.path.join(preferences().materials_directory,"blender_assets.cats.txt"), mode='r',newline="\n",encoding="UTF-8") as catalog_file:
            for l in catalog_file.readlines():
                if not l.startswith("#") and not l.startswith("VERSION") and ":" in l:
                    catalogs.append((l[:l.index(":")],l[l.index(":")+1:l.rindex(":")]))
    if os.path.isfile(os.path.join(preferences().worlds_directory,"blender_assets.cats.txt")):
        with open(os.path.join(preferences().worlds_directory,"blender_assets.cats.txt"), mode='r',newline="\n",encoding="UTF-8") as catalog_file:
            for l in catalog_file.readlines():
                if not l.startswith("#") and not l.startswith("VERSION") and ":" in l:
                    catalogs.append((l[:l.index(":")],l[l.index(":")+1:l.rindex(":")]))
    if os.path.isfile(os.path.join(preferences().node_groups_directory,"blender_assets.cats.txt")):
        with open(os.path.join(preferences().node_groups_directory,"blender_assets.cats.txt"), mode='r',newline="\n",encoding="UTF-8") as catalog_file:
            for l in catalog_file.readlines():
                if not l.startswith("#") and not l.startswith("VERSION") and ":" in l:
                    catalogs.append((l[:l.index(":")],l[l.index(":")+1:l.rindex(":")]))
    catalogs=sorted(catalogs,key=lambda x:get_name_to_display(x[1]).lower())
    return [("Filename","Filename","Filename"),("Directory","Directory Name","Directory Name"),("NONE","None","None"),("NEW","New/Search","New/Search")]+[(b,get_name_to_display(b),b) for a,b in catalogs]
def get_object_catalogs(self, context):
    catalogs=[]
    if os.path.isfile(os.path.join(preferences().objects_directory,"blender_assets.cats.txt")):
        with open(os.path.join(preferences().objects_directory,"blender_assets.cats.txt"), mode='r',newline="\n",encoding="UTF-8") as catalog_file:
            for l in catalog_file.readlines():
                if not l.startswith("#") and not l.startswith("VERSION") and ":" in l:
                    catalogs.append((l[:l.index(":")],l[l.index(":")+1:l.rindex(":")]))
    catalogs=sorted(catalogs,key=lambda x:get_name_to_display(x[1]).lower())
    return [("NONE","None","None"),("NEW","New/Search","New/Search")]+[(a,get_name_to_display(b),b) for a,b in catalogs]
def create_new_catalog(path,name):
    uuids=[]
    #uuid=get_random_uuid()
    result = [_.start() for _ in re.finditer("/",name )] +[len(name),]
    for a in range(len(result)):
        uuids.append(get_random_uuid())
    if not os.path.isfile(os.path.join(path,"blender_assets.cats.txt")):
        with open(os.path.join(path,"blender_assets.cats.txt"), mode='w+',newline="\n",encoding="UTF-8") as catalog_file:
            catalog_file.writelines(["\nVERSION 1",])
            for index,i in enumerate(result):
                    if not get_catalog_by_name(path,name[:i]):
                        
                        catalog_file.writelines(["\n"+uuids[index]+f":{name[:i]}:{name[:i]}\n",])
    else:
        
        if os.path.isfile(os.path.join(path,"blender_assets.cats.txt")):
            with open(os.path.join(path,"blender_assets.cats.txt"), mode='a',newline="\n",encoding="UTF-8") as catalog_file:
                for index,i in enumerate(result):
                    if not get_catalog_by_name(path,name[:i]):
                        catalog_file.writelines(["\n"+uuids[index]+f":{name[:i]}:{name[:i]}\n",])
    
    return get_catalog_by_name(path,name) if get_catalog_by_name(path,name) else "00000000-0000-0000-0000-000000000000"
def find_scene_from_object(object):
    for s in bpy.data.scenes:
        if object in s.objects[:]:
            return s
    return None
def delete_object_with_data(obj):
    if obj and obj.name in bpy.data.objects:
        data = obj.data
        isMesh = obj.type == 'MESH'
        bpy.data.objects.remove(obj, do_unlink=True)
        if isMesh:
            bpy.data.meshes.remove(data)
def get_bounding_box_of_collection(col):
    corners=[]
    for ob in col.all_objects:
        corners.extend([(ob.matrix_world @ mathutils.Vector(b)) for b in ob.bound_box])
    min_z=min(((mathutils.Vector(b)).z for b in corners))
    max_z=max(((mathutils.Vector(b)).z for b in corners))
    min_x=min(((mathutils.Vector(b)).x for b in corners))
    max_x=max(((mathutils.Vector(b)).x for b in corners))
    min_y=min(((mathutils.Vector(b)).y for b in corners))
    max_y=max(((mathutils.Vector(b)).y for b in corners))
    return mathutils.Vector((min_x,min_y,min_z)),mathutils.Vector((max_x,max_y,max_z))
def get_collections_from_scene(scene):
    return [c for c in bpy.data.collections if scene.user_of_id(c)]
def get_biggest_object_from(col):
    if col.all_objects:

        max_volume_object=col.all_objects[0]
        for object in col.all_objects:
            if object.dimensions.x*object.dimensions.y*object.dimensions.z>max_volume_object.dimensions.x*max_volume_object.dimensions.y*max_volume_object.dimensions.z:
                max_volume_object=object
        return object
    return None
def capture_thumbnail_for_collection(context,col,path, shading="Material",angle='X',add_plane=True):
    #print(object)
    hidden = []
    # if object not in context.window_manager.windows[0].scene.objects[:]:
    #     context.window_manager.windows[0].scene=find_scene_from_object(object) if find_scene_from_object(object) else context.window_manager.windows[0].scene
    #     print("New Scene",context.window_manager.windows[0].scene)
    scene_backup=context.window.scene
    file_format=context.window.scene.render.image_settings.file_format
    scene=context.scene  
    if col not in get_collections_from_scene(scene):
        scene.collection.children.link(col)
    view_area=None  
    filepath=scene.render.filepath

    render_path=os.path.join(path,bpy.path.clean_name(col.name+"_"+get_id(8))+'.png')
    scene.render.filepath=render_path
    cameraBackUp = scene.camera
    camera_data = bpy.data.cameras.new(name='Camera')
    camera_data.clip_end=10000000
    camera_object = bpy.data.objects.new('Camera', camera_data)
    scene.collection.objects.link(camera_object)
    track_object=bpy.data.objects.new("Track",object_data=None)
    scene.collection.objects.link(track_object)
    bottom_left,top_right=get_bounding_box_of_collection(col)
    local_bbox_center =(bottom_left+top_right)/2
    location_backup=[]
    # for ob in col.all_objects:
    #     location_backup.append((ob,ob.location.copy()))
        # ob.location=ob.location-mathutils.Vector((local_bbox_center.x,local_bbox_center.y,0))
    ground_plane=None
    if add_plane:
        mesh_data = bpy.data.meshes.new("TA_Plane")
        polycoords=[(1000,1000,0),(1000,-1000,0),(-1000,-1000,0),(-1000,1000,0)]
        polyIndices = [(i, (i+1) % (len(polycoords)))
                        for i in range(0, len(polycoords))]
        mesh_data.from_pydata(polycoords, polyIndices, [
                                [i for i in range(0, len(polycoords))]])
        mesh_data.validate()
        ground_plane = bpy.data.objects.new("TA_Plane", mesh_data)
        subsurf_mod=ground_plane.modifiers.new(type='SUBSURF',name="TA_SUBSURF")
        subsurf_mod.levels=4
        subsurf_mod.render_levels=4
        subsurf_mod.subdivision_type='SIMPLE'
        scene.collection.objects.link(ground_plane)
        min_z=bottom_left.z
        ground_plane.location=(local_bbox_center.x, local_bbox_center.y,min_z)
    global_bbox_center = local_bbox_center
    track_object.location=(global_bbox_center.x,global_bbox_center.y,global_bbox_center.z)
    dist=max([abs(top_right.x-bottom_left.x),abs(top_right.y-bottom_left.y),abs(top_right.z-bottom_left.z)])*2
    object=get_biggest_object_from(col)
    # print(object)
    if object:
        object_matrix=mathutils.Matrix.LocRotScale(track_object.location, object.rotation_quaternion,[1,1,1])
        camera_object.location=object_matrix@mathutils.Vector((-dist if '-X' in angle else (dist if 'X' in angle else 0),dist if '-Y' in angle else (-dist if 'Y' in angle else 0),dist*(-1 if '-Z' in angle else 1) ))
    else:
        camera_object.location=mathutils.Vector((-dist if '-X' in angle else (dist if 'X' in angle else 0),dist if '-Y' in angle else (-dist if 'Y' in angle else 0),dist*(-1 if '-Z' in angle else 1) ))   
    #object_matrix=mathutils.Matrix.LocRotScale(track_object.location, object.rotation_quaternion,[1,1,1])
    # camera_object.location=mathutils.Vector((-dist if '-X' in angle else (dist if 'X' in angle else 0),dist if '-Y' in angle else (-dist if 'Y' in angle else 0),dist*(-1 if '-Z' in angle else 1) ))
    #camera_object.location=camera_object.location+local_bbox_center
    
    constraint=camera_object.constraints.new('TRACK_TO')
    constraint.name="TA_Track_To"
    constraint.target=track_object
    constraint.up_axis='UP_Y'
    #override = context.copy()
    #override["object"] = camera_object
    with context.temp_override(object=camera_object):
        bpy.ops.constraint.apply(constraint=constraint.name)
    bpy.data.objects.remove(track_object)
    scene.camera = camera_object
    camera_data.lens = 80
    transparent_backup=scene.render.film_transparent
    scene.render.film_transparent=True
    rendexXbackUp = scene.render.resolution_x
    rendexYbackUp = scene.render.resolution_y
    scene.render.resolution_x = 256
    scene.render.resolution_y = 256
    use_nodes=scene.use_nodes
    scene.use_nodes=False
    if not bpy.data.worlds.get("TA_Thumbnail_World"):
        asset_file=os.path.join(os.path.dirname(
                    os.path.abspath(__file__)), "Assets", "Assets.blend" if bpy.app.version>=(3,4,0) else "Assets_Old.blend","World")
        bpy.ops.wm.append(
                directory=asset_file,
                filename="TA_Thumbnail_World", autoselect=False
            )
    backup_world=context.scene.world
    context.scene.world=bpy.data.worlds.get("TA_Thumbnail_World")        
    map_node=context.scene.world.node_tree.nodes.get("Mapping")
    context.scene.world.node_tree.nodes.get("Background").inputs[1].default_value=preferences().world_strength
    if map_node:
        map_node.inputs[2].default_value=[map_node.inputs[2].default_value[0],map_node.inputs[2].default_value[1],0.785398-camera_object.rotation_euler.z]          
    if shading=='Material':
        
        context.space_data.region_3d.view_perspective = 'CAMERA'
        shading = context.space_data.shading.type
        context.space_data.shading.type = 'MATERIAL'
        overlays = context.space_data.overlay.show_overlays
        context.space_data.overlay.show_overlays = False
        unselected=[a for a in scene.objects if a.name not in col.all_objects]
        for obj in unselected:
            if obj.hide_viewport == False:
                obj.hide_viewport = True
                hidden.append(obj)
        bpy.ops.render.opengl(write_still=True)
        for obj in hidden:
            obj.hide_viewport = False
        context.space_data.overlay.show_overlays = overlays
        
        context.space_data.shading.type = shading
    elif shading=='Eevee':
        engine = scene.render.engine
        scene.render.engine = 'BLENDER_EEVEE'
        unselected=[a for a in scene.objects if a.name not in col.all_objects]
        for obj in unselected:
            if obj.hide_render == False:
                obj.hide_render = True
                hidden.append(obj)
        bpy.ops.render.render(write_still=True)
        scene.render.engine = engine
        for obj in hidden:
            obj.hide_render = False
    else:
        samples=context.scene.cycles.samples
        context.scene.cycles.samples=64
        engine = scene.render.engine
        scene.render.engine = 'CYCLES'
        unselected=[a for a in scene.objects if a not in col.all_objects[:]]
        
        for obj in unselected:
            if obj.hide_render == False:
                obj.hide_render = True
                hidden.append(obj)
        bpy.ops.render.render(write_still=True)
        context.scene.cycles.samples=samples
        scene.render.engine = engine
        for obj in hidden:
            obj.hide_render = False
    # for ob,loc in location_backup:
    #     ob.location=loc
    context.scene.render.resolution_x = rendexXbackUp
    context.scene.render.resolution_y = rendexYbackUp
    context.scene.render.film_transparent=transparent_backup
    context.scene.render.filepath=filepath
    context.scene.world=backup_world
    if bpy.data.worlds.get("TA_Thumbnail_World"):
        if bpy.data.worlds.get("TA_Thumbnail_World").node_tree.nodes.get("Environment Texture").image:
            bpy.data.images.remove(bpy.data.worlds.get("TA_Thumbnail_World").node_tree.nodes.get("Environment Texture").image)
        bpy.data.worlds.remove(bpy.data.worlds.get("TA_Thumbnail_World"))
    if camera_object:
           bpy.data.objects.remove(camera_object, do_unlink=True)
    context.scene.camera = cameraBackUp
    context.window.scene=scene_backup
    context.scene.use_nodes=use_nodes
    context.scene.render.image_settings.file_format = file_format
    if ground_plane:
        delete_object_with_data(ground_plane)
    return render_path
def capture_thumbnail(context,object,path, shading="Material",angle='X',add_plane=True):
    hidden = []
    scene_backup=context.window.scene
    
    file_format=context.window.scene.render.image_settings.file_format
    context.window.scene.render.image_settings.file_format = 'PNG'
    file_format=context.scene.render.image_settings.file_format
    context.scene.render.image_settings.file_format = 'PNG'
    if object not in context.scene.objects[:]:
        context.window.scene=find_scene_from_object(object) if find_scene_from_object(object) else context.scene
    filepath=context.scene.render.filepath
    render_path=os.path.join(path,object.name+"_"+get_id(8)+".png")
    context.scene.render.filepath=render_path
    use_nodes=context.scene.use_nodes
    context.scene.use_nodes=False
    cameraBackUp = context.scene.camera
    camera_data = bpy.data.cameras.new(name='Camera')
    camera_data.clip_end=10000000
    camera_object = bpy.data.objects.new('Camera', camera_data)
    context.scene.collection.objects.link(camera_object)
    track_object=bpy.data.objects.new("Track",object_data=None)
    
    context.scene.collection.objects.link(track_object)
    context.scene.camera = camera_object
    local_bbox_center = 0.125 * sum((mathutils.Vector(b) for b in object.bound_box), mathutils.Vector())
    ground_plane=None
    if add_plane:
        mesh_data = bpy.data.meshes.new("TA_Plane")
        polycoords=[(1000,1000,0),(1000,-1000,0),(-1000,-1000,0),(-1000,1000,0)]
        polyIndices = [(i, (i+1) % (len(polycoords)))
                        for i in range(0, len(polycoords))]
        mesh_data.from_pydata(polycoords, polyIndices, [
                                [i for i in range(0, len(polycoords))]])
        mesh_data.validate()
        ground_plane = bpy.data.objects.new("TA_Plane", mesh_data)
        subsurf_mod=ground_plane.modifiers.new(type='SUBSURF',name="TA_SUBSURF")
        subsurf_mod.levels=4
        subsurf_mod.render_levels=4
        subsurf_mod.subdivision_type='SIMPLE'
        context.scene.collection.objects.link(ground_plane)
        min_z=min(((object.matrix_world @ mathutils.Vector(b)).z for b in object.bound_box))
        ground_plane.location=(object.location.x, object.location.y,min_z)
    global_bbox_center = object.matrix_world @ local_bbox_center
    track_object.location=global_bbox_center
    dist=max(object.dimensions[:])*2
    object_matrix=mathutils.Matrix.LocRotScale(object.location, object.rotation_quaternion,[1,1,1])
    #print(object_matrix,object.matrix_world)
    if angle!='VIEW':
        
        object_matrix=mathutils.Matrix.LocRotScale(track_object.location, object.rotation_quaternion,[1,1,1])
        #print(object_matrix,dist)
        camera_object.location=object_matrix@mathutils.Vector((-dist if '-X' in angle else (dist if 'X' in angle else 0),dist if '-Y' in angle else (-dist if 'Y' in angle else 0),dist*(-1 if '-Z' in angle else 1) ))
    else:
        camera_object.matrix_world = context.space_data.region_3d.view_matrix
        camera_object.matrix_world.invert()
        camera_object.location=camera_object.location.normalized()*dist*2
        #bpy.ops.view3d.camera_to_view_selected()
        #camera_object.location,_,_=context.space_data.region_3d.view_matrix.decompose()#*context.space_data.region_3d.view_distance
    #camera_object.location=(camera_object.location[0]*object.scale[0],camera_object.location[1]*object.scale[1],camera_object.location[2]*object.scale[2])
    constraint=camera_object.constraints.new('TRACK_TO')
    constraint.name="TA_Track_To"
    constraint.target=track_object
    constraint.up_axis='UP_Y'
    #override = context.copy()
    #override["object"] = camera_object
    with context.temp_override(object=camera_object):
        bpy.ops.constraint.apply(constraint=constraint.name)
    bpy.data.objects.remove(track_object)
    
    camera_data.lens = 80 
    # camera_data.type='ORTHO'
    # camera_data.ortho_scale=dist/2
    camera_data.clip_start=0.0001
    transparent_backup=context.scene.render.film_transparent
    context.scene.render.film_transparent=True
    rendexXbackUp = context.scene.render.resolution_x
    rendexYbackUp = context.scene.render.resolution_y
    context.scene.render.resolution_x = THUMBNAIL_RESOLUTION
    context.scene.render.resolution_y = THUMBNAIL_RESOLUTION
    if not bpy.data.worlds.get("TA_Thumbnail_World"):
        asset_file=os.path.join(os.path.dirname(
                    os.path.abspath(__file__)), "Assets", "Assets.blend" if bpy.app.version>=(3,4,0) else "Assets_Old.blend","World")
        bpy.ops.wm.append(
                directory=asset_file,
                filename="TA_Thumbnail_World", autoselect=False
            )
    backup_world=context.scene.world
    context.scene.world=bpy.data.worlds.get("TA_Thumbnail_World")
    map_node=context.scene.world.node_tree.nodes.get("Mapping")
    context.scene.world.node_tree.nodes.get("Background").inputs[1].default_value=preferences().world_strength
    #print(camera_object.rotation_euler.z,map_node.inputs[2].default_value[2])
    if map_node:
        map_node.inputs[2].default_value=[map_node.inputs[2].default_value[0],map_node.inputs[2].default_value[1],0.785398-camera_object.rotation_euler.z]
    if shading=="Material":
        context.space_data.region_3d.view_perspective = 'CAMERA'
        shading = context.space_data.shading.type
        context.space_data.shading.type = 'MATERIAL'
        overlays = context.space_data.overlay.show_overlays
        context.space_data.overlay.show_overlays = False
        unselected =[a for a in context.scene.objects if a !=object and a!=ground_plane]
        for obj in unselected:
            if obj.hide_get() == False:
                try:
                    obj.hide_set(True)
                    hidden.append(obj)
                except:
                    pass
        bpy.ops.render.opengl(write_still=True)
        for obj in hidden:
            obj.hide_set(False)
        context.space_data.overlay.show_overlays = overlays
        
        context.space_data.shading.type = shading
    elif shading=='Cycles':
        samples=context.scene.cycles.samples
        context.scene.cycles.samples=64
        engine = context.scene.render.engine
        context.scene.render.engine = 'CYCLES'
        unselected = [a for a in context.scene.objects if a !=object and a!=ground_plane]
        for obj in unselected:
            if obj.type != "LIGHT" and obj.hide_render == False:
                obj.hide_render = True
                hidden.append(obj)
        og_hide_render=object.hide_render
        object.hide_render=False
        bpy.ops.render.render(write_still=True)
        object.hide_render=og_hide_render
        context.scene.render.engine = engine
        context.scene.cycles.samples=samples
        for obj in hidden:
            obj.hide_render = False
    else:
        engine = context.scene.render.engine
        context.scene.render.engine = 'BLENDER_EEVEE'
        unselected = [a for a in context.scene.objects if a !=object and a!=ground_plane]
        for obj in unselected:
            if obj.type != "LIGHT" and obj.hide_render == False:
                obj.hide_render = True
                hidden.append(obj)
        og_hide_render=object.hide_render
        object.hide_render=False
        bpy.ops.render.render(write_still=True)
        object.hide_render=og_hide_render
        context.scene.render.engine = engine
        for obj in hidden:
            obj.hide_render = False
    
    context.scene.render.resolution_x = rendexXbackUp
    context.scene.render.resolution_y = rendexYbackUp
    context.scene.render.film_transparent=transparent_backup
    context.scene.render.filepath=filepath
    context.scene.world=backup_world
    if bpy.data.worlds.get("TA_Thumbnail_World"):
        if bpy.data.worlds.get("TA_Thumbnail_World").node_tree.nodes.get("Environment Texture").image:
            bpy.data.images.remove(bpy.data.worlds.get("TA_Thumbnail_World").node_tree.nodes.get("Environment Texture").image)
        bpy.data.worlds.remove(bpy.data.worlds.get("TA_Thumbnail_World"))
    if camera_object:
           bpy.data.objects.remove(camera_object, do_unlink=True)
    context.scene.camera = cameraBackUp
    context.window.scene=scene_backup
    context.scene.use_nodes=use_nodes
    context.scene.render.image_settings.file_format = file_format
    if ground_plane:
        delete_object_with_data(ground_plane)
    return render_path
def create_thumbnail(context, ob,shading='Solid',angle='X',add_plane=False):
    if not os.path.isdir(os.path.join(preferences().objects_directory,'Thumbnails')):
        os.mkdir(os.path.join(preferences().objects_directory,'Thumbnails'))
    if shading!='Solid':
        if isinstance(ob,bpy.types.Collection):
            path=capture_thumbnail_for_collection(context,ob,os.path.join(preferences().objects_directory,'Thumbnails'),shading,angle=angle,add_plane=add_plane)
        else:
        #capture_thumbnail_for_collection(context,bpy.data.collections['Test'],os.path.join(preferences().objects_directory,'Thumbnails'),shading,angle=angle,add_plane=add_plane)
            path=capture_thumbnail(context,ob,os.path.join(preferences().objects_directory,'Thumbnails'),shading,angle=angle,add_plane=add_plane)
        if os.path.isfile(path):
            with context.temp_override(id=ob):
                bpy.ops.ed.lib_id_load_custom_preview(filepath=path)
            # os.remove(path)
    else:
        ob.asset_generate_preview()
def get_all_blend_files(path):
    blends = []
    for path, subdirs, files in os.walk(path):
        for name in files:
            if name.endswith('.blend'):
                blends.append(os.path.join(path, name))
                #print(os.path.join(path, name))
    return blends
def get_all_files_with_extension(path,ext):
    blends = []
    for path, subdirs, files in os.walk(path):
        for name in files:
            if len(os.path.splitext(name))>1 and os.path.splitext(name)[1].replace(".","").lower() == ext.lower():
                
                blends.append(os.path.join(path, name))
                #print(os.path.join(path, name))
    return blends
def remove_library(path):
    for index,a in enumerate(bpy.context.preferences.filepaths.asset_libraries[:]):
        if a.path==path:
            bpy.ops.preferences.asset_library_remove(index=index)
            break
    
def ensure_library(name,path):
    lib_path=[a for a in bpy.context.preferences.filepaths.asset_libraries if a.name==name]
    if not lib_path:
        same_path=[a for a in bpy.context.preferences.filepaths.asset_libraries if a.path==path]
        if not same_path:
            init_libs=bpy.context.preferences.filepaths.asset_libraries[:]
            try:
                bpy.ops.preferences.asset_library_add(directory='')
            except:
                pass
            new_lib=[a for a in bpy.context.preferences.filepaths.asset_libraries[:] if a not in init_libs]
            if new_lib:
                new_lib=new_lib[0]
                new_lib.name=name
                new_lib.path=path
                return new_lib.path
            else:
                None
            
        else:
            same_path[0].name=name
            return same_path[0].path
    else:
        lib_path[0].path=path
        return lib_path[0].path
def get_value(a):
    try:
        value=[round(b,2) for b in a]
    except:
        value=round(a,2)
    return value
def preferences():
    return bpy.context.preferences.addons[__package__].preferences

# preview_already_called=[]
# def create_material_preview2(objects,objects_to_write):
#     obs=objects[:]
#     ob=obs[0]
#     global preview_already_called
    
#     if ob not in preview_already_called:
#         preview_already_called.append(ob)
#         ob.asset_generate_preview()
#     if any(ob.preview.image_pixels[:]):
#         obs.pop(0)
#     else:
#         return 0.1
#     if obs:
#         bpy.app.timers.register(functools.partial(create_material_preview,obs,objects_to_write))
#     else:
#         global st
#         print(time.time()-st)
#         write_to_blend('MATS',objects_to_write,'COPY',"True_Assets_Materials")
#     return None
def create_material_preview(objects,images,pack=False,ambient_cg=False,remove_mats=False):
    obs=objects[:]
    while obs:
        ob=obs[0]
        if any(ob.preview.image_pixels[:]):
            obs.pop(0)
        else:
            return 0.1
    packed_images=[]
    if pack:
        for i in images:
            packed_images.append(i)
            i.pack()
    write_to_blend('AMBIENTCG_MATS' if ambient_cg else 'MATS',objects+images,'COPY',"True_Assets_Materials")
    for ob in objects:
        remove_duplicate_images(ob)
        if remove_mats:
            try:
                bpy.data.materials.remove(ob)
            except Exception:
                pass
            
    for i in packed_images:
        i.unpack(method='REMOVE')
    return None
def sortTextures(textures,name=""):
    BaseColor = None
    Roughness = None
    Normal = None
    Height = None
    Metal = None
    Specular = None
    opacity = None
    ao=None
    emit=None
    len = 0
    preview=None
    BaseColorTags = ['diffuse', 'diff', 'albedo', 'base', 'col', 'color']
    RoughnessTags = ['roughness', 'rough', 'rgh','gloss']
    NormalTags = ['normal', 'nor', 'nrm', 'nrml', 'norm']
    HeightTags = ['displacement', 'displace', 'disp',
                  'dsp', 'height', 'heightmap', 'bmp', 'bump']
    MetallicTags = ['metallic', 'metalness', 'mtl']
    SpecularTags = ['specularity', 'specular', 'spec', 'spc']
    opacityTags = ['opacity', 'alpha', ' transparen']
    aoTags=['ao','occlusion']
    emitTags=['emis','emit']
    previewTags=['preview',]
    
    for tex in textures:
        if any(x in tex.lower() for x in BaseColorTags) and BaseColor is None:
            BaseColor = tex
            len = len+1
        elif any(x in tex.lower() for x in RoughnessTags) and Roughness is None:
            Roughness = tex
            len = len+1
        elif any(x in tex.lower() for x in NormalTags) and Normal is None:
            Normal = tex
            len = len+1
        elif any(x in tex.lower() for x in HeightTags) and Height is None:
            Height = tex
            len = len+1
        elif any(x in tex.lower() for x in MetallicTags) and Metal is None:
            Metal = tex
            len = len+1
        elif any(x in tex.lower() for x in SpecularTags) and Specular is None:
            Specular = tex
            len = len+1
        elif any(x in tex.lower() for x in opacityTags) and opacity is None:
            opacity = tex
            len = len+1
        elif any(x in tex.lower() for x in aoTags) and ao is None:
            ao = tex
            len = len+1
        elif any(x in tex.lower() for x in emitTags) and emit is None:
            emit = tex
            len = len+1
        elif (any(x in tex.lower() for x in previewTags) or os.path.splitext(tex.lower())[0]==name.lower()) and preview is None:
            
            preview = tex
            len = len+1
    return[BaseColor, Roughness, Normal, Height, Metal, Specular, opacity,ao,emit, len,preview]
from pathlib import Path
def load_textures_from_dir(dir_path,name,quality):
    all_maps=os.listdir(dir_path)
    maps=sortTextures(all_maps,name)
    new_texture=bpy.context.scene.true_terrain.tt.assets.textures.get(name.lower(),None)
    if not new_texture:
        new_texture = bpy.context.scene.true_terrain.tt.assets.textures.new(name.lower(),False)
    # quality=get_quality(maps)[0]
    if new_texture and quality:
        if quality not in [a.display_name for a in new_texture.quality_levels.values()]:
            tex=new_texture
            tex.display_name=name
            ql=tex.quality_levels.new(quality)
            
            
            if maps[0]:
                ql.col=Path(os.path.join(dir_path,maps[0]))
            if maps[1]:
                ql.rough=Path(os.path.join(dir_path,maps[1]))
            if maps[2]:
                ql.nrm=Path(os.path.join(dir_path,maps[2]))
            if maps[3]:
                ql.disp=Path(os.path.join(dir_path,maps[3]))
            if maps[5]:
                ql.spec=Path(os.path.join(dir_path,maps[5]))
            if maps[7]:
                ql.ao=Path(os.path.join(dir_path,maps[7]))
            if maps[10]:
                tex.icon_path=Path(os.path.join(dir_path,maps[10]))
            elif os.path.exists(os.path.join(ADDONS_DIR, 'TA-Previews', name+".png")):
                tex.icon_path=Path(os.path.join(ADDONS_DIR, 'TA-Previews', name+".png"),)
            bpy.context.scene.true_terrain.tt.save_json()
        bpy.context.scene.true_terrain.tt.assets.textures.reload()
    return new_texture
def get_quality(content):
    if content:
        if "1K" in content[0] or "1024" in content[0]:
            return "1K", 1
        elif "2K" in content[0] or "2048" in content[0]:
            return "2K", 2
        elif "4K" in content[0] or "4096" in content[0]:
            return "4K", 4
        elif "8K" in content[0] or "8192" in content[0]:
            return "8K", 8
        elif "6K" in content[0] or "6144" in content[0]:
            return "6K", 6
        elif "12K" in content[0]:
            return "12K", 12
        elif "16K" in content[0]:
            return "16K", 16
    return "", 4
def largest_substr(str1, str2):
    #print(str1,str2)
    matcher = difflib.SequenceMatcher(None, str1, str2)
    match = matcher.find_longest_match(0, len(str1), 0, len(str2))
    #print(match,str1[0: match.size])
    if match.a!=0 or match.b!=0:
        return largest_substr(str1[:max(match.a,match.b)],str2[:max(match.a,match.b)])
    return str1[0: match.size] if match.a == 0 and match.b == 0 else ""
def longest_substr(str1, str2):
    for i in range(min(len(str1), len(str2))):
        if str1[i] != str2[i]: return str1[0:i]
    return str1[0:i+1]
def distribute_mat_names(original_strings):
    strings = original_strings.copy()
    groups = []
    while len(strings) > 0:
        group = [strings[0]]
        rem_strings = []
        substr = strings[0]
        strings.pop(0)
        for string in strings:
            new_substr = longest_substr(string, substr)
            #print(new_substr)
            res = re.findall("[A-Za-z]*", string)[0]
            max_length=len(res)
            if new_substr and len(new_substr) > max_length:
                group.append(string)
                substr = new_substr
            else:
                rem_strings.append(string)
        groups.append({"contents": group, "common": substr})
        strings = rem_strings
        #print(groups)
    return groups
def distribute(original_strings):
    strings = original_strings.copy()
    groups = []
    while len(strings) > 0:
        group = [strings[0]]
        rem_strings = []
        substr = strings[0]
        strings.pop(0)
        for string in strings:
            new_substr = longest_substr(string, substr)
            #print(new_substr)
            if "_" in string:
                max_length=string.index("_")
            else:
                max_length=1
            if new_substr and len(new_substr) > max_length:
                group.append(string)
                substr = new_substr
            else:
                rem_strings.append(string)
        groups.append({"contents": group, "common": substr})
        strings = rem_strings
        #print(groups)
    return groups
def clear_mat_name(name):
    return name.replace("1K","").replace("2K","").replace("4K","").replace("6K","").replace("8K","").replace("12K","").replace("12K","").strip()
def create_world_preview(objects,pack,ambient_cg=False):
    obs=objects[:]
    while obs:
        ob=obs[0]
        if any(ob.preview.image_pixels[:]):
            obs.pop(0)
        else:
            return 0.1
    #if ob:
    images=[]
    packed_images=[]
    if pack:
        images=[ob.node_tree.nodes.get("True HDRI").node_tree.nodes.get("Environment Texture").image for ob in objects]
        for i in images:
            packed_images.append(i)
            i.pack()
    write_to_blend('AMBIENTCG_WORLDS' if ambient_cg else 'WORLDS',objects+images,'COPY',"True-Assets_HDRIs" if len(objects)>1 else ob.name)
    for i in packed_images:
        i.unpack(method='REMOVE')
        #path=os.path.join(preferences().worlds_directory,os.path.basename(ob.name)+".blend")
        #bpy.data.libraries.write(path
        #            , set(objects), fake_user=True)
    return None
def setup_hdris(context,path,selected_files=[],mark_asset=False,image_filters=["hdr","exr","png","jpeg","jpg"],subfolders=True,catalog=None,pack=False,search="",shader_values={},shader_type='True-HDRI'):
    if not os.path.isdir(os.path.join(preferences().worlds_directory,'Thumbnails')):
        os.mkdir(os.path.join(preferences().worlds_directory,'Thumbnails'))
    created_worlds=[]
    for i, j, y in os.walk(path):
        files = []
        for name in os.listdir(i):
            _,extension=os.path.splitext(name)
            extension=extension.replace(".","")
            if selected_files:
                
                if extension in image_filters and name in [os.path.basename(s) for s in selected_files] and search.lower() in name.lower():
                #if (name.endswith('jpg') or name.endswith('hdr') or name.endswith('exr') or name.endswith('jpeg') or name.endswith('png')) and name in selected_files:
                    files.append(os.path.basename(name))
            else:
                if extension in image_filters and search.lower() in name.lower():
                #if name.endswith('jpg') or name.endswith('hdr') or name.endswith('exr') or name.endswith('jpeg') or name.endswith('png'):
                    files.append(os.path.basename(name))
        for f in files:
            hdri_path=os.path.join(i,f)
            if bpy.data.worlds.get(shader_type) is None:
                path = os.path.join(os.path.dirname(os.path.abspath(
                    __file__)), "Assets", "Assets.blend" if bpy.app.version>=(3,4,0) else "Assets_Old.blend", "World")
                init_worlds=bpy.data.worlds[:]
                bpy.ops.wm.append(
                    directory=path,
                    filename=shader_type, autoselect=False
                )
                new_world=[a for a in bpy.data.worlds if a not in init_worlds]

                if new_world:
                    new_world=new_world[0]
                    
                    new_world.name=os.path.splitext(f)[0]
                    node_group=new_world.node_tree.nodes.get(shader_type)
                    set_toggles(node_group)
                    for inp,value in shader_values.items():
                        try:
                            inp=[a.name for a in node_group.inputs if  a.name.lower().startswith(inp.lower())][0]
                            node_group.inputs[inp].default_value=eval(value)
                        except Exception as e:
                            print(inp,value,e)
                    image_node=node_group.node_tree.nodes.get("Environment Texture")
                    image_node.image=bpy.data.images.load(hdri_path)
                    preview_image=image_node.image.copy()
                    #if preview_image.size[0]*preview_image.size[0]>100000000:
                    preview_image.scale(1000,500)
                    og_file_format=context.scene.render.image_settings.file_format
                    context.scene.render.image_settings.file_format = 'PNG'

                    preview_image.save_render(os.path.join(preferences().worlds_directory,'Thumbnails',image_node.image.name), scene=None)
                    context.scene.render.image_settings.file_format=og_file_format
                    bpy.data.images.remove(preview_image)
                    created_worlds.append((new_world,os.path.join(preferences().worlds_directory,'Thumbnails',image_node.image.name)))
                    if context.scene.world:
                        new_world.cycles.sampling_method=context.scene.world.cycles.sampling_method
                        new_world.cycles.sample_map_resolution=context.scene.world.cycles.sample_map_resolution
                        new_world.cycles.max_bounces=context.scene.world.cycles.max_bounces
        if not subfolders:
            break
    if mark_asset and created_worlds:
        for w,h in created_worlds:
            w.asset_mark()
            w.asset_data.tags.new(w.name)
            if catalog!="00000000-0000-0000-0000-000000000000":
                w.asset_data.catalog_id=catalog
            with context.temp_override(id=w):
                bpy.ops.ed.lib_id_load_custom_preview(filepath=h)
            # os.remove(h)
            
            # w.asset_generate_preview()
            bpy.app.timers.register(functools.partial(create_world_preview,[w,],pack))
def import_hdri(og_world,path,quality,mark_asset=False,catalog=None,pack=False,shader_values={},context=None,shader_type='True-HDRI'):
    if not os.path.isdir(os.path.join(preferences().worlds_directory,'Thumbnails')):
        os.mkdir(os.path.join(preferences().worlds_directory,'Thumbnails'))
    created_worlds=[]
    override = context.copy()
    temp_override=context.temp_override
    hdri_path=path
    if bpy.data.worlds.get(shader_type) is None:
        path = os.path.join(os.path.dirname(os.path.abspath(
            __file__)), "Assets", "Assets.blend" if bpy.app.version>=(3,4,0) else "Assets_Old.blend", "World")
        init_worlds=bpy.data.worlds[:]
        bpy.ops.wm.append(
            directory=path,
            filename=shader_type, autoselect=False
        )
        new_world=[a for a in bpy.data.worlds if a not in init_worlds]

        if new_world:
            new_world=new_world[0]
            
            new_world.name=os.path.splitext(os.path.basename(hdri_path))[0]
            new_world.name=og_world.name
            new_world.name=new_world.name+'---'+quality
            new_world.ta_quality=quality
            new_world.ta_shader_type=shader_type
            node_group=new_world.node_tree.nodes.get(shader_type)
            node_group.is_ta_node=True
            for inp,value in shader_values.items():
                try:
                    if 'Tonemapped' in inp:
                        value='1' if 'tonemap' in new_world.name.lower() else '0'
                    inp=[a.name for a in node_group.inputs if  a.name.lower().startswith(inp.lower())][0]
                    node_group.inputs[inp].default_value=eval(value)
                except Exception as e:
                    print(inp,value,e)
            set_toggles(node_group)
            image_node=node_group.node_tree.nodes.get("Environment Texture")
            image_node.image=bpy.data.images.load(hdri_path)
            preview_image=image_node.image.copy()
            #if preview_image.size[0]*preview_image.size[0]>100000000:
            preview_image.scale(1000,500)
            og_file_format=context.scene.render.image_settings.file_format
            context.scene.render.image_settings.file_format = 'PNG'

            preview_image.save_render(os.path.join(preferences().worlds_directory,'Thumbnails',image_node.image.name), scene=None)
            context.scene.render.image_settings.file_format=og_file_format
            bpy.data.images.remove(preview_image)
            created_worlds.append((new_world,os.path.join(preferences().worlds_directory,'Thumbnails',image_node.image.name)))

            if context and context.scene.world:
                new_world.cycles.sampling_method=context.scene.world.cycles.sampling_method
                new_world.cycles.sample_map_resolution=context.scene.world.cycles.sample_map_resolution
                new_world.cycles.max_bounces=context.scene.world.cycles.max_bounces

    if mark_asset and created_worlds:
        for w,h in created_worlds:
            w.asset_mark()
            w.asset_data.tags.new(w.name)
            if catalog!="00000000-0000-0000-0000-000000000000":
                w.asset_data.catalog_id=create_new_catalog(preferences().downloads_directory,f"HDRI/{quality}/{catalog}")
            
            with temp_override(id=w):
                bpy.ops.ed.lib_id_load_custom_preview(filepath=h)
            bpy.app.timers.register(functools.partial(create_world_preview,[w,],pack,True))
    bpy.data.worlds.remove(og_world)
    return created_worlds[0][0]
def get_matching_catalog(name,catalogs):
    for c in sorted(catalogs,key=lambda x:len(x)):
        if c.lower() in name.lower():
            return c
    return None
def set_toggles(node_group):
    i=1
    for inp in node_group.inputs:
        if '*t' in inp.name:
            if inp.default_value==1:
                setattr(node_group.ta_toggles,f'toggle_{i}',True)
            i+=1
def setup_materials(type,folder_path,selected_files,only_basecolor=False, lib_path=None, mark_asset=False,catalog=None,pack=False,search="",shader_values={}):
    mats_created = []
    filename = type
    images_used=[]
    file_path = os.path.join(os.path.dirname(
        os.path.abspath(__file__)), "Assets", "Assets.blend" if bpy.app.version>=(3,4,0) else "Assets_Old.blend", "Material")
    if os.path.isdir(folder_path):
        matInit = []

        for m in bpy.data.materials:
            matInit.append(m.name)
        bpy.ops.wm.append(
            directory=file_path,
            filename=filename, autoselect=False
        )
        matAfter = []

        for m in bpy.data.materials:
            matAfter.append(m.name)
        mat = [a for a in matAfter if a not in matInit]
        mat = bpy.data.materials.get(mat[0])
        for i, j, y in os.walk(folder_path):
            files = []
            
            for name in os.listdir(i):
                if selected_files:
                    if (name.endswith('jpg') or name.endswith('png') or name.endswith('jpeg') or name.endswith('exr') or name.endswith('tif')) and name in [os.path.basename(s) for s in selected_files] and search.lower() in name.lower():
                        
                        files.append(os.path.basename(name))
                else:
                    if (name.endswith('jpg') or name.endswith('png') or name.endswith('jpeg') or name.endswith('exr') or name.endswith('tif')) and search.lower() in name.lower():
                        
                        files.append(os.path.basename(name))
            if len(files) > 1:
                groups = distribute(files)
                for group in groups:
                    tex = sortTextures(group['contents'])
                    if tex[9] > 1 or (tex[0] is not None and only_basecolor):
                        newMat = mat.copy()
                        newMat.use_fake_user = True
                        nodes_to_arrange=[]
                        newMat.name = group['common'].replace(
                            '_', ' ')+get_quality(group['contents'])[0]
                        if type!='Standard':
                            node_group=newMat.node_tree.nodes.get(type)
                            node_group.node_tree=node_group.node_tree.copy()
                            node_group.name=newMat.name
                            node_group.node_tree.name=newMat.name
                            node_group.is_ta_node=True
                            for inp,value in shader_values.items():
                                try:
                                    inp=[a.name for a in node_group.inputs if  a.name.lower().startswith(inp.lower())][0]
                                    node_group.inputs[inp].default_value=eval(value)
                                except Exception as e:
                                    print(inp,value,e)
                            if tex[2]:
                                try:
                                    if 'dx' in tex[2].lower():
                                        node_group.inputs['DX to OGL Normals*t'].default_value=0
                                    if 'gl' in tex[2].lower():
                                        node_group.inputs['DX to OGL Normals*t'].default_value=1
                                except Exception:
                                    pass
                                
                            set_toggles(node_group)
                        else:
                            node_group=newMat
                        
                        if node_group:
                            if node_group.node_tree.nodes.get("AO"):
                                if tex[7] is not None:
                                    node_group.node_tree.nodes["AO"].image = bpy.data.images.load(
                                        os.path.join(i, tex[7]))
                                    node_group.node_tree.links.new(node_group.node_tree.nodes["AO"].outputs[0],node_group.node_tree.nodes["AO Mix"].inputs[7 if len(node_group.node_tree.nodes["AO Mix"].inputs)>5 else 2])
                                    images_used.append(node_group.node_tree.nodes["AO"].image)
                                    nodes_to_arrange.append(node_group.node_tree.nodes["AO"])
                            if node_group.node_tree.nodes.get("BaseColor"):
                                if tex[0] is not None:
                                    node_group.node_tree.nodes["BaseColor"].image = bpy.data.images.load(
                                        os.path.join(i, tex[0]))
                                    images_used.append(node_group.node_tree.nodes["BaseColor"].image)
                                    nodes_to_arrange.append(node_group.node_tree.nodes["BaseColor"])
                            if  node_group.node_tree.nodes.get("Metallic"):
                                if tex[4] is not None:
                                    node_group.node_tree.nodes["Metallic"].image = bpy.data.images.load(
                                        os.path.join(i, tex[4]))
                                    images_used.append(node_group.node_tree.nodes["Metallic"].image)
                                    node_group.node_tree.nodes["Metallic"].image.colorspace_settings.is_data = True
                                    nodes_to_arrange.append(node_group.node_tree.nodes["Metallic"])
                                else:
                                    node_group.node_tree.nodes.remove(
                                        node_group.node_tree.nodes["Metallic"])
                            if node_group.node_tree.nodes.get("Specular"):
                                if tex[5] is not None:
                                    node_group.node_tree.nodes["Specular"].image = bpy.data.images.load(
                                        os.path.join(i, tex[5]))
                                    images_used.append(node_group.node_tree.nodes["Specular"].image)
                                    node_group.node_tree.nodes["Specular"].image.colorspace_settings.is_data = True
                                    nodes_to_arrange.append(node_group.node_tree.nodes["Specular"])
                                else:
                                    node_group.node_tree.nodes.remove(
                                        node_group.node_tree.nodes["Specular"])
                            if node_group.node_tree.nodes.get("Roughness"):
                                if tex[1] is not None:
                                    node_group.node_tree.nodes["Roughness"].image = bpy.data.images.load(
                                        os.path.join(i, tex[1]))
                                    images_used.append(node_group.node_tree.nodes["Roughness"].image)
                                    node_group.node_tree.nodes["Roughness"].image.colorspace_settings.is_data = True
                                    nodes_to_arrange.append(node_group.node_tree.nodes["Roughness"])
                                else:
                                    node_group.node_tree.nodes.remove(
                                        node_group.node_tree.nodes["Roughness"])
                            if node_group.node_tree.nodes.get("Emission"):
                                if tex[8] is not None:
                                    node_group.node_tree.nodes["Emission"].image = bpy.data.images.load(
                                        os.path.join(i, tex[8]))
                                    images_used.append(node_group.node_tree.nodes["Emission"].image)
                                    nodes_to_arrange.append(node_group.node_tree.nodes["Emission"])
                                else:
                                    node_group.node_tree.nodes.remove(
                                        node_group.node_tree.nodes["Emission"])
                            if node_group.node_tree.nodes.get("Height"):
                                if tex[3] is not None:
                                    node_group.node_tree.nodes["Height"].image = bpy.data.images.load(
                                        os.path.join(i, tex[3]))
                                    images_used.append(node_group.node_tree.nodes["Height"].image)
                                    node_group.node_tree.nodes["Height"].image.colorspace_settings.is_data = True
                                    nodes_to_arrange.append(node_group.node_tree.nodes["Height"])
                                else:
                                    node_group.node_tree.nodes.remove(
                                        node_group.node_tree.nodes["Height"])
                            if node_group.node_tree.nodes.get("Normal"):
                                if tex[2] is not None:
                                    node_group.node_tree.nodes["Normal"].image = bpy.data.images.load(
                                        os.path.join(i, tex[2]))
                                    images_used.append(node_group.node_tree.nodes["Normal"].image)
                                    node_group.node_tree.nodes["Normal"].image.colorspace_settings.is_data = True
                                    nodes_to_arrange.append(node_group.node_tree.nodes["Normal"])
                                    
                                else:
                                    node_group.node_tree.nodes.remove(
                                        node_group.node_tree.nodes["Normal"])
                            
                                
                            if node_group.node_tree.nodes.get("Opacity"):
                                if tex[6] is not None:
                                    newMat.blend_method = 'CLIP' if newMat.blend_method == 'OPAQUE' else newMat.blend_method
                                    node_group.node_tree.nodes["Opacity"].image = bpy.data.images.load(
                                        os.path.join(i, tex[6]))
                                    images_used.append(node_group.node_tree.nodes["Opacity"].image)
                                    node_group.node_tree.nodes["Opacity"].image.colorspace_settings.is_data = True
                                    nodes_to_arrange.append(node_group.node_tree.nodes["Opacity"])
                                else:
                                    node_group.node_tree.nodes.remove(
                                        node_group.node_tree.nodes["Opacity"])
                            if len(nodes_to_arrange)>1:
                                for index,n in enumerate(nodes_to_arrange[1:]):
                                    n.location.y=nodes_to_arrange[index].location.y-(nodes_to_arrange[index].height*(3 if not nodes_to_arrange[index].hide else 1))
                            if mark_asset:

                                newMat.asset_mark()
                                newMat.asset_data.tags.new(newMat.name)
                                if catalog!="00000000-0000-0000-0000-000000000000" and catalog!='AUTO':
                                    newMat.asset_data.catalog_id=catalog
                                newMat.asset_generate_preview()
                            mats_created.append(newMat)
        
        cats = distribute_mat_names([clear_mat_name(m.name) for m in mats_created])
        catalogs=[]
        for cat in cats:
            catalogs.append(''.join([i for i in cat['common'] if not i.isdigit()]).title())
        if mark_asset and catalog=="AUTO":
            for m in mats_created:
                cat=get_matching_catalog(m.name,catalogs)
                if cat:
                    m.asset_data.catalog_id=create_new_catalog(preferences().materials_directory,cat)
        if mark_asset and mats_created:
            bpy.app.timers.register(functools.partial(create_material_preview,mats_created,images_used,pack))
        bpy.data.materials.remove(mat)
    return mats_created
def append_material(file_path,name):
    with bpy.data.libraries.load(file_path) as (data_from,data_to):
            for mat in data_from.materials:
                if mat==name:
                    data_to.materials.append(mat)
    if data_to.materials:
        return data_to.materials[0]
    return None
def append_world(file_path,name):
    with bpy.data.libraries.load(file_path) as (data_from,data_to):
            for mat in data_from.worlds:
                if mat==name:
                    data_to.worlds.append(mat)
    if data_to.worlds:
        return data_to.worlds[0]
    return None
def import_material(og_mat,type,folder_path, lib_path=None, mark_asset=True,catalog=None,pack=False,shader_values={},object=None,thumbnail=None,context=None,quality='All',author="AmbientCG",use_as_tt_layer=False):
    mats_created = []
    temp_override=context.temp_override
    filename = type
    images_used=[]
    file_path = os.path.join(os.path.dirname(
        os.path.abspath(__file__)), "Assets", "Assets.blend" if bpy.app.version>=(3,4,0) else "Assets_Old.blend")
    if os.path.isdir(folder_path):
        
        for i, j, y in os.walk(folder_path):
            files = []
            
            for name in os.listdir(i):
                if (name.endswith('jpg') or name.endswith('png') or name.endswith('jpeg') or name.endswith('exr') or name.endswith('tif')):
                    
                    files.append(os.path.basename(name))
            
            if len(files) > 1:
                groups = distribute(files)
                for group in groups:
                    tex = sortTextures(group['contents'])
                    
                    if tex[9] > 1 or (tex[0] is not None):
                        mat=append_material(file_path,filename)
                        newMat = mat.copy()
                        
                        bpy.data.materials.remove(mat)
                        newMat.use_fake_user = True
                        nodes_to_arrange=[]
                        newMat.name = og_mat.name+"---"+get_quality(group['contents'])[0]+f"-{os.path.splitext(group['contents'][0])[1].replace('.','').upper()}"
                        
                        newMat.ta_quality=get_quality(group['contents'])[0]+f"-{os.path.splitext(group['contents'][0])[1].replace('.','').upper()}"
                        newMat.ta_shader_type=type
                        
                        if type!='Standard':
                            node_group=newMat.node_tree.nodes.get(type)
                            node_group.node_tree=node_group.node_tree.copy()
                            node_group.name=og_mat.name
                            node_group.node_tree.name=og_mat.name
                            node_group.is_ta_node=True
                            for inp,value in shader_values.items():
                                try:
                                    inp=[a.name for a in node_group.inputs if  a.name.lower().startswith(inp.lower())][0]
                                    node_group.inputs[inp].default_value=eval(value)
                                except Exception as e:
                                    print(inp,value,e)
                            if tex[2]:
                                try:
                                    if 'dx' in tex[2].lower():
                                        node_group.inputs['DX to OGL Normals*t'].default_value=0
                                    if 'gl' in tex[2].lower():
                                        node_group.inputs['DX to OGL Normals*t'].default_value=1
                                    if author=='ShareTextures':
                                        node_group.inputs['DX to OGL Normals*t'].default_value=1
                                except Exception:
                                    pass
                                
                            set_toggles(node_group)
                        else:
                            node_group=newMat
                        
                        if node_group:
                            if node_group.node_tree.nodes.get("AO"):
                                if tex[7] is not None:
                                    node_group.node_tree.nodes["AO"].image = bpy.data.images.load(
                                        os.path.join(i, tex[7]))
                                    node_group.node_tree.links.new(node_group.node_tree.nodes["AO"].outputs[0],node_group.node_tree.nodes["AO Mix"].inputs[7 if len(node_group.node_tree.nodes["AO Mix"].inputs)>5 else 2])
                                    images_used.append(node_group.node_tree.nodes["AO"].image)
                                    nodes_to_arrange.append(node_group.node_tree.nodes["AO"])
                            if node_group.node_tree.nodes.get("BaseColor"):
                                if tex[0] is not None:
                                    node_group.node_tree.nodes["BaseColor"].image = bpy.data.images.load(
                                        os.path.join(i, tex[0]))
                                    images_used.append(node_group.node_tree.nodes["BaseColor"].image)
                                    nodes_to_arrange.append(node_group.node_tree.nodes["BaseColor"])
                            if  node_group.node_tree.nodes.get("Metallic"):
                                if tex[4] is not None:
                                    node_group.node_tree.nodes["Metallic"].image = bpy.data.images.load(
                                        os.path.join(i, tex[4]))
                                    images_used.append(node_group.node_tree.nodes["Metallic"].image)
                                    node_group.node_tree.nodes["Metallic"].image.colorspace_settings.is_data = True
                                    nodes_to_arrange.append(node_group.node_tree.nodes["Metallic"])
                                else:
                                    node_group.node_tree.nodes.remove(
                                        node_group.node_tree.nodes["Metallic"])
                            if node_group.node_tree.nodes.get("Specular"):
                                if tex[5] is not None:
                                    node_group.node_tree.nodes["Specular"].image = bpy.data.images.load(
                                        os.path.join(i, tex[5]))
                                    images_used.append(node_group.node_tree.nodes["Specular"].image)
                                    node_group.node_tree.nodes["Specular"].image.colorspace_settings.is_data = True
                                    nodes_to_arrange.append(node_group.node_tree.nodes["Specular"])
                                else:
                                    node_group.node_tree.nodes.remove(
                                        node_group.node_tree.nodes["Specular"])
                            if node_group.node_tree.nodes.get("Roughness"):
                                if tex[1] is not None:
                                    node_group.node_tree.nodes["Roughness"].image = bpy.data.images.load(
                                        os.path.join(i, tex[1]))
                                    images_used.append(node_group.node_tree.nodes["Roughness"].image)
                                    node_group.node_tree.nodes["Roughness"].image.colorspace_settings.is_data = True
                                    nodes_to_arrange.append(node_group.node_tree.nodes["Roughness"])
                                else:
                                    node_group.node_tree.nodes.remove(
                                        node_group.node_tree.nodes["Roughness"])
                            if node_group.node_tree.nodes.get("Emission"):
                                if tex[8] is not None:
                                    node_group.node_tree.nodes["Emission"].image = bpy.data.images.load(
                                        os.path.join(i, tex[8]))
                                    images_used.append(node_group.node_tree.nodes["Emission"].image)
                                    nodes_to_arrange.append(node_group.node_tree.nodes["Emission"])
                                else:
                                    node_group.node_tree.nodes.remove(
                                        node_group.node_tree.nodes["Emission"])
                            if node_group.node_tree.nodes.get("Height"):
                                if tex[3] is not None:
                                    node_group.node_tree.nodes["Height"].image = bpy.data.images.load(
                                        os.path.join(i, tex[3]))
                                    images_used.append(node_group.node_tree.nodes["Height"].image)
                                    node_group.node_tree.nodes["Height"].image.colorspace_settings.is_data = True
                                    nodes_to_arrange.append(node_group.node_tree.nodes["Height"])
                                else:
                                    node_group.node_tree.nodes.remove(
                                        node_group.node_tree.nodes["Height"])
                            if node_group.node_tree.nodes.get("Normal"):
                                if tex[2] is not None:
                                    node_group.node_tree.nodes["Normal"].image = bpy.data.images.load(
                                        os.path.join(i, tex[2]))
                                    images_used.append(node_group.node_tree.nodes["Normal"].image)
                                    node_group.node_tree.nodes["Normal"].image.colorspace_settings.is_data = True
                                    nodes_to_arrange.append(node_group.node_tree.nodes["Normal"])
                                else:
                                    node_group.node_tree.nodes.remove(
                                        node_group.node_tree.nodes["Normal"])
                            
                                
                            if node_group.node_tree.nodes.get("Opacity"):
                                if tex[6] is not None:
                                    newMat.blend_method = 'CLIP' if newMat.blend_method == 'OPAQUE' else newMat.blend_method
                                    node_group.node_tree.nodes["Opacity"].image = bpy.data.images.load(
                                        os.path.join(i, tex[6]))
                                    images_used.append(node_group.node_tree.nodes["Opacity"].image)
                                    node_group.node_tree.nodes["Opacity"].image.colorspace_settings.is_data = True
                                    nodes_to_arrange.append(node_group.node_tree.nodes["Opacity"])
                                else:
                                    node_group.node_tree.nodes.remove(
                                        node_group.node_tree.nodes["Opacity"])
                            if len(nodes_to_arrange)>1:
                                for index,n in enumerate(nodes_to_arrange[1:]):
                                    n.location.y=nodes_to_arrange[index].location.y-(nodes_to_arrange[index].height*(3 if not nodes_to_arrange[index].hide else 1))
                            
                            if mark_asset:

                                newMat.asset_mark()
                                newMat.asset_data.tags.new(newMat.name.replace('--',''))
                                newMat.asset_data.author=author
                                if catalog and catalog!="00000000-0000-0000-0000-000000000000" and catalog!='AUTO':
                                    newMat.asset_data.catalog_id=create_new_catalog(preferences().downloads_directory,f"Materials/{quality}/{catalog}")
                                if thumbnail and os.path.isfile(thumbnail):
                                    with temp_override(id=newMat):
                                        bpy.ops.ed.lib_id_load_custom_preview(filepath=thumbnail)
                                else:
                                    newMat.asset_generate_preview()
                            mats_created.append(newMat)
                            break
                            
        
        # cats = distribute_mat_names([clear_mat_name(m.name) for m in mats_created])
        # catalogs=[]
        # for cat in cats:
        #     catalogs.append(''.join([i for i in cat['common'] if not i.isdigit()]).title())
        # if mark_asset and catalog=="AUTO":
        #     for m in mats_created:
        #         cat=get_matching_catalog(m.name,catalogs)
        #         if cat:
        #             m.asset_data.catalog_id=create_new_catalog(preferences().downloads_directory,cat)
        if mark_asset and mats_created:
            bpy.app.timers.register(functools.partial(create_material_preview,mats_created,images_used,pack,True,use_as_tt_layer))
        if object:
            objects=[object]
        else:
            objects=bpy.data.objects[:]
        if not use_as_tt_layer:
            for obj in objects:
                for slot in obj.material_slots:
                    if slot.material==og_mat:
                        slot.material=mats_created[0]
        bpy.data.materials.remove(og_mat)
    return mats_created[0]

def get_images_from_material(mat):
    images=[]
    for n in mat.node_tree.nodes:
        if n.type=='GROUP':
            images.extend(get_images_from_material(n))
        else:
            if (n.type=='TEX_IMAGE' or n.type=='TEX_ENVIRONMENT') and n.image:
                images.append(n.image)
    return images
def switch_world_shader(og_world,shader_type,shader_values={},context=None):
    if not os.path.isdir(os.path.join(preferences().worlds_directory,'Thumbnails')):
        os.mkdir(os.path.join(preferences().worlds_directory,'Thumbnails'))
    created_worlds=[]
    
    images=get_images_from_material(og_world)
    if images and bpy.data.worlds.get(shader_type) is None:
        image=images[0]
        path = os.path.join(os.path.dirname(os.path.abspath(
            __file__)), "Assets", "Assets.blend" if bpy.app.version>=(3,4,0) else "Assets_Old.blend", "World")
        init_worlds=bpy.data.worlds[:]
        bpy.ops.wm.append(
            directory=path,
            filename=shader_type, autoselect=False
        )
        new_world=[a for a in bpy.data.worlds if a not in init_worlds]

        if new_world:
            new_world=new_world[0]
            temp_name=og_world.name
            og_world.name=og_world.name+"."
            new_world.name = temp_name
            new_world.ta_quality=og_world.ta_quality
            new_world.ta_shader_type=shader_type
            node_group=new_world.node_tree.nodes.get(shader_type)
            node_group.is_ta_node=True
            for inp,value in shader_values.items():
                try:
                    if 'Tonemapped' in inp:
                        value='1' if 'tonemap' in image.name.lower() else '0'
                    inp=[a.name for a in node_group.inputs if  a.name.lower().startswith(inp.lower())][0]
                    node_group.inputs[inp].default_value=eval(value)
                except Exception as e:
                    print(inp,value,e)
            set_toggles(node_group)
            image_node=node_group.node_tree.nodes.get("Environment Texture")
            image_node.image=image
            created_worlds.append(new_world)

            if context and context.scene.world:
                new_world.cycles.sampling_method=context.scene.world.cycles.sampling_method
                new_world.cycles.sample_map_resolution=context.scene.world.cycles.sample_map_resolution
                new_world.cycles.max_bounces=context.scene.world.cycles.max_bounces
        for node in og_world.node_tree.nodes:
            if node.label in ['True-HDRI','True-HDRI (Standard)']:
                if getattr(node,'node_tree',None):
                    bpy.data.node_groups.remove(node.node_tree,do_unlink=True)

        bpy.data.worlds.remove(og_world)
    return created_worlds[0]
def switch_material_shader(og_mat,type,shader_values={}):
    mats_created = []
    filename = type
    images_used=[]
    file_path = os.path.join(os.path.dirname(
        os.path.abspath(__file__)), "Assets", "Assets.blend" if bpy.app.version>=(3,4,0) else "Assets_Old.blend")
    matInit = []

    for m in bpy.data.materials:
        matInit.append(m.name)
    
    append_material(file_path,filename)
    matAfter = []

    for m in bpy.data.materials:
        matAfter.append(m.name)
    mat = [a for a in matAfter if a not in matInit]
    mat = bpy.data.materials.get(mat[0])
    images = get_images_from_material(og_mat)
    if len(images) > 1:
        groups = distribute([a.name for a in images])
        for group in groups:
            tex = sortTextures(group['contents'])
            if tex[9] > 1 or (tex[0] is not None):
                newMat = mat.copy()
                bpy.data.materials.remove(mat)
                newMat.use_fake_user = True
                nodes_to_arrange=[]
                temp_name=og_mat.name
                og_mat.name=og_mat.name+"."
                newMat.name = temp_name
                newMat.ta_quality=og_mat.ta_quality
                newMat.ta_shader_type=type
                if type!='Standard':
                    node_group=newMat.node_tree.nodes.get(type)
                    node_group.node_tree=node_group.node_tree.copy()

                    node_group.name=newMat.name
                    node_group.node_tree.name=newMat.name
                    node_group.is_ta_node=True
                    for inp,value in shader_values.items():
                        try:
                            inp=[a.name for a in node_group.inputs if  a.name.lower().startswith(inp.lower())][0]
                            node_group.inputs[inp].default_value=eval(value)
                        except Exception as e:
                            print(inp,value,e)
                else:
                    node_group=newMat
                
                if node_group:
                    if node_group.node_tree.nodes.get("AO"):
                        if tex[7] is not None:
                            node_group.node_tree.nodes["AO"].image = bpy.data.images.get(tex[7])
                            node_group.node_tree.links.new(node_group.node_tree.nodes["AO"].outputs[0],node_group.node_tree.nodes["AO Mix"].inputs[7 if len(node_group.node_tree.nodes["AO Mix"].inputs)>5 else 2])
                            images_used.append(node_group.node_tree.nodes["AO"].image)
                            nodes_to_arrange.append(node_group.node_tree.nodes["AO"])
                    if node_group.node_tree.nodes.get("BaseColor"):
                        if tex[0] is not None:
                            node_group.node_tree.nodes["BaseColor"].image = bpy.data.images.get(tex[0])
                            images_used.append(node_group.node_tree.nodes["BaseColor"].image)
                            nodes_to_arrange.append(node_group.node_tree.nodes["BaseColor"])
                    if  node_group.node_tree.nodes.get("Metallic"):
                        if tex[4] is not None:
                            node_group.node_tree.nodes["Metallic"].image = bpy.data.images.get(tex[4])
                            images_used.append(node_group.node_tree.nodes["Metallic"].image)
                            node_group.node_tree.nodes["Metallic"].image.colorspace_settings.is_data = True
                            nodes_to_arrange.append(node_group.node_tree.nodes["Metallic"])
                        else:
                            node_group.node_tree.nodes.remove(
                                node_group.node_tree.nodes["Metallic"])
                    if node_group.node_tree.nodes.get("Specular"):
                        if tex[5] is not None:
                            node_group.node_tree.nodes["Specular"].image = bpy.data.images.get(tex[5])
                            images_used.append(node_group.node_tree.nodes["Specular"].image)
                            node_group.node_tree.nodes["Specular"].image.colorspace_settings.is_data = True
                            nodes_to_arrange.append(node_group.node_tree.nodes["Specular"])
                        else:
                            node_group.node_tree.nodes.remove(
                                node_group.node_tree.nodes["Specular"])
                    if node_group.node_tree.nodes.get("Roughness"):
                        if tex[1] is not None:
                            node_group.node_tree.nodes["Roughness"].image = bpy.data.images.get(tex[1])
                            images_used.append(node_group.node_tree.nodes["Roughness"].image)
                            node_group.node_tree.nodes["Roughness"].image.colorspace_settings.is_data = True
                            nodes_to_arrange.append(node_group.node_tree.nodes["Roughness"])
                            if "gloss" in tex[1].lower():
                                    roughness_socket=node_group.node_tree.nodes["Roughness"].outputs[0].links[0].to_socket
                                    invert_node=node_group.node_tree.nodes.new('ShaderNodeInvert')
                                    invert_node.location=node_group.node_tree.nodes["Roughness"].location.x+300,node_group.node_tree.nodes["Roughness"].location.y
                                    node_group.node_tree.links.new(node_group.node_tree.nodes["Roughness"].outputs[0],invert_node.inputs[1])
                                    node_group.node_tree.links.new(invert_node.outputs[0],roughness_socket)
                                    node_group.node_tree.nodes["Roughness"].label="Glossiness"
                                    node_group.node_tree.nodes["Roughness"].name="Glossiness"
                        else:
                            node_group.node_tree.nodes.remove(
                                node_group.node_tree.nodes["Roughness"])
                    if node_group.node_tree.nodes.get("Emission"):
                        if tex[8] is not None:
                            node_group.node_tree.nodes["Emission"].image = bpy.data.images.get(tex[8])
                            images_used.append(node_group.node_tree.nodes["Emission"].image)
                            nodes_to_arrange.append(node_group.node_tree.nodes["Emission"])
                        else:
                            node_group.node_tree.nodes.remove(
                                node_group.node_tree.nodes["Emission"])
                    if node_group.node_tree.nodes.get("Height"):
                        if tex[3] is not None:
                            node_group.node_tree.nodes["Height"].image = bpy.data.images.get(tex[3])
                            images_used.append(node_group.node_tree.nodes["Height"].image)
                            node_group.node_tree.nodes["Height"].image.colorspace_settings.is_data = True
                            nodes_to_arrange.append(node_group.node_tree.nodes["Height"])
                        else:
                            node_group.node_tree.nodes.remove(
                                node_group.node_tree.nodes["Height"])
                    if node_group.node_tree.nodes.get("Normal"):
                        if tex[2] is not None:
                            node_group.node_tree.nodes["Normal"].image = bpy.data.images.get(tex[2])
                            images_used.append(node_group.node_tree.nodes["Normal"].image)
                            node_group.node_tree.nodes["Normal"].image.colorspace_settings.is_data = True
                            nodes_to_arrange.append(node_group.node_tree.nodes["Normal"])
                        else:
                            node_group.node_tree.nodes.remove(
                                node_group.node_tree.nodes["Normal"])
                    
                        
                    if node_group.node_tree.nodes.get("Opacity"):
                        if tex[6] is not None:
                            newMat.blend_method = 'CLIP' if newMat.blend_method == 'OPAQUE' else newMat.blend_method
                            node_group.node_tree.nodes["Opacity"].image = bpy.data.images.get(tex[6])
                            images_used.append(node_group.node_tree.nodes["Opacity"].image)
                            node_group.node_tree.nodes["Opacity"].image.colorspace_settings.is_data = True
                            nodes_to_arrange.append(node_group.node_tree.nodes["Opacity"])
                        else:
                            node_group.node_tree.nodes.remove(
                                node_group.node_tree.nodes["Opacity"])
                    if len(nodes_to_arrange)>1:
                        for index,n in enumerate(nodes_to_arrange[1:]):
                            n.location.y=nodes_to_arrange[index].location.y-(nodes_to_arrange[index].height*(3 if not nodes_to_arrange[index].hide else 1))
                    
                    mats_created.append(newMat)
        
        objects=bpy.data.objects[:]
        for obj in objects:
            for slot in obj.material_slots:
                if slot.material==og_mat:
                    slot.material=mats_created[0]
        for node in og_mat.node_tree.nodes:
            if node.label in ['True-Plastic','True-Uber','True-BSDF','True-Organic']:
                if getattr(node,'node_tree',None):
                    bpy.data.node_groups.remove(node.node_tree,do_unlink=True)

        bpy.data.materials.remove(og_mat)
    return mats_created[0]
def get_value(a):
    try:
        value=[round(b,2) for b in a]
    except:
        value=round(a,2)
    return value

def check_if_default(mat):
    if set([m.type for m in mat.node_tree.nodes if not m.mute]).difference(set(["BSDF_PRINCIPLED","OUTPUT_MATERIAL"])):
        return False
    else:
        bsdf=[m for m in mat.node_tree.nodes if not m.mute and m.type=='BSDF_PRINCIPLED']
        if bsdf:
            bsdf=bsdf[0]
            for i in range(len(bsdf.inputs)-3):
                if i not in inputs_to_skip_compare and get_value(bsdf.inputs[i].default_value)!=default_values_bsdf[i]:
                    return False
                
            return True
def does_support_thumbnails(ob):
    return isinstance(ob,bpy.types.Material) or isinstance(ob,bpy.types.World) or  (isinstance(ob,bpy.types.Object) and ob.type in supports_thumbnails and not (isinstance(ob,bpy.types.Object) and ob.type=='MESH' and len(ob.evaluated_get(bpy.context.evaluated_depsgraph_get()).data.polygons)<1)) or (isinstance(ob,bpy.types.NodeTree) and ob.outputs)

def get_available_name(path,name):
    name=name.replace(".blend","")
    new_name=name
    if os.path.isdir(path):
        files = [file for file in os.listdir(path)]
        final_name = name
        i = 1
        while (final_name+".blend" in files):
            final_name = name+f"_{i}"
            i = i+1
        new_name = final_name
    return new_name+".blend"
def write_to_blend(type, objects,override,custom_name=None):
    blend_name=os.path.basename(bpy.data.filepath)
    imported_data=[]
    name=None
    if objects:
        
        if custom_name:
            blend_name=custom_name+".blend"
        if type=="OBJECTS" or type=="COLLECTIONS":
            if override=='COPY':
                blend_name=get_available_name(preferences().objects_directory,blend_name)
            path=os.path.join(preferences().objects_directory,blend_name)
                
        elif type=="MATS":
            if override=='COPY':
                blend_name=get_available_name(preferences().materials_directory,blend_name)
            path=os.path.join(preferences().materials_directory,blend_name)
        elif type=="AMBIENTCG_MATS":
            blend_name=objects[0].name
            mat=objects[0]

            name=mat.name.split('---')[0]
            quality=mat.name.split('---')[1]
            mat.name=name
            if is_duplicate(mat.name):
                mat.name=mat.name[:-4]
                name=mat.name
            # mat.name=name+'-'+quality
            # if override=='COPY':
            #     blend_name=get_available_name(preferences().downloads_directory,blend_name)
            if not os.path.isdir(os.path.join(preferences().downloads_directory,name)):
                os.makedirs(os.path.join(preferences().downloads_directory,name))
            path=os.path.join(preferences().downloads_directory,name,quality+'.blend')
        elif type=="AMBIENTCG_WORLDS":
            blend_name=objects[0].name
            world=objects[0]

            name=world.name.split('---')[0]
            quality=world.name.split('---')[1]
            world.name=name
            if is_duplicate(world.name):
                world.name=mat.name[:-4]
                name=world.name
            # if override=='COPY':
            #     blend_name=get_available_name(preferences().downloads_directory,blend_name)
            if not os.path.isdir(os.path.join(preferences().downloads_directory,name)):
                os.makedirs(os.path.join(preferences().downloads_directory,name))
            path=os.path.join(preferences().downloads_directory,name,quality+'.blend')
                
        elif type=="WORLDS":
            if override=='COPY':
                blend_name=get_available_name(preferences().worlds_directory,blend_name)
            path=os.path.join(preferences().worlds_directory,blend_name)    
        else:
            if override=='COPY':
                blend_name=get_available_name(preferences().node_groups_directory,blend_name)
            path=os.path.join(preferences().node_groups_directory,blend_name)
        bpy.data.libraries.write(path
                        , set(objects), fake_user=True)
        if name:
            objects[0].name=name+f"-{quality}"
def ensure_preview_and_save_asset(type,objects,objects_to_check_previews,mark_materials,name=None,min_nodes=3,override='COPY',location_backup=[]):

    obs=objects_to_check_previews[:]
    while obs:
        ob=obs[0]
        if any(ob.preview.image_pixels[:]):
            obs.pop(0)
        else:
            return 1
    if objects:
        write_to_blend(type, objects,override,custom_name=name)
        for obj in objects:
            if isinstance(obj,bpy.types.Collection):
                obj.instance_offset=obj.og_instance_offset
        # for obj,loc in location_backup:
        #     obj.location=loc
    if type=='OBJECTS' and mark_materials:
        mats=[]
        for ob in objects:
            for mat in ob.data.materials:
                if mat.name!='Dots Stroke' and mat.users>0:
                    if  mat and mat.use_nodes:
                        if check_if_default(mat) or len(mat.node_tree.nodes)<min_nodes:
                            continue
                    mats.append(mat)
                    if not mat.asset_data:
                        mat.asset_mark()
                        mat.asset_data.tags.new(ob.name)
                    if not (mat.preview and any(mat.preview.image_pixels[:])):
                        mat.asset_generate_preview()
        bpy.app.timers.register(functools.partial(ensure_preview_and_save_asset,"MATS",mats,mats,False,location_backup=location_backup),first_interval=0)
    
    return None
def check_if_volumetric(mat):
    if mat.use_nodes:
        output_nodes=[node for node in mat.node_tree.nodes if node.bl_idname=='ShaderNodeOutputMaterial']
        for node in output_nodes:
            if node.inputs[1].is_linked:
                return True
RENDERING_THUMB_FOR=None
def create_preview(context,type,objects,objects_to_check_previews,filters,force_previews=False,min_nodes=3,quit=False,engine='BLENDER_EEVEE',override='OVERRIDE',clear_materials=True,shading='',angle='XY',add_plane=False,what_to_save='Everything',location_backup=[],catalog='NONE',origin_to_bottom=False):
    global RENDERING_THUMB_FOR
    obs=objects_to_check_previews[:]
    while obs:
        ob=obs[0]
        if any(ob.preview.image_pixels[:]) or ((type=='COLLECTIONS' or type=='OBJECTS') and shading!="Solid") or (type=='MATS' and check_if_volumetric(ob)): 
            
            obs.pop(0)
            if RENDERING_THUMB_FOR==ob:
                RENDERING_THUMB_FOR=None
        else:
            if not RENDERING_THUMB_FOR:
                RENDERING_THUMB_FOR=ob
                
                ob.asset_generate_preview()
            return 1
    if objects and type!='COLLECTIONS':
        write_to_blend(type, objects,override)
        for obj in objects:
            if isinstance(obj,bpy.types.Collection):
                obj.instance_offset=obj.og_instance_offset
        # for obj,loc in location_backup:
        #     obj.location=loc
    if type=="COLLECTIONS":
        Create_Mesh_Assets(context,filters,force_previews,min_nodes,clear_materials,quit,shading,engine=engine,override=override,angle=angle,add_plane=add_plane,collection_objects=objects,what_to_save=what_to_save,location_backup=location_backup,catalog=catalog,origin_to_bottom=origin_to_bottom)
    if type=="OBJECTS":
        Create_Material_Assets(context,filters,force_previews,min_nodes,quit,engine,what_to_save=what_to_save,catalog=catalog)
    elif type=="MATS":
        Create_World_Assets(context,filters,force_previews,min_nodes,quit,engine,what_to_save=what_to_save,catalog=catalog)
    elif type=="WORLDS":
        Create_NodeGroup_Assets(context,filters,force_previews,min_nodes,quit,engine,what_to_save=what_to_save,catalog=catalog)
    else:
        if quit:
            bpy.ops.wm.save_mainfile()
            bpy.ops.wm.quit_blender()
    return None
def is_hidden_in_all_viewlayers(ob):
    for scene in bpy.data.scenes:
        for v in scene.view_layers:
            if ob in v.objects[:] and not ob.hide_get(view_layer=v):
                return False
    return True
def get_all_nodegroups_from_material(node_tree):
    groups=[]
    if node_tree:
        for node in node_tree.nodes:
            if node.type=='GROUP':
                if node.node_tree:
                    groups.append(node.node_tree)
                    groups.extend(get_all_nodegroups_from_material(node.node_tree))
    return groups

def Create_Collection_Assets(context,filters,force_previews,min_nodes,clear_materials=False,quit=False,shading="Solid",engine='BLENDER_EEVEE',override='OVERRIDE',angle='X',add_plane=False,what_to_save='Everything',catalog='NONE',origin_to_bottom=False):
    
    objects=[]
    data_used=[]
    location_backup=[]
    objects_to_check_previews=[]
    if "COLLECTIONS" in filters:
        for scene in bpy.data.scenes:
            if scene.name!='TA_Node_Group_Render_Setup':
                for col in [c for c in bpy.data.collections if scene.user_of_id(c)]:
                    if col.all_objects:
                        bottom_left,top_right=get_bounding_box_of_collection(col)
                        bbox_center=(bottom_left+top_right)/2
                        col.og_instance_offset=col.instance_offset.copy()
                        col.instance_offset=mathutils.Vector((bbox_center.x,bbox_center.y,0))
                        if not col.hide_viewport and col not in data_used and (what_to_save=='Everything' or (what_to_save=='Non-Asset' and not col.asset_data) or (what_to_save=='Marked Assets' and col.asset_data)):
                            data_used.append(col)
                            if not col.asset_data:
                                col.asset_mark()
                                col.asset_data.tags.new(col.name)
                            if catalog!='NONE':
                                col.asset_data.catalog_id=create_new_catalog(preferences().objects_directory,os.path.splitext(os.path.basename(bpy.data.filepath))[0] if catalog=='Filename' else catalog)
                            objects.append(col)
                            if [a for a in col.all_objects if a.type!='EMPTY' and a.type!='LIGHT' and a.display_type!='WIRE'and a.display_type!='BOUNDS' and a.hide_render==False]:
                                objects_to_check_previews.append(col)
                            if not (col.preview and any(col.preview.image_pixels[:])) or force_previews:
                                create_thumbnail(context,col,shading,angle,add_plane=add_plane)
                                #ob.asset_generate_preview()
                        # for obj,loc in location_backup:
                        #     obj.location=loc
    bpy.app.timers.register(functools.partial(
    create_preview, context,"COLLECTIONS",objects,objects_to_check_previews,filters,force_previews,min_nodes,quit,engine=engine,override=override,clear_materials=clear_materials,shading=shading,angle=angle,add_plane=add_plane,what_to_save=what_to_save,location_backup=location_backup,catalog=catalog,origin_to_bottom=origin_to_bottom), first_interval=0)
def set_new_origin_of_object(obj):
    corners=[(mathutils.Vector(b)) for b in obj.bound_box]
    min_z=min(((mathutils.Vector(b)).z for b in corners))
    max_z=max(((mathutils.Vector(b)).z for b in corners))
    min_x=min(((mathutils.Vector(b)).x for b in corners))
    max_x=max(((mathutils.Vector(b)).x for b in corners))
    min_y=min(((mathutils.Vector(b)).y for b in corners))
    max_y=max(((mathutils.Vector(b)).y for b in corners))
    origin= mathutils.Vector(((min_x+max_x)/2,(min_y+max_y)/2,min_z))
    to_move=(obj.matrix_world.inverted() @obj.location)-origin
    obj.data.transform(mathutils.Matrix.Translation(to_move))
    _, rot, sca = obj.matrix_world.decompose()
    matrix_without_location = mathutils.Matrix.LocRotScale([0,0,0], rot, sca)
    obj.location-=matrix_without_location @to_move
def Create_Mesh_Assets(context,filters,force_previews,min_nodes,clear_materials=False,quit=False,shading="Solid",engine='BLENDER_EEVEE',override='OVERRIDE',angle='X',add_plane=False,collection_objects=[],what_to_save='Everything',location_backup=[],catalog='NONE',origin_to_bottom=False):
    
    objects=[]
    data_used=[]
    objects_to_check_previews=[]
    if "OBJECTS" in filters:
        for scene in bpy.data.scenes:
            if scene.name!='TA_Node_Group_Render_Setup':
                for ob in scene.objects:
                    if ob.users>0 and ob.type in filters and not is_hidden_in_all_viewlayers(ob) and not ob.hide_viewport and  ob.data not in data_used  and (what_to_save=='Everything' or (what_to_save=='Non-Asset' and not ob.asset_data) or (what_to_save=='Marked Assets' and ob.asset_data)):
                        data_used.append(ob.data)
                        if ob.type=='MESH' and (ob.display_type == 'WIRE' or ob.display_type == 'BOUNDS'):
                            continue    
                        if ob.type in {'MESH','CURVE','SURFACE','TEXT','META'}:
                            ob.display_type='TEXTURED'
                            if clear_materials:
                                for mat in ob.data.materials:
                                    if mat:
                                        if mat.use_nodes:
                                            for n in get_all_nodegroups_from_material(mat.node_tree):
                                                    n.asset_clear()
                                        mat.asset_clear()
                        if not ob.asset_data:
                            if origin_to_bottom:
                                set_new_origin_of_object(ob)
                            ob.asset_mark()
                            ob.asset_data.tags.new(ob.name)
                            ob.asset_data.tags.new(ob.type.title())
                        if catalog!='NONE':
                                ob.asset_data.catalog_id=create_new_catalog(preferences().objects_directory,os.path.splitext(os.path.basename(bpy.data.filepath))[0] if catalog=='Filename' else catalog)
                        objects.append(ob)
                        if does_support_thumbnails(ob):
                            objects_to_check_previews.append(ob)
                            if not (ob.preview and any(ob.preview.image_pixels[:])) or force_previews:
                                create_thumbnail(context,ob,shading,angle,add_plane=add_plane)
                                #ob.asset_generate_preview()
    
    bpy.app.timers.register(functools.partial(
    create_preview, context,"OBJECTS",objects+collection_objects,objects_to_check_previews,filters,force_previews,min_nodes,quit,engine=engine,override=override,what_to_save=what_to_save,location_backup=location_backup,catalog=catalog), first_interval=0)
def Create_Material_Assets(context,filters,force_previews,min_nodes,quit=False,engine='BLENDER_EEVEE',override='OVERRIDE',what_to_save='Everything',catalog='NONE'):
    mats=[]
    if "MATS" in filters:
        for ob in bpy.data.materials:
            if ob.name!='Dots Stroke' and ob.users>0 and (what_to_save=='Everything' or (what_to_save=='Non-Asset' and not ob.asset_data) or (what_to_save=='Marked Assets' and ob.asset_data)):
                if ob.use_nodes:
                    if check_if_default(ob) or len(ob.node_tree.nodes)<min_nodes:
                        continue
                mats.append(ob)
                if not ob.asset_data:
                    ob.asset_mark()
                    ob.asset_data.tags.new(ob.name)
                if catalog!='NONE':
                    ob.asset_data.catalog_id=create_new_catalog(preferences().materials_directory,os.path.splitext(os.path.basename(bpy.data.filepath))[0] if catalog=='Filename' else catalog)
                if not (ob.preview and any(ob.preview.image_pixels[:])) or force_previews:
                    # bpy.ops.ed.lib_id_generate_preview({'id':ob})
                    # ob.asset_generate_preview()
                    pass
    
    bpy.app.timers.register(functools.partial(create_preview,context,"MATS",mats,mats,filters,force_previews,min_nodes,quit,engine=engine,override=override,what_to_save=what_to_save,catalog=catalog),first_interval=0)

def Create_World_Assets(context,filters,force_previews,min_nodes,quit=False,engine='BLENDER_EEVEE',override='OVERRIDE',what_to_save='Everything',catalog='NONE'):
    worlds=[]
    if "WORLDS" in filters:
        for ob in bpy.data.worlds:
            if (what_to_save=='Everything' or (what_to_save=='Non-Asset' and not ob.asset_data) or (what_to_save=='Marked Assets' and ob.asset_data)):
                worlds.append(ob)
                if not ob.asset_data:
                    ob.asset_mark()
                    ob.asset_data.tags.new(ob.name)
                if catalog!='NONE':
                    ob.asset_data.catalog_id=create_new_catalog(preferences().worlds_directory,os.path.splitext(os.path.basename(bpy.data.filepath))[0] if catalog=='Filename' else catalog)
                if not (ob.preview and any(ob.preview.image_pixels[:])) or force_previews:
                    ob.asset_generate_preview()
    bpy.app.timers.register(functools.partial(create_preview,context,"WORLDS",worlds,worlds,filters,force_previews,min_nodes,quit,engine=engine,override=override,what_to_save=what_to_save,catalog=catalog),first_interval=0)

def Create_NodeGroup_Assets(context,filters,force_previews,min_nodes,quit,engine,override='OVERRIDE',what_to_save='Everything',catalog='NONE'):
    groups=[]
    groups_to_check_previews=[]
    if "NODEGROUPS" in filters:
        for ob in bpy.data.node_groups:
            if ob.bl_idname in {'ShaderNodeTree','GeometryNodeTree'} and (what_to_save=='Everything' or (what_to_save=='Non-Asset' and not ob.asset_data) or (what_to_save=='Marked Assets' and ob.asset_data)):
                
                groups.append(ob)
                ob.asset_mark()
                ob.asset_data.tags.new(ob.name)
                if catalog!='NONE':
                        ob.asset_data.catalog_id=create_new_catalog(preferences().node_groups_directory,os.path.splitext(os.path.basename(bpy.data.filepath))[0] if catalog=='Filename' else catalog)
                if does_support_thumbnails(ob):
                    groups_to_check_previews.append(ob)
                    if not (ob.preview and any(ob.preview.image_pixels[:])) or force_previews:
                        if ob.bl_idname in {'ShaderNodeTree',}:
                            Create_Node_Group_Asset(context,None,ob,engine)
                        else:
                            with context.temp_override(id=ob):
                                bpy.ops.ed.lib_id_load_custom_preview( filepath=os.path.join(os.path.dirname(os.path.abspath(__file__)), "Assets","GeometryNodeGroup.png"))
    bpy.app.timers.register(functools.partial(create_preview,context,"NODEGROUPS",groups,groups_to_check_previews,filters,force_previews,min_nodes,quit=quit,engine=engine,override=override,what_to_save=what_to_save,catalog=catalog),first_interval=0)
def copy_attributes( old_node, new_node):
        for i in range(len(old_node.inputs)):
            
            new_node.inputs[i].default_value=old_node.inputs[i].default_value 
            # if new_node.inputs[i].type=='RGBA':
            #     new_node.inputs[i].default_value=[random.random() for a in range(len(new_node.inputs[i].default_value))]
random_id_source = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789"
def get_id(length=6):
    return "".join(random.sample(random_id_source, length))
def Create_Node_Group_Asset(context,node,node_group,engine="CYCLES",tags="",description=""):
    if not os.path.isdir(os.path.join(preferences().node_groups_directory,"Thumbnails")):
        os.mkdir(os.path.join(preferences().node_groups_directory,"Thumbnails"))
    scene_backup=context.scene
    if not bpy.data.scenes.get('TA_Node_Group_Render_Setup'):
        setup_file_path=os.path.join(os.path.dirname(
                os.path.abspath(__file__)), "Assets", "NodeGroupStudio.blend","Scene")
        bpy.ops.wm.append(
                directory=setup_file_path,
                filename="TA_Node_Group_Render_Setup", autoselect=False
            )
    context.window.scene=bpy.data.scenes.get('TA_Node_Group_Render_Setup')
    material=bpy.data.materials.get('TA_Node_Group_Shader')
    if not material.node_tree.nodes.get("NodeGroup"):
        group_node=material.node_tree.nodes.new(type='ShaderNodeGroup')
        group_node.name="NodeGroup"
    else:
        group_node=material.node_tree.nodes.get("NodeGroup")    
    group_node.node_tree=node_group
    if node:
        copy_attributes(node,group_node)
    if group_node.outputs[0].type=='SHADER':
        material.node_tree.links.new(group_node.outputs[0],material.node_tree.nodes.get('Material Output').inputs[0])
    else:
        material.node_tree.links.new(group_node.outputs[0],material.node_tree.nodes.get('ToConnect').inputs[0])
    r=context.scene.render
    r.engine=engine
    r.image_settings.file_format = 'PNG'
    r.image_settings.quality = 95
    r.resolution_x = THUMBNAIL_RESOLUTION
    r.resolution_y =  THUMBNAIL_RESOLUTION
    r.resolution_percentage = 100
    path=os.path.join(preferences().node_groups_directory,"Thumbnails",node_group.name+get_id(8)+".png")
    r.filepath = path
    bpy.ops.render.render(write_still=True)
    context.window.scene =scene_backup
    try:
        bpy.data.scenes.remove(bpy.data.scenes.get('TA_Node_Group_Render_Setup'))
    except:
        print("Could not remove scene!")
    # cmd = [bpy.app.binary_path]
    # cmd.append("--background")
    # cmd.append("--factory-startup")
    # cmd.append("--python")
    # cmd.append(os.path.join(os.path.dirname(
    #     os.path.abspath(__file__)), "create_nodegroup_preview.py"))
    # cmd.append('--')
    # cmd.append(os.path.join(os.path.dirname(
    #     os.path.abspath(__file__)), "Assets", "NodeGroupStudio.blend"))
    # cmd.append(path)
    # cmd.append(shadertype)
    # cmd.append(os.path.join(bpy.data.filepath,'NodeTree',node_group.name))
    # cmd.append(engine)
    # run(cmd)
    if not node_group.asset_data:
        node_group.asset_mark()
        for t in tags.split(","):
            node_group.asset_data.tags.new(t)    
        node_group.asset_data.description=description
        node_group.asset_data.tags.new(node_group.name)
    if os.path.isfile(path):
        with context.temp_override(id=node_group):
            bpy.ops.ed.lib_id_load_custom_preview(filepath=path)
        # os.remove(path)
    
def create_assets(context,filters,force_previews=False,min_nodes=3,clear_materials=False,quit=False,shading="Solid",engine='BLENDER_EEVEE',override='OVERRIDE',angle='X',add_plane=False,what_to_save='Everything',catalog='NONE',origin_to_bottom=False):
    
    Create_Collection_Assets(context,filters,force_previews,min_nodes,clear_materials,quit,shading,engine=engine,override=override,angle=angle,add_plane=add_plane,what_to_save=what_to_save,catalog=catalog,origin_to_bottom=origin_to_bottom)
def create_assets_from_path(path,filters,objects_path,mats_path,worlds_path,node_groups_path,blends_path,override="SKIP",min_nodes=3,shading="Solid",engine='BLENDER_EEVEE',max_time=300,force_previews=False,clear_materials=False,angle='X',join_children=False,apply_rotation=True,clear_parent=True,catalog="Filename",add_plane=False,join_method='Collection',what_to_save='Everything',pack=False,change_displacement_type=True,remove_subdiv=True,origin_to_bottom=False,save_as_blend=False,skip_hidden=True):
    cmd = [bpy.app.binary_path]
    #cmd.append("--background")
    #cmd.append(path)
    cmd.append("--factory-startup")
    cmd.append("--python")
    cmd.append(os.path.join(os.path.dirname(
        os.path.abspath(__file__)), "mark_assets.py"))
    cmd.append('--')
    cmd.append(":-path-:".join(path))
    cmd.append(objects_path)
    cmd.append(mats_path)
    cmd.append(worlds_path)
    cmd.append(node_groups_path)
    cmd.append(",".join(filters))
    cmd.append(override)
    cmd.append(str(min_nodes))
    cmd.append(shading)
    cmd.append(engine)
    cmd.append(str(max_time))
    cmd.append(str(force_previews))
    cmd.append(str(clear_materials))
    cmd.append(angle)
    cmd.append(str(join_children))
    cmd.append(str(apply_rotation))
    cmd.append(str(clear_parent))
    cmd.append(catalog)
    cmd.append(str(add_plane))
    cmd.append(str(preferences().world_strength))
    cmd.append(join_method)
    cmd.append(what_to_save)
    cmd.append(str(pack))
    cmd.append(str(change_displacement_type))
    cmd.append(str(remove_subdiv))
    cmd.append(str(origin_to_bottom))
    cmd.append(str(save_as_blend))
    cmd.append(str(skip_hidden))
    cmd.append(blends_path)
    cmd.append(bpy.context.preferences.addons['cycles'].preferences.compute_device_type if 'cycles' in bpy.context.preferences.addons.keys() else 'NONE')
    #t1=threading.Thread(target=functools.partial(run,cmd))
    #t1.start()
    run(cmd)
def create_usd_assets_from_path(path,objects_path,override="SKIP",shading="Solid",engine='BLENDER_EEVEE',max_time=300,force_previews=False,angle='X',catalog="Filename",add_plane=False,scale=None,set_frame_range=None,import_cameras=None,import_curves=None,import_lights=None,import_materials=None,import_meshes=None,import_volumes=None,import_subdiv=None,import_instance_proxies=None,import_visible_only=None,create_collection=None,read_mesh_uvs=None,read_mesh_colors=None,prim_path_mask=None,import_guide=None,import_proxy=None,import_render=None,import_usd_preview=None,set_material_blend=None,light_intensity_scale=None,mtl_name_collision_mode=None,import_all_materials=False,import_textures_dir='',import_textures_mode='IMPORT_PACK',tex_name_collision_mode='USE_EXISTING',pack=False,make_collection=True):
    cmd = [bpy.app.binary_path]
    #cmd.append("--background")
    #cmd.append(path)
    cmd.append("--factory-startup")
    
    cmd.append("--python")

    cmd.append(os.path.join(os.path.dirname(
        os.path.abspath(__file__)), "mark_assets_usd.py"))
    cmd.append('--')
    cmd.append(":-path-:".join(path))
    cmd.append(objects_path)
    cmd.append(override)
    cmd.append(shading)
    cmd.append(engine)
    cmd.append(str(max_time))
    cmd.append(str(force_previews))
    cmd.append(angle)
    cmd.append(catalog)
    cmd.append(str(add_plane))
    cmd.append(str(preferences().world_strength))
    cmd.append(str(scale))
    cmd.append(str(set_frame_range))
    cmd.append(str(import_cameras))
    cmd.append(str(import_curves))
    cmd.append(str(import_lights))
    cmd.append(str(import_materials))
    cmd.append(str(import_meshes))
    cmd.append(str(import_volumes))
    cmd.append(str(import_subdiv))
    cmd.append(str(import_instance_proxies))
    cmd.append(str(import_visible_only))
    cmd.append(str(create_collection))
    cmd.append(str(read_mesh_uvs))
    cmd.append(str(read_mesh_colors))
    cmd.append(str(prim_path_mask))
    cmd.append(str(import_guide))
    cmd.append(str(import_proxy))
    cmd.append(str(import_render))
    cmd.append(str(import_usd_preview))
    cmd.append(str(set_material_blend))
    cmd.append(str(light_intensity_scale))
    cmd.append(str(mtl_name_collision_mode))
    cmd.append(str(import_all_materials))
    cmd.append(str(import_textures_mode))
    cmd.append(str(import_textures_dir))
    cmd.append(str(tex_name_collision_mode))
    cmd.append(str(pack))
    cmd.append(str(make_collection))
    cmd.append(bpy.context.preferences.addons['cycles'].preferences.compute_device_type if 'cycles' in bpy.context.preferences.addons.keys() else 'NONE')
    #t1=threading.Thread(target=functools.partial(run,cmd))
    #t1.start()
    # print(cmd)
    run(cmd)
def create_fbx_assets_from_path(path,objects_path,override="SKIP",shading="Solid",engine='BLENDER_EEVEE',max_time=300,force_previews=False,angle='X',catalog="Filename",add_plane=False,use_auto_bone_orientation= None, my_calculate_roll= None, my_bone_length= None, my_leaf_bone= None, use_fix_bone_poses= None, use_fix_attributes= None, use_only_deform_bones= None, use_vertex_animation= None, use_animation= None, my_animation_offset= None, use_animation_prefix= None, use_triangulate= None, my_import_normal= None, use_auto_smooth= None, my_angle= None, my_shade_mode= None, my_scale= None, use_optimize_for_blender= None, use_reset_mesh_origin= None, use_edge_crease= None, my_edge_crease_scale= None, my_edge_smoothing= None, use_import_materials= None, use_rename_by_filename= None, my_rotation_mode= None, use_manual_orientation= None, global_scale= None, bake_space_transform= None, use_custom_normals= None, use_image_search= None, use_alpha_decals= None, decal_offset= None, use_anim= None, anim_offset= None, use_subsurf= None, use_custom_props= None, use_custom_props_enum_as_string= None, ignore_leaf_bones= None, force_connect_children= None, automatic_bone_orientation= None, primary_bone_axis= None, secondary_bone_axis= None, use_prepost_rot= None, axis_forward= None, axis_up= None,importer='Blender',pack=False,single_file=False,make_collection=True):
    cmd = [bpy.app.binary_path]
    #cmd.append("--background")
    #cmd.append(path)
    cmd.append("--factory-startup")
    if importer=='Better FBX':
        cmd.append("--addons")
        cmd.append("better_fbx")
    cmd.append("--python")

    cmd.append(os.path.join(os.path.dirname(
        os.path.abspath(__file__)), "mark_assets_fbx.py"))
    cmd.append('--')
    cmd.append(":-path-:".join(path))
    cmd.append(objects_path)
    cmd.append(override)
    cmd.append(shading)
    cmd.append(engine)
    cmd.append(str(max_time))
    cmd.append(str(force_previews))
    cmd.append(angle)
    cmd.append(catalog)
    cmd.append(str(add_plane))
    cmd.append(str(preferences().world_strength))
    cmd.append(str(use_auto_bone_orientation))
    cmd.append(str(my_calculate_roll))
    cmd.append(str(my_bone_length))
    cmd.append(str(my_leaf_bone))
    cmd.append(str(use_fix_bone_poses))
    cmd.append(str(use_fix_attributes))
    cmd.append(str(use_only_deform_bones))
    cmd.append(str(use_vertex_animation))
    cmd.append(str(use_animation))
    cmd.append(str(my_animation_offset))
    cmd.append(str(use_animation_prefix))
    cmd.append(str(use_triangulate))
    cmd.append(str(my_import_normal))
    cmd.append(str(use_auto_smooth))
    cmd.append(str(my_angle))
    cmd.append(str(my_shade_mode))
    cmd.append(str(my_scale))
    cmd.append(str(use_optimize_for_blender))
    cmd.append(str(use_reset_mesh_origin))
    cmd.append(str(use_edge_crease))
    cmd.append(str(my_edge_crease_scale))
    cmd.append(str(my_edge_smoothing))
    cmd.append(str(use_import_materials))
    cmd.append(str(use_rename_by_filename))
    cmd.append(str(my_rotation_mode))
    cmd.append(str(use_manual_orientation))
    cmd.append(str(global_scale))
    cmd.append(str(bake_space_transform))
    cmd.append(str(use_custom_normals))
    cmd.append(str(use_image_search))
    cmd.append(str(use_alpha_decals))
    cmd.append(str(decal_offset))
    cmd.append(str(use_anim))
    cmd.append(str(anim_offset))
    cmd.append(str(use_subsurf))
    cmd.append(str(use_custom_props))
    cmd.append(str(use_custom_props_enum_as_string))
    cmd.append(str(ignore_leaf_bones))
    cmd.append(str(force_connect_children))
    cmd.append(str(automatic_bone_orientation))
    cmd.append(str(primary_bone_axis))
    cmd.append(str(secondary_bone_axis))
    cmd.append(str(use_prepost_rot))
    cmd.append(str(axis_forward))
    cmd.append(str(axis_up))
    cmd.append(str(importer))
    cmd.append(str(pack))
    cmd.append(str(single_file))
    cmd.append(str(make_collection))
    cmd.append(bpy.context.preferences.addons['cycles'].preferences.compute_device_type if 'cycles' in bpy.context.preferences.addons.keys() else 'NONE')
    #t1=threading.Thread(target=functools.partial(run,cmd))
    #t1.start()
    run(cmd)
def create_obj_assets_from_path(path,objects_path,override="SKIP",shading="Solid",engine='BLENDER_EEVEE',max_time=300,force_previews=False,angle='X',catalog="Filename",add_plane=False,use_auto_bone_orientation= None, my_calculate_roll= None, my_bone_length= None, my_leaf_bone= None, use_fix_bone_poses= None, use_fix_attributes= None, use_only_deform_bones= None, use_vertex_animation= None, use_animation= None, my_animation_offset= None, use_animation_prefix= None, use_triangulate= None, my_import_normal= None, use_auto_smooth= None, my_angle= None, my_shade_mode= None, my_scale= None, use_optimize_for_blender= None, use_reset_mesh_origin= None, use_edge_crease= None, my_edge_crease_scale= None, my_edge_smoothing= None, use_import_materials= None, use_rename_by_filename= None, my_rotation_mode= None,use_edges= None, use_smooth_groups= None, use_split_objects= None, use_split_groups= None, use_groups_as_vgroups= None, use_image_search= None, split_mode= None, global_clamp_size= None,clamp_size=0.52,global_scale=1.0,import_vertex_groups=False,validate_meshes=False,axis_forward='-Z',axis_up='Y',importer='Blender',pack=False,single_file=False,make_collection=True):
    cmd = [bpy.app.binary_path]
    #cmd.append("--background")
    #cmd.append(path)
    cmd.append("--factory-startup")
    if importer=='Better FBX':
        cmd.append("--addons")
        cmd.append("better_fbx")
    cmd.append("--python")

    cmd.append(os.path.join(os.path.dirname(
        os.path.abspath(__file__)), "mark_assets_obj.py"))
    cmd.append('--')
    cmd.append(":-path-:".join(path))
    cmd.append(objects_path)
    cmd.append(override)
    cmd.append(shading)
    cmd.append(engine)
    cmd.append(str(max_time))
    cmd.append(str(force_previews))
    cmd.append(angle)
    cmd.append(catalog)
    cmd.append(str(add_plane))
    cmd.append(str(preferences().world_strength))
    cmd.append(str(use_auto_bone_orientation))
    cmd.append(str(my_calculate_roll))
    cmd.append(str(my_bone_length))
    cmd.append(str(my_leaf_bone))
    cmd.append(str(use_fix_bone_poses))
    cmd.append(str(use_fix_attributes))
    cmd.append(str(use_only_deform_bones))
    cmd.append(str(use_vertex_animation))
    cmd.append(str(use_animation))
    cmd.append(str(my_animation_offset))
    cmd.append(str(use_animation_prefix))
    cmd.append(str(use_triangulate))
    cmd.append(str(my_import_normal))
    cmd.append(str(use_auto_smooth))
    cmd.append(str(my_angle))
    cmd.append(str(my_shade_mode))
    cmd.append(str(my_scale))
    cmd.append(str(use_optimize_for_blender))
    cmd.append(str(use_reset_mesh_origin))
    cmd.append(str(use_edge_crease))
    cmd.append(str(my_edge_crease_scale))
    cmd.append(str(my_edge_smoothing))
    cmd.append(str(use_import_materials))
    cmd.append(str(use_rename_by_filename))
    cmd.append(str(my_rotation_mode))
    cmd.append(str(use_edges))
    cmd.append(str(use_smooth_groups))
    cmd.append(str(use_split_objects))
    cmd.append(str(use_split_groups))
    cmd.append(str(use_groups_as_vgroups))
    cmd.append(str(use_image_search))
    cmd.append(str(split_mode))
    cmd.append(str(global_clamp_size))
    cmd.append(str(global_scale))
    cmd.append(str(clamp_size))
    
    cmd.append(str(import_vertex_groups))
    cmd.append(str(validate_meshes))
    cmd.append(str(axis_forward))
    cmd.append(str(axis_up))
    
    cmd.append(str(importer))
    cmd.append(str(pack))
    cmd.append(str(single_file))
    cmd.append(str(make_collection))
    cmd.append(bpy.context.preferences.addons['cycles'].preferences.compute_device_type if 'cycles' in bpy.context.preferences.addons.keys() else 'NONE')
    #t1=threading.Thread(target=functools.partial(run,cmd))
    #t1.start()
    run(cmd)
def clear_assets_from_external_file(path,f,delete=False,remove_data=False,remove_linked_data=False):
    cmd = [bpy.app.binary_path]
    cmd.append("--background")
    cmd.append(path)
    cmd.append("--factory-startup")
    cmd.append("--python")
    cmd.append(os.path.join(os.path.dirname(
        os.path.abspath(__file__)), "clear_asset_from_external_file.py"))
    cmd.append('--')
    cmd.append(path)
    cmd.append(":--separator--:".join([a[1] for a in f]))
    cmd.append(":--separator--:".join([a[0] for a in f]))
    cmd.append(str(delete))
    cmd.append(str(remove_data))
    cmd.append(str(remove_linked_data))
    if preferences().non_blocking:
        t1=threading.Thread(target=functools.partial(run,cmd))
        t1.start()
        return t1
    else:
        run(cmd)
        return None
def clear_broken_materials(path,f):
    cmd = [bpy.app.binary_path]
    cmd.append("--background")
    cmd.append(path)
    cmd.append("--factory-startup")
    cmd.append("--python")
    cmd.append(os.path.join(os.path.dirname(
        os.path.abspath(__file__)), "clear_broken_materials.py"))
    cmd.append('--')
    cmd.append(path)
    cmd.append(":--separator--:".join([a[1] for a in f]))
    cmd.append(":--separator--:".join([a[0] for a in f]))
    if preferences().non_blocking:
        t1=threading.Thread(target=functools.partial(run,cmd))
        t1.start()
        return t1
    else:
        run(cmd)
        return None
def fix_kitbash_assets(path,f,change_displacement_type=True,remove_subdiv=True):
    cmd = [bpy.app.binary_path]
    cmd.append("--background")
    cmd.append(path)
    cmd.append("--factory-startup")
    cmd.append("--python")
    cmd.append(os.path.join(os.path.dirname(
        os.path.abspath(__file__)), "clean_k3d_objects.py"))
    cmd.append('--')
    cmd.append(path)
    cmd.append(":--separator--:".join([a[1] for a in f]))
    cmd.append(":--separator--:".join([a[0] for a in f]))
    cmd.append(str(change_displacement_type))
    cmd.append(str(remove_subdiv))
    if preferences().non_blocking:
        t1=threading.Thread(target=functools.partial(run,cmd))
        t1.start()
        return t1
    else:
        run(cmd)
        return None
def edit_assets_from_external_file(path,f,workspace):
    cmd = [bpy.app.binary_path]
    #cmd.append("--background")
    cmd.append(path)
    #cmd.append("--factory-startup")
    cmd.append("--python")
    cmd.append(os.path.join(os.path.dirname(
        os.path.abspath(__file__)), "edit_asset.py"))
    cmd.append('--')
    cmd.append(path)
    cmd.append(":--separator--:".join([a[1] for a in f]))
    cmd.append(":--separator--:".join([a[0] for a in f]))
    cmd.append(workspace)
    t1=threading.Thread(target=functools.partial(run,cmd))
    t1.start()
    #run(cmd)
def fetch_ambient_cg_assets(path):
    cmd = [bpy.app.binary_path]
    cmd.append("--background")
    #cmd.append(path)
    cmd.append("--factory-startup")
    cmd.append("--python")
    cmd.append(os.path.join(os.path.dirname(
        os.path.abspath(__file__)), "ambient_cg_fetch.py"))
    cmd.append('--')
    cmd.append(path)
    cmd.append(preferences().downloads_directory)
    run(cmd)
def fetch_assets_for_module(path,module):
    cmd = [bpy.app.binary_path]
    cmd.append("--background")
    #cmd.append(path)
    cmd.append("--factory-startup")
    cmd.append("--python")
    module_path=get_ta_module(module)
    print(module_path)
    cmd.append(os.path.join(module_path, "fetch.py"))
    cmd.append('--')
    cmd.append(path)
    cmd.append(preferences().downloads_directory)
    run(cmd)
def download_asset(path,asset_url,asset_name,asset_type='zip'):
    # cmd = [bpy.app.binary_path]
    # cmd.append("--background")
    # #cmd.append(path)
    # cmd.append("--factory-startup")
    # cmd.append("--python")
    cmd=[sys.executable]
    cmd.append(os.path.join(os.path.dirname(
        os.path.abspath(__file__)), "download_material.py"))
    cmd.append('--')
    cmd.append(path)
    cmd.append(asset_url)
    cmd.append(asset_name)
    cmd.append(asset_type)
    cmd.append(bpy.app.version_string)
    t1=threading.Thread(target=functools.partial(run,cmd))
    t1.start()
    return t1
def create_materials_in_separate_thread(type,folder_path,selected_files,catalog=None,pack=False,search='',shader_values={}):
    cmd = [bpy.app.binary_path]
    #cmd.append("--background")
    #cmd.append(path)
    cmd.append("--factory-startup")
    cmd.append("--python")
    cmd.append(os.path.join(os.path.dirname(
        os.path.abspath(__file__)), "create_materials.py"))
    cmd.append('--')
    cmd.append(type)
    cmd.append(folder_path)
    cmd.append(":--separator--:".join(selected_files) if selected_files else 'None')

    cmd.append(preferences().materials_directory)
    cmd.append(catalog)
    cmd.append(str(pack))
    cmd.append(search)
    cmd.append(":--separator--:".join([a for a,b in shader_values.items()]) if shader_values else 'None')
    cmd.append(":--separator--:".join([str(b) for a,b in shader_values.items()]) if shader_values else 'None')
    run(cmd)
def create_worlds_in_separate_thread(path,selected_files=None,image_filters=["hdr","exr","png","jpeg","jpg"],subfolders=True,catalog=None,pack=False,search="",shader_values={},shader_type='True-HDRI'):
    cmd = [bpy.app.binary_path]
    #cmd.append("--background")
    #cmd.append(path)
    cmd.append("--factory-startup")
    cmd.append("--python")
    cmd.append(os.path.join(os.path.dirname(
        os.path.abspath(__file__)), "create_worlds.py"))
    cmd.append('--')
    cmd.append(path)
    cmd.append(":--separator--:".join(selected_files) if selected_files else 'None')

    cmd.append(preferences().worlds_directory)
    cmd.append(str(subfolders))
    cmd.append(catalog)
    cmd.append(str(pack))
    cmd.append(":--separator--:".join(image_filters))
    cmd.append(":--separator--:".join([a for a,b in shader_values.items()]) if shader_values else 'None')
    cmd.append(":--separator--:".join([str(b) for a,b in shader_values.items()]) if shader_values else 'None')
    cmd.append(search)
    cmd.append(shader_type)
    run(cmd)
def rename_assets(path,object,names,prefix,suffix):
    cmd = [bpy.app.binary_path]
    cmd.append("--background")
    cmd.append(path)
    cmd.append("--factory-startup")
    cmd.append("--python")
    cmd.append(os.path.join(os.path.dirname(
        os.path.abspath(__file__)), "rename_assets.py"))
    cmd.append('--')
    cmd.append(path)
    cmd.append(":--separator--:".join([a[1] for a in object]))
    cmd.append(":--separator--:".join([a[0] for a in object]))
    cmd.append(":--separator--:".join(names))
    cmd.append(prefix)
    cmd.append(suffix)
    if preferences().non_blocking:
        t1=threading.Thread(target=functools.partial(run,cmd))
        t1.start()
        return t1
    else:
        run(cmd)
        return None
def move_assets_to_catalog(path,object,uid):
    cmd = [bpy.app.binary_path]
    cmd.append("--background")
    cmd.append(path)
    cmd.append("--factory-startup")
    cmd.append("--python")
    cmd.append(os.path.join(os.path.dirname(
        os.path.abspath(__file__)), "move_asset_to_catalog.py"))
    cmd.append('--')
    cmd.append(path)
    cmd.append(":--separator--:".join([a[1] for a in object]))
    cmd.append(":--separator--:".join([a[0] for a in object]))
    cmd.append(uid)
    if preferences().non_blocking:
        t1=threading.Thread(target=functools.partial(run,cmd))
        t1.start()
        return t1
    else:
        run(cmd)
        return None
def update_asset_data(path,object,tags,author,desc,add_or_replace,replacement_tags=[]):
    cmd = [bpy.app.binary_path]
    cmd.append("--background")
    cmd.append(path)
    cmd.append("--factory-startup")
    cmd.append("--python")
    cmd.append(os.path.join(os.path.dirname(
        os.path.abspath(__file__)), "update_asset_data.py"))
    cmd.append('--')
    cmd.append(path)
    cmd.append(":--separator--:".join([a[1] for a in object]))
    cmd.append(":--separator--:".join([a[0] for a in object]))
    cmd.append(tags)
    cmd.append(author)
    cmd.append(desc)
    cmd.append(add_or_replace)
    if replacement_tags:
        cmd.append(":--separator--:".join([a[0] for a in replacement_tags]))
        cmd.append(":--separator--:".join([a[1] for a in replacement_tags]))
    if preferences().non_blocking:
        t1=threading.Thread(target=functools.partial(run,cmd))
        t1.start()
        return t1
    else:
        run(cmd)
        return None    
def rerender_thumbnail(path,directory,objects,shading,angle='X',add_plane=False):
    cmd = [bpy.app.binary_path]
    #cmd.append("--background")
    #cmd.append(path[0])
    cmd.append("--factory-startup")
    cmd.append("--python")
    cmd.append(os.path.join(os.path.dirname(
        os.path.abspath(__file__)), "rerender_thumbnails.py"))
    cmd.append('--')
    cmd.append(":--separator--:".join(path))
    names=[]
    types=[]
    for o in objects:
        names.append(":--separator2--:".join([a[0] for a in o]))
        types.append(":--separator2--:".join([a[1] for a in o]))
    cmd.append(":--separator--:".join(names))
    cmd.append(":--separator--:".join(types))
    cmd.append(shading)
    cmd.append(directory)
    cmd.append(angle)
    cmd.append(str(add_plane))
    cmd.append(str(preferences().world_strength))
    cmd.append(bpy.context.preferences.addons['cycles'].preferences.compute_device_type if 'cycles' in bpy.context.preferences.addons.keys() else 'NONE')
    if preferences().non_blocking:
        t1=threading.Thread(target=functools.partial(run,cmd))
        t1.start()
        return t1
    else:
        run(cmd)
        return None
def recheck_if_object_file_exists(self, context):
    if os.path.isfile(os.path.join(preferences().objects_directory,self.name+".blend")):
        self.exists = True
    else:
        self.exists = False
def recheck_if_node_group_file_exists(self, context):
    if os.path.isfile(os.path.join(preferences().node_groups_directory,self.name+".blend")):
        self.exists = True
    else:
        self.exists = False
def ensure_directory_is_set():
    #print(preferences().objects_directory ,os.path.isdir(preferences().materials_directory), os.path.isdir(preferences().worlds_directory),os.path.isdir(preferences().node_groups_directory))
    return os.path.isdir(preferences().objects_directory) and os.path.isdir(preferences().materials_directory) and os.path.isdir(preferences().worlds_directory) and os.path.isdir(preferences().node_groups_directory)
def ensure_all_libraries(refresh=True):
        if refresh:
            preferences().ta_directory=preferences().ta_directory
        objects_lib_path=ensure_library(preferences().object_library_name,path=preferences().objects_directory)
        mats_lib_path=ensure_library(preferences().material_library_name,path=preferences().materials_directory)
        worlds_lib_path=ensure_library(preferences().world_library_name,path=preferences().worlds_directory)
        node_groups_lib_path=ensure_library(preferences().node_group_library_name,path=preferences().node_groups_directory)
        ambientcg_lib_path=ensure_library(preferences().online_library_name,path=preferences().online_assets_directory)
        blends_lib_path=ensure_library(preferences().blends_library_name,path=preferences().blends_directory)
        return objects_lib_path,mats_lib_path,worlds_lib_path,node_groups_lib_path,blends_lib_path,ambientcg_lib_path
def create_key(self,ext,key_names,script_names):
    import winreg as reg
    cwd = os.path.dirname(__file__)
    python_exe = sys.executable
    version_string=bpy.app.version_string[:bpy.app.version_string.rindex(".")]
    ext=f".{ext}" if ext!='blendfile'  else ext
    key_path = f"{ext}"+r'\\shell\\TrueAssets Blender('+f"{version_string})"+r"\\"  
    try:
    # Check if outer key exists
        key = reg.OpenKey(reg.HKEY_CLASSES_ROOT, key_path)
        # existing_scripts=reg.QueryValueEx(key,'SubCommands')[0].split(';')
        # existing_scripts.append(key_name)
        # reg.SetValueEx(key, 'SubCommands',0, reg.REG_SZ, f'{";".join(existing_scripts)};')
    except FileNotFoundError:
        try:
            key = reg.CreateKey(reg.HKEY_CLASSES_ROOT, key_path)
            reg.SetValueEx(key, 'Icon',0, reg.REG_SZ, f'"{bpy.app.binary_path}"')
            reg.SetValueEx(key, 'MUIVerb',0, reg.REG_SZ, f'True Assets')
            reg.SetValueEx(key, 'SubCommands',0, reg.REG_SZ,f'{";".join(key_names)};')
            reg.SetValueEx(key, 'Position',0, reg.REG_SZ, f'Bottom')
        except PermissionError:
            self.report({'WARNING'},"Access is denied! Please start Blender as Administrator!")
        except Exception as e:
            print(e)
    for i,key_name in enumerate(key_names):
        key_path=r"SOFTWARE\Microsoft\Windows\CurrentVersion\Explorer\CommandStore\Shell\\"+key_name
        try:
            # Check if outer key exists
            key = reg.OpenKey(reg.HKEY_LOCAL_MACHINE, key_path)
            command_key=reg.OpenKey(key,'command',access=reg.KEY_ALL_ACCESS)
            if reg.QueryValue(command_key,'')!=python_exe + f' "{cwd}\\call_script_with_multiple_files.py" "{script_names[i]}" "%1"':
                reg.SetValue(command_key, '', reg.REG_SZ,python_exe + f' "{cwd}\\call_script_with_multiple_files.py" "{script_names[i]}" "%1"')
                self.report({'INFO'},"Entry Updated!")
            else:
                self.report({'INFO'},'Entry Already Exists!')
        except FileNotFoundError:
            try:
                key = reg.CreateKey(reg.HKEY_LOCAL_MACHINE, key_path)

                reg.SetValue(key, '', reg.REG_SZ, f'&{key_name}')
                reg.SetValueEx(key, 'Icon',0, reg.REG_SZ, f'"{bpy.app.binary_path}"')

                # create inner key
                key1 = reg.CreateKey(key, r"command")
                reg.SetValue(key1, '', reg.REG_SZ,python_exe + f' "{cwd}\\call_script_with_multiple_files.py" "{script_names[i]}" "%1"')
                self.report({'INFO'},"Entry created!")
            except PermissionError:
                        self.report({'WARNING'},"Access is denied! Please start Blender as Administrator!")
            except Exception as e:
                print(e) 
def delete_key(self,ext,key_names):
    import winreg as reg
    version_string=bpy.app.version_string[:bpy.app.version_string.rindex(".")]
    ext=f".{ext}" if ext!='blendfile'  else ext
    key_path = f"{ext}"+r'\\shell\\TrueAssets Blender('+f"{version_string})"+r"\\"  
    try:
        # Check if outer key exists
        open_key=reg.OpenKey(reg.HKEY_CLASSES_ROOT, key_path)
        reg.DeleteKey(reg.HKEY_CLASSES_ROOT, key_path)
        self.report({'INFO'},"Entry Deleted Successfully!")
    except PermissionError:
        self.report({'WARNING'},"Access is denied! Please start Blender as Administrator!")
    except Exception as e:
        print(e)
    for key_name in key_names:
        key_path=r"SOFTWARE\Microsoft\Windows\CurrentVersion\Explorer\CommandStore\Shell\\"+key_name
        try:
            # Check if outer key exists
            open_key = reg.OpenKey(reg.HKEY_LOCAL_MACHINE, key_path)
            reg.DeleteKey(open_key,"command")
            reg.DeleteKey(reg.HKEY_LOCAL_MACHINE, key_path)
            self.report({'INFO'},"Entry Deleted Successfully!")
        except PermissionError:
            self.report({'WARNING'},"Access is denied! Please start Blender as Administrator!")
        except Exception as e:
            print(e)
def get_library_cat_file(name):
    cats_path=[a for a in bpy.context.preferences.filepaths.asset_libraries if a.name==name]
    if cats_path:
        cats_path=cats_path[0]
    if os.path.isfile(os.path.join(cats_path.path,'blender_assets.cats.txt')):
        return os.path.join(cats_path.path,'blender_assets.cats.txt')
    return None
def isConnected():
    try:
        socket.create_connection(("www.bing.com", 80), timeout=2)
        return True
    except :
        pass
    try:
        response = s.get("http://www.bing.com", timeout=5)
        return True
    except :
        return False

def get_asset_url(name,quality,asset_type='zip'):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        if isConnected():
            try:
                try:
                    response=s.get("http://ambientcg.com/api/v2/full_json",params={'id':name ,'include':'downloadData,imageData,tagData'},timeout=10,verify=False)
                except:
                    response=s.get("http://ambientcg.com/api/v2/full_json",params={'id':name ,'include':'downloadData,imageData,tagData'},timeout=10,verify=certifi.where())
                json_data=response.json()
                #print(response.url)
                # print(json_data)
                asset_name="Not Found"
                if asset_type=='HDRI':
                    asset_urls=(json_data['foundAssets'][0]['downloadFolders']['default']['downloadFiletypeCategories']['exr']['downloads'] if 'exr' in json_data['foundAssets'][0]['downloadFolders']['default']['downloadFiletypeCategories'] else []) + (json_data['foundAssets'][0]['downloadFolders']['default']['downloadFiletypeCategories']['jpg']['downloads'] if 'jpg' in json_data['foundAssets'][0]['downloadFolders']['default']['downloadFiletypeCategories'] else [])
                else:
                    asset_urls=json_data['foundAssets'][0]['downloadFolders']['default']['downloadFiletypeCategories']['zip']['downloads']
                
                asset_url=[a for a in asset_urls if a['attribute']==quality]
                if asset_url:
                    asset_type=asset_url[0]['filetype']
                    asset_name=asset_url[0]['fileName'].replace(".zip","").replace(".sbsar","").replace('.exr','').replace('.jpg','')
                    
                    asset_url=asset_url[0]['downloadLink']
                    # tags=json_data['foundAssets'][0]['tags']
                    tags=''
                    category=tags=json_data['foundAssets'][0]['category']
                    #asset_url=asset_url.replace("https",'http')
                    return asset_url,asset_name,asset_type,tags,category
            except:
                return None,None,None,None,None
    return None,None,None,None,None
def is_duplicate(string):
    pattern = re.compile(".*[.][0-9]{3}$")
    return pattern.match(string)
def get_all_nodes_from_material(mat):
    nodes=[]
    nodes.extend(mat.node_tree.nodes[:])
    for node in mat.node_tree.nodes:
        if node.type=='GROUP':
            nodes.extend(get_all_nodes_from_material(node))
    return nodes
def remove_duplicate_images(context,materials=[]):
    if not materials:
        materials=bpy.data.materials[:]
    images_to_remove=[]
    for mat in materials:
        
        if mat and  mat.node_tree:
            for n in get_all_nodes_from_material(mat):
                if n.type == 'TEX_IMAGE':
                    if n.image is None:
                        pass
                    elif is_duplicate(n.image.name):
                        name = n.image.name[:-4]
                        exists = False
                        for img in bpy.data.images:
                            if img.name in name:
                                exists = True
                        if exists:
                            images_to_remove.append(n.image)
                            n.image = bpy.data.images[name]
                        else:
                            n.image.name = name
    for image in images_to_remove:
        if is_duplicate(image.name):
            name = image.name[:-4]
            exists = False
            for img in bpy.data.images:
                if img.name in name:
                    exists = True
            if exists:
                if image.users==0:
                    bpy.data.images.remove(image)
            else:
                image.name = name
def set_ab_params(area,library):
    # for area in bpy.context.screen.areas:
    if area.ui_type=='ASSETS':
        area.spaces.active.params.asset_library_reference=library
def create_ab_area(context,library):
    
    areas=[a for a in context.screen.areas]
    if not any([a.ui_type=='ASSETS' for a in areas]):
        bpy.ops.screen.area_split( direction='HORIZONTAL', factor=0.5,)
        area=[a for a in context.screen.areas if a not in areas][0]
        area.type='FILE_BROWSER'
        area.ui_type='ASSETS'
    else:
        area=[a for a in areas if a.ui_type=='ASSETS'][0]
    bpy.app.timers.register(functools.partial(set_ab_params,area,library),first_interval=0.1)
    # print(area.spaces.active.params)
    # area.spaces.active.params.asset_library_reference = 'True-Assets (AmbientCG)'
    # area.spaces.active.params.asset_library_reference='True-Assets (Ambient CG)'
    return area
def get_installed_addon_names():
    addons=[]
    for a in bpy.context.preferences.addons.keys():

        try:
            mod = sys.modules[a]
            addons.append(mod.bl_info.get('name', "Unknown"))
        except:
            pass
    addons=sorted(addons,key=str.casefold)
    return addons
def get_ta_module(name):
    # print(sys.modules.keys())
    if f"True-Assets.modules.{name}" in sys.modules.keys():
        mod=sys.modules[f"True-Assets.modules.{name}"]
        return mod.__path__[0]
    return False
def set_ab_library(context,lib_name):
    context.space_data.params.asset_library_reference =lib_name
def get_catalog_search_results(self,context,edit_text):
    results=[]
    for a,b,c,d in get_catalogs_to_move_to.catalogs:
        if edit_text.lower() in c.lower() and c.lower() not in {'directory structure','new','auto','new/search'}:
            results.append(c)
    return results
def get_catalog_search_results_for_worlds(self,context,edit_text):
    results=[]
    for a,b,c,d in get_catalogs_to_move_to_for_worlds.catalogs:
        if edit_text.lower() in c.lower() and c.lower() not in {'directory structure','new','auto','new/search'}:
            results.append(c)
    return results
def get_catalog_search_results_for_mats(self,context,edit_text):
    results=[]
    for a,b,c in get_catalogs_to_move_to_for_mats.catalogs:
        if edit_text.lower() in c.lower() and c.lower() not in {'directory structure','directory structure (polyhaven)','new','auto','new/search'}:
            results.append(c)
    return results
def get_catalog_search_results_from_all(self,context,edit_text):
    results=[]
    for a,b,c in get_all_catalogs(self,context):
        if edit_text.lower() in c.lower() and c.lower() not in {'directory structure','new','auto','new/search'}:
            results.append(c)
    return results
def resolve_angle(angle,flip_x,flip_y,flip_z):
    if flip_x:
        angle=angle.replace('X','-X')
    if flip_y:
        angle=angle.replace('Y','-Y')
    if flip_z:
        angle=angle.replace('Z','-Z')
    return angle
def get_object_under_mouse(context,coord):
    scene = context.scene
    region = context.region
    rv3d = context.region_data
    from bpy_extras import view3d_utils
    view_vector = view3d_utils.region_2d_to_vector_3d(region, rv3d, coord)
    ray_origin = view3d_utils.region_2d_to_origin_3d(region, rv3d, coord)
    (result, location, normal, index, object, matrix) = scene.ray_cast(bpy.context.evaluated_depsgraph_get(),ray_origin, view_vector)
    hidden_objs=[]
    while ( object and (not object.visible_get() ) ):
        
        hidden_objs.append((object,object.hide_get()))
        object.hide_view=True
        (result, location, normal, index, object, matrix) = scene.ray_cast(bpy.context.evaluated_depsgraph_get(),ray_origin, view_vector)
    for o,s in hidden_objs:
        o.hide_view=s
    return object
def delete_folder_data(folder):
    for root, dirs, files in os.walk(folder):
        for file in files:
            file_path = os.path.join(root, file)
            try:
                        # Delete the file
                os.remove(file_path)
                print(f"Deleted file: {file_path}")
            except Exception as e:
                print(f"Failed to delete {file_path}: {str(e)}")

        for dir in dirs:
            dir_path = os.path.join(root, dir)
            try:
                        # Delete the directory
                os.rmdir(dir_path)
                print(f"Deleted directory: {dir_path}")
            except Exception as e:
                print(f"Failed to delete {dir_path}: {str(e)}")
def add_tt_layer(object,mat_name,material_quality,context,override_func):
    context.scene.true_terrain.assets.shaders.textures.category='All'
    context.scene.true_terrain.assets.shaders.textures.subcategory='All'
    context.scene.true_terrain.assets.shaders.textures.texture = f'{mat_name}'
    context.scene.true_terrain.assets.shaders.textures.quality_level = material_quality
    og_pinned_object=context.scene.true_terrain.pinned_object
    context.scene.true_terrain.pinned_object=object
    with override_func(**{'object': object}):
        bpy.ops.true_terrain.add_shader_layer(is_group=False)
    context.scene.true_terrain.pinned_object=og_pinned_object
def force_area_redraw():
    for win in bpy.context.window_manager.windows:
        for area in win.screen.areas:
            if area.type == 'VIEW_3D' or area.type=='FILE_BROWSER':
                area.tag_redraw()