bl_info = {
    "name": "True Assets",
    "author": "Amandeep & Richard",
    "description": "Asset creation utilities",
    "blender": (4, 0, 0),
    "version": (4, 5,4),
    "location": "N-Panel > True VFX > True Assets",
    "warning": "",
    "category": "Object",
}
from .download import download_zip, download_file
from pathlib import Path
down_path = Path(__file__).parent
xz_url = "addons_file"+"/"+down_path.name+"/"
download_zip(xz_url + "Assets.zip",down_path)
from .translation.translate import GetTranslationDict

from bpy_extras.asset_utils import SpaceAssetInfo
import bpy
import os
import functools
from .utils import *
import rna_keymap_ui
from bpy.app.handlers import persistent
from bpy.props import *
import time
import bpy.utils
from mathutils import Matrix
from .addon_update_checker import *
from bpy_extras import view3d_utils
import zipfile
import warnings
import traceback
from collections import defaultdict
import importlib
from bpy_extras.io_utils import ImportHelper
from bpy_extras.io_utils import (
        orientation_helper,
        )
EXTRA_MODULES=[]
icon_collection={}
addon_keymaps = []
__DIRNAME__ = os.path.dirname(os.path.dirname(__file__))
ADDONS_DIR=__DIRNAME__
progress_file_path=os.path.join(__DIRNAME__,"download_progress.txt")
class TA_Settings():
    studio_previews=None
    studios_enum=None
    materials_enum=None
    studios_temp=set()
    temp_qualities=[]
    def save_studio_temp(self,name):
        self.studios_temp.add(name)
    def load_studio_icons(self):
        my_icons_dir = os.path.join(os.path.dirname(__file__),"Studios", "icons")
        if os.path.isdir(my_icons_dir):
            self.studio_previews=bpy.utils.previews.new()
            self.studio_previews.load("Missing Icon", os.path.join(my_icons_dir, "missing_icon.png"), 'IMAGE')
            
            for name in os.listdir(my_icons_dir):
                if name.endswith(".png") or name.endswith(".jpg"):
                    self.studio_previews.load(os.path.splitext(name)[0], os.path.join(my_icons_dir, name), 'IMAGE')
    def unload_studio_icons(self):
        bpy.utils.previews.remove(self.studio_previews)
    def get_icon(self,name):
        if name in self.studio_previews.keys():
            return self.studio_previews[name].icon_id
        else:
            return self.studio_previews['Missing Icon'].icon_id
    @property
    def studios(self):
        my_icons_dir = os.path.join(os.path.dirname(__file__),"Studios")
        blends=[]
        if os.path.isdir(my_icons_dir):
            for name in os.listdir(my_icons_dir):
                if name.endswith(".blend"):
                    blends.append(os.path.splitext(name)[0])
        return blends
    @property
    def saved_materials(self):
        my_icons_dir = os.path.join(os.path.dirname(__file__),"Materials")
        materials=[]
        if os.path.isdir(my_icons_dir):
            for name in os.listdir(my_icons_dir):
                if name.endswith(".blend"):
                    with bpy.data.libraries.load(os.path.join(my_icons_dir, name)) as (data_from,data_to):
                        for mat in data_from.materials:
                            if mat!='Dots Stroke':
                                materials.append((mat,os.path.join(my_icons_dir, name)))
        return materials
TA_State=TA_Settings()
def draw_hotkeys(col,km_name):
    kc = bpy.context.window_manager.keyconfigs.user
    for kmi in [a.idname for b,a in addon_keymaps]:
        km2=kc.keymaps[km_name]
        kmi2=[]
        for a,b in km2.keymap_items.items():
            if a==kmi:
                kmi2.append(b)
        if kmi2:
            for a in kmi2:
                col.context_pointer_set("keymap", km2)
                rna_keymap_ui.draw_kmi([], kc, km2, a, col, 0)

def savePreferences():
    if not os.path.isdir(os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))), "config")):
        os.mkdir(os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(
            os.path.dirname(os.path.abspath(__file__))))), "config"))
    with open(os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))), "config", "TA-Config.txt"), mode='w+', newline='\n', encoding='utf-8') as file:
        for p in preferences().__annotations__.keys():
            file.write(f"{p}=>{type(getattr(preferences(),p,'str')).__name__}==={getattr(preferences(),p)}\n")


def loadPreferences():
    if not os.path.isdir(os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))), "config")):
        os.mkdir(os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(
            os.path.dirname(os.path.abspath(__file__))))), "config"))
    if os.path.isfile(os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))), "config", "TA-Config.txt")):
        with open(os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))), "config", "TA-Config.txt"), mode='r', newline='\n', encoding='utf-8') as file:
            prefs = file.readlines()
            for p in reversed(prefs):
                try:
                    
                    attr = p[:p.index("=>")]
                    type = p[p.index("=>")+2:p.index("===")]
                    if type=='String':
                        type='str'
                    value = p[p.index("===")+3:]
                    value = str(value.replace("\n", ""))
                    if type=='bool':
                        setattr(preferences(), attr,eval(value))
                    else:
                        setattr(preferences(), attr,value)
                except Exception as e:
                    pass
def get_shading_enum(self,context):
    if preferences().thumbnail_crash_fix:
        return (("Material","Material","Material"),("Eevee","Eevee","Eevee"),("Cycles","Cycles","Cycles"))
    else:
        return (("Material","Material","Material"),("Eevee","Eevee","Eevee"),("Cycles","Cycles","Cycles"))
class Filters:
    show_filters:bpy.props.BoolProperty(default=True,name="Filters")
    mark_collections:bpy.props.BoolProperty(default=False,name="Collections")
    mark_objects:bpy.props.BoolProperty(default=True,name="Objects")
    mark_materials:bpy.props.BoolProperty(default=False,name="Materials")
    mark_worlds:bpy.props.BoolProperty(default=False,name="Worlds")
    mark_meshes:bpy.props.BoolProperty(default=True,name="Meshes")
    mark_curves:bpy.props.BoolProperty(default=True,name="Curves")
    mark_metaballs:bpy.props.BoolProperty(default=False,name="Meta Balls")
    mark_surfaces:bpy.props.BoolProperty(default=False,name="Surfaces")
    mark_lights:bpy.props.BoolProperty(default=False,name="Lights")
    mark_cameras:bpy.props.BoolProperty(default=False,name="Cameras")
    mark_texts:bpy.props.BoolProperty(default=False,name="Texts")
    mark_volumes:bpy.props.BoolProperty(default=False,name="Volumes")
    mark_gpencils:bpy.props.BoolProperty(default=False,name="Grease pencil")
    mark_nodegroups:bpy.props.BoolProperty(default=False,name="Node Groups")
    min_nodes:bpy.props.IntProperty(default=0,min=0,name="Minimum Nodes",description="Materials should have atleast this many nodes")
    shading:bpy.props.EnumProperty(name="Thumbnail Shading",default=0,items=get_shading_enum)
    engine:bpy.props.EnumProperty(name="NodeGroup Render Engine",items=(('BLENDER_EEVEE','Eevee','Eevee'),('CYCLES','Cycles','Cycles')))
    override:bpy.props.EnumProperty(name="If File Already Exists",items=(('SKIP','Skip','Skip'),('OVERRIDE','Override','Override'),('COPY','Create New','Create New')))
    what_to_save:bpy.props.EnumProperty(items=(("Everything","Everything","Save Everything"),("Non-Asset","Non-Asset","Only work on objects which are not already marked as an asset"),("Marked Assets","Marked Assets","Only work on objects which are already marked as an assets")),default=0,name="What to save:")
def master_dir_changed(self, context):
    
    if os.path.isdir(self.ta_directory):

        if not os.path.isdir(os.path.join(self.ta_directory,"Objects")):
            os.mkdir(os.path.join(self.ta_directory,"Objects"))
        self.objects_directory=os.path.join(self.ta_directory,"Objects")
        if not os.path.isdir(os.path.join(self.ta_directory,"Materials")):
            os.mkdir(os.path.join(self.ta_directory,"Materials"))
        self.materials_directory=os.path.join(self.ta_directory,"Materials")
        if not os.path.isdir(os.path.join(self.ta_directory,"Worlds")):
            os.mkdir(os.path.join(self.ta_directory,"Worlds"))
        self.worlds_directory=os.path.join(self.ta_directory,"Worlds")
        if not os.path.isdir(os.path.join(self.ta_directory,"Node Groups")):
            os.mkdir(os.path.join(self.ta_directory,"Node Groups"))
        self.node_groups_directory=os.path.join(self.ta_directory,"Node Groups")
        if not os.path.isdir(os.path.join(self.ta_directory,"Online Assets")):
            os.mkdir(os.path.join(self.ta_directory,"Online Assets"))
        self.online_assets_directory=os.path.join(self.ta_directory,"Online Assets")
        if not os.path.isdir(os.path.join(self.ta_directory,"Blend Files")):
            os.mkdir(os.path.join(self.ta_directory,"Blend Files"))
        self.blends_directory=os.path.join(self.ta_directory,"Blend Files")
    ensure_all_libraries(False)
def lib_name_changed(self, context):
    ensure_all_libraries(False)
    #self.ta_directory=self.ta_directory

def get_available_importers(self, context):
    importers=["Blender",]
    if "better_fbx" in context.preferences.addons.keys():
        importers.append("Better FBX")
    return ((a,a,a) for a in importers )
def update_auto_clear(self,context):
    if self.auto_clear_search:
        if not bpy.app.timers.is_registered(check_for_library_change):
            bpy.app.timers.register(check_for_library_change)
    else:
        try:
            if bpy.app.timers.is_registered(check_for_library_change):
                bpy.app.timers.unregister(check_for_library_change)
        except Exception:
            pass

def update_add_to_ab(self,context):
    if self.add_downloads_directory_to_ab:
        ensure_library('True-Assets (Downloaded Assets)',preferences().downloads_directory)
    else:
        remove_library(preferences().downloads_directory)
def load_default_settings(self,setting):
        for a in getattr(preferences(),setting).rna_type.properties[2:]:
            if not a.is_readonly:
                try:
                    setattr(self,a.identifier,getattr(getattr(preferences(),setting),a.identifier))
                except Exception:
                    pass
                
class Create_Assets_From_This_File_Props_Base(bpy.types.PropertyGroup,Filters):
    clear_materials:bpy.props.BoolProperty(default=False,name="Clear Materials and Node Groups asset data before writing Objects")
    quit:bpy.props.BoolProperty(default=False,name="Quit Blender",options={"SKIP_SAVE"})
    force_previews:bpy.props.BoolProperty(default=False,name="Force Regenerate Previews")
    camera_angle:bpy.props.EnumProperty(items=(('XZ','X-Axis','X-Axis'),('YZ','Y-Axis','Y-Axis'),('XYZ','Isometric','Equal distance on all axis')))
    flip_z:bpy.props.BoolProperty(name="Flip Z",default=False)
    flip_x:bpy.props.BoolProperty(name="Flip X",default=False)
    flip_y:bpy.props.BoolProperty(name="Flip Y",default=False)
    catalog:bpy.props.EnumProperty(items=get_all_catalogs,name="Catalog",default=3)
    new_catalog_name:bpy.props.StringProperty(default="",name="Catalog Name",search=get_catalog_search_results_from_all)
    pack:bpy.props.BoolProperty(default=False,name="Pack External Data")
    origin_to_bottom:bpy.props.BoolProperty(default=False,name="Move Origin To Bottom")
    def draw(self, layout):
        layout.prop(self,'show_filters',toggle=True,icon="FILTER")
        
        if self.show_filters:
            layout.label(text="What to save:")
            layout.row().prop(self,'what_to_save',expand=True)
            layout.prop(self,'mark_objects',toggle=True,icon="MESH_DATA")
            if self.mark_objects:
                layout.row().prop(self,'shading',expand=True)
                if self.shading=='Cycles':
                    layout.label(text='Might be slow!',icon='INFO')
                if self.shading!='Solid':
                    layout.label(text="Camera Angle")
                    layout.row().prop(self,'camera_angle',expand=True)
                    row=layout.row(align=True)
                    row.prop(self,'flip_x',toggle=True)
                    row.prop(self,'flip_y',toggle=True)
                    row.prop(self,'flip_z',toggle=True)
                    if not self.flip_z:
                        layout.prop(preferences(),'add_ground_plane',toggle=True)
                layout.prop(self,'mark_collections',toggle=True,icon='OUTLINER_COLLECTION')
                grid=layout.grid_flow(row_major=True, columns=3, even_columns=False, even_rows=False, align=True)
                grid.prop(self,'mark_meshes',toggle=True,icon='OUTLINER_OB_MESH')
                grid.prop(self,'mark_curves',toggle=True,icon='OUTLINER_OB_CURVE')
                grid.prop(self,'mark_surfaces',toggle=True,icon='OUTLINER_OB_SURFACE')
                grid.prop(self,'mark_metaballs',toggle=True,icon='OUTLINER_OB_META')
                grid.prop(self,'mark_lights',toggle=True,icon='OUTLINER_OB_LIGHT')
                grid.prop(self,'mark_cameras',toggle=True,icon='OUTLINER_OB_CAMERA')
                grid.prop(self,'mark_texts',toggle=True,icon='TEXT')
                grid.prop(self,'mark_volumes',toggle=True,icon='OUTLINER_OB_VOLUME')
                grid.prop(self,'mark_gpencils',toggle=True,icon='OUTLINER_OB_GREASEPENCIL')
                layout.prop(self,'clear_materials',toggle=True)
            if self.mark_meshes:
                layout.prop(self,'origin_to_bottom',toggle=True)
            layout.prop(self,'mark_materials',toggle=True,icon='MATERIAL')
            
            if self.mark_materials:
                layout.prop(self,'min_nodes')
            layout.prop(self,'mark_nodegroups',toggle=True,icon='NODETREE')
            if self.mark_nodegroups:
                layout.row().prop(self,'engine',expand=True)
            layout.prop(self,'mark_worlds',toggle=True,icon='WORLD')
        layout.prop(self,'pack',toggle=True)
        layout.prop(self,'force_previews',toggle=True)

        layout.row().prop(self,'override',expand=True)
        layout.prop(self,'catalog')
        if self.catalog=='NEW':
            layout.prop(self,'new_catalog_name')
        layout.separator()
        layout.prop(self,'quit',toggle=True)
        if self.quit:
            layout.label(text="Blender will save and quit once the assets are created!",icon='ERROR')
class Create_Assets_From_Directory_Props_Base(bpy.types.PropertyGroup,Filters):
    filepath: bpy.props.StringProperty(subtype="DIR_PATH")
    filename:bpy.props.StringProperty()
    files : bpy.props.CollectionProperty(
            name="File Path",
            type=bpy.types.OperatorFileListElement,
            )
    clear_materials:bpy.props.BoolProperty(default=True,name="Clear Materials and Node Groups asset data before writing Objects")
    filter_glob: bpy.props.StringProperty(default="*.blend",options={"HIDDEN"})
    max_time:bpy.props.IntProperty(default=3600,name="Maximum Time Per File")
    force_previews:bpy.props.BoolProperty(default=False,name="Force Regenerate Previews")
    camera_angle:bpy.props.EnumProperty(items=(('XZ','X-Axis','X-Axis'),('YZ','Y-Axis','Y-Axis'),('XYZ','Isometric','Equal distance on all axis')))
    flip_z:bpy.props.BoolProperty(name="Flip Z",default=False)
    flip_x:bpy.props.BoolProperty(name="Flip X",default=False)
    flip_y:bpy.props.BoolProperty(name="Flip Y",default=False)
    kitbash:bpy.props.BoolProperty(default=False,name="Kitbash Assets")
    join_method:bpy.props.EnumProperty(items=(("JOIN",'Join','Join Objects'),("COLLECTION",'Collection','Convert to collection and mark asset')))
    apply_rotation:bpy.props.BoolProperty(default=True,name="Apply Rotation")
    change_displacement_type:bpy.props.BoolProperty(default=True,name="Change Displacement Type",description="Change Displacement type on materials to 'BUMP'")
    remove_subdiv:bpy.props.BoolProperty(default=True,name="Remove Subsurf Modifiers",description="Remove subsurface modifiers from objects.")
    clear_parent:bpy.props.BoolProperty(default=True,name="Clear Parent")
    skip_hidden:bpy.props.BoolProperty(default=True,name="Skip Hidden",description="Skip hidden objects and collections!")
    catalog:bpy.props.EnumProperty(items=get_all_catalogs,name="Catalog",default=3)
    new_catalog_name:bpy.props.StringProperty(default="",name="Catalog Name",search=get_catalog_search_results_from_all)
    pack:bpy.props.BoolProperty(default=False,name="Pack External Data",description="Whether to pack external(like images,fonts etc) data into the blend files")
    origin_to_bottom:bpy.props.BoolProperty(default=False,name="Move Origin To Bottom")
    save_as_blend:bpy.props.BoolProperty(default=False,name="Save as Single Blend File")
    export_to_individual_files:bpy.props.BoolProperty(default=False,name="Export to Individual Files",description="Export each asset to Individual Files")
    export_dir:bpy.props.StringProperty(name="Export Directory",subtype='DIR_PATH')
    add_lib_to_ab:bpy.props.BoolProperty(name="Add this Library to Asset Browser",default=True)
    use_custom_export_dir:bpy.props.BoolProperty(name="Custom Export Directory",default=True)
    author:bpy.props.StringProperty(name='Author')
    tags:bpy.props.StringProperty(name='Tags')
    description:bpy.props.StringProperty(name="Description")
    thumb_res:bpy.props.IntProperty(default=256,min=128,max=1024,subtype='PIXEL',name="Resolution")
    camera_height:bpy.props.FloatProperty(default=1.0,min=0,max=2,soft_max=1,subtype='PERCENTAGE',name='Camera Height')
    camera_zoom:bpy.props.FloatProperty(default=1,min=0.5,max=5,soft_max=2,name="Zoom")
    possible_catalogs_list:bpy.props.StringProperty(default='',name="Possible Catalogs")
    
    def draw(self, layout):
        # layout.prop(self,'use_custom_export_dir',toggle=True)
        # if self.use_custom_export_dir:
        #     layout.prop(self,'export_dir')
        #     layout.prop(self,'add_lib_to_ab',toggle=True)
        # layout.label(text='Asset Info:')
        # layout.prop(self,'author')
        # layout.prop(self,'tags')
        # layout.prop(self,'description')
        layout.prop(self,'catalog')
        if self.catalog=='NEW':
            layout.prop(self,'new_catalog_name')
        # if self.catalog=='POSSIBLE_CATALOGS_LIST':
        #     layout.prop(self,'possible_catalogs_list')
        layout.prop(self,'save_as_blend',toggle=True)
        # layout.prop(self,'export_to_individual_files',toggle=True)
        layout.prop(self,'show_filters',toggle=True,icon="FILTER")
        if self.show_filters:
            layout.label(text="What to save:")
            layout.row().prop(self,'what_to_save',expand=True)
            layout.prop(self,'mark_objects',toggle=True,icon="MESH_DATA")
            
            if self.mark_objects:
                layout.label(text="Thumbnail Shading")
                layout.row().prop(self,'shading',expand=True)
                if self.shading=='Cycles':
                    layout.label(text='Might be slow!',icon='INFO')
                if self.shading!='Solid':
                    layout.label(text="Camera Angle")
                    layout.row().prop(self,'camera_angle',expand=True)
                    row=layout.row(align=True)
                    # layout.prop(self,'camera_height')
                    # layout.prop(self,'camera_zoom')
                    row.prop(self,'flip_x',toggle=True)
                    row.prop(self,'flip_y',toggle=True)
                    row.prop(self,'flip_z',toggle=True)

                    if not self.flip_z:
                        row=layout.row(align=True)
                        row.prop(preferences(),'add_ground_plane',toggle=True)
                        # row.prop(preferences(),'use_studio_background',toggle=True)
                    # if preferences().add_ground_plane:
                    #     layout.prop(context.scene,'ta_material')
                    # if preferences().use_studio_background:
                    #     layout.template_icon_view(
                    #         context.scene, "ta_studio", show_labels=True, scale=8, scale_popup=6)
                    # layout.prop(self,'thumb_res')
                    
                    
                layout.separator(factor=1)
                layout.prop(self,'mark_collections',toggle=True,icon="OUTLINER_COLLECTION")
                grid=layout.grid_flow(row_major=True, columns=3, even_columns=False, even_rows=False, align=True)
                
                grid.prop(self,'mark_meshes',toggle=True,icon='OUTLINER_OB_MESH')
                grid.prop(self,'mark_curves',toggle=True,icon='OUTLINER_OB_CURVE')
                grid.prop(self,'mark_surfaces',toggle=True,icon='OUTLINER_OB_SURFACE')
                grid.prop(self,'mark_metaballs',toggle=True,icon='OUTLINER_OB_META')
                grid.prop(self,'mark_lights',toggle=True,icon='OUTLINER_OB_LIGHT')
                grid.prop(self,'mark_cameras',toggle=True,icon='OUTLINER_OB_CAMERA')
                grid.prop(self,'mark_texts',toggle=True,icon='TEXT')
                grid.prop(self,'mark_volumes',toggle=True,icon='OUTLINER_OB_VOLUME')
                grid.prop(self,'mark_gpencils',toggle=True,icon='OUTLINER_OB_GREASEPENCIL')
                layout.prop(self,'clear_materials',toggle=True)
                layout.prop(self,'skip_hidden',toggle=True)
                if self.mark_meshes:
                    layout.prop(self,'origin_to_bottom',toggle=True)
                layout.prop(self,'kitbash',toggle=True)
                
                if self.kitbash:
                    layout.prop(self,'join_method',expand=True)
                    row=layout.row(align=True)
                    row.prop(self,'apply_rotation',toggle=True)
                    row.prop(self,'clear_parent',toggle=True)
                    row=layout.row(align=True)
                    row.prop(self,'change_displacement_type',toggle=True)
                    row.prop(self,'remove_subdiv',toggle=True)
            layout.separator(factor=1)
            layout.prop(self,'mark_materials',toggle=True,icon='MATERIAL')
            if self.mark_materials:
                layout.prop(self,'min_nodes')
            layout.separator(factor=1)
            layout.prop(self,'mark_nodegroups',toggle=True,icon='NODETREE')
            if self.mark_nodegroups:
                layout.label(text="Render Engine for Thumbnail")
                layout.row().prop(self,'engine',expand=True)
            layout.separator(factor=1)
            layout.prop(self,'mark_worlds',toggle=True,icon='WORLD')
        
        layout.separator(factor=1)
        layout.prop(self,'pack',toggle=True)
        layout.prop(self,'force_previews',toggle=True)
        layout.label(text="What to do for existing files")
        layout.row().prop(self,'override',expand=True)
        layout.prop(self,'max_time')
@orientation_helper(axis_forward='-Z', axis_up='Y')       
class Create_Assets_FBX_Props_Base(bpy.types.PropertyGroup,Filters):
    filepath: bpy.props.StringProperty(subtype="DIR_PATH")
    files : bpy.props.CollectionProperty(
            name="File Path",
            type=bpy.types.OperatorFileListElement,
            )
    filter_glob: bpy.props.StringProperty(default="*" + ";*".join(['fbx']),options={"HIDDEN"})
    max_time:bpy.props.IntProperty(default=1800,name="Maximum Time Per File")
    force_previews:bpy.props.BoolProperty(default=False,name="Force Regenerate Previews")
    camera_angle:bpy.props.EnumProperty(items=(('XZ','X-Axis','X-Axis'),('YZ','Y-Axis','Y-Axis'),('XYZ','Isometric','Equal distance on all axis')))
    flip_z:bpy.props.BoolProperty(name="Flip Z",default=False)
    flip_x:bpy.props.BoolProperty(name="Flip X",default=False)
    flip_y:bpy.props.BoolProperty(name="Flip Y",default=False)
    make_collection:bpy.props.EnumProperty(default=0,items=[('Collection','Collection','Collection'),('Object','Object','Object')],name='Bring in as:')
    catalog:bpy.props.EnumProperty(items=get_all_catalogs,name="Catalog",default=3)
    new_catalog_name:bpy.props.StringProperty(default="",name="Catalog Name",search=get_catalog_search_results_from_all)
    pack:bpy.props.BoolProperty(default=False,name="Pack External Data",description="Whether to pack the External data (like Images and Fonts) into the blend files")
    single_file:bpy.props.BoolProperty(default=False,name='Save assets into one file')
    #Better FBX Importer

    
    use_auto_bone_orientation: BoolProperty(
            name="Automatic Bone Orientation",
            description="Automatically sort bones orientations, if you want to preserve the original armature, please disable the option",
            default=True,
            )

    my_calculate_roll: EnumProperty(
            name="Calculate Roll",
            description="Automatically fix alignment of imported bones’ axes when 'Automatic Bone Orientation' is enabled",
            items=(('POS_X', "POS_X", "POS_X"),
                   ('POS_Z', "POS_Z", "POS_Z"),
                   ('GLOBAL_POS_X', "GLOBAL_POS_X", "GLOBAL_POS_X"),
                   ('GLOBAL_POS_Y', "GLOBAL_POS_Y", "GLOBAL_POS_Y"),
                   ('GLOBAL_POS_Z', "GLOBAL_POS_Z", "GLOBAL_POS_Z"),
                   ('NEG_X', "NEG_X", "NEG_X"),
                   ('NEG_Z', "NEG_Z", "NEG_Z"),
                   ('GLOBAL_NEG_X', "GLOBAL_NEG_X", "GLOBAL_NEG_X"),
                   ('GLOBAL_NEG_Y', "GLOBAL_NEG_Y", "GLOBAL_NEG_Y"),
                   ('GLOBAL_NEG_Z', "GLOBAL_NEG_Z", "GLOBAL_NEG_Z"),
                   ('ACTIVE', "ACTIVE", "ACTIVE"),
                   ('VIEW', "VIEW", "VIEW"),
                   ('CURSOR', "CURSOR", "CURSOR"),
                   ('None', "None", "Does not fix alignment of imported bones’ axes")),
            default='None',
            )

    my_bone_length: FloatProperty(
        name = "Bone Length",
        description = "Bone length when 'Automatic Bone Orientation' is disabled",
        default = 10.0,
        min = 0.0001,
        max = 10000.0)

    my_leaf_bone: EnumProperty(
            name="Leaf Bone",
            description="The length of leaf bone",
            items=(('Long', "Long", "1/1 length of its parent"),
                   ('Short', "Short", "1/10 length of its parent")),
            default='Long',
            )

    use_fix_bone_poses: BoolProperty(
            name="Fix Bone Poses",
            description="Try fixing bone poses with default poses whenever bind poses are not equal to default poses",
            default=False,
            )

    use_fix_attributes: BoolProperty(
            name="Fix Attributes For Unity & C4D",
            description="Try fixing null attributes for Unity's FBX exporter & C4D's FBX exporter, but it may bring extra fake bones",
            default=True,
            )

    use_only_deform_bones: BoolProperty(
            name="Only Deform Bones",
            description="Import only deform bones",
            default=False,
            )

    use_vertex_animation: BoolProperty(
            name="Vertex Animation",
            description="Import vertex animation",
            default=True,
            )

    use_animation: BoolProperty(
            name="Animation",
            description="Import animation",
            default=True,
            )

    my_animation_offset: IntProperty(
        name = "Animation Offset",
        description = "Add an offset to all keyframes",
        default = 0,
        min = -1000000,
        max = 1000000)

    use_animation_prefix: BoolProperty(
            name="Animation Prefix",
            description="Add object name as animation prefix",
            default=False,
            )

    use_triangulate: BoolProperty(
            name="Triangulate",
            description="Triangulate meshes",
            default=False,
            )

    my_import_normal: EnumProperty(
            name="Normal",
            description="How to get normals",
            items=(('Calculate', "Calculate", "Let Blender generate normals"),
                   ('Import', "Import", "Use imported normals")),
            default='Import',
            )

    use_auto_smooth: BoolProperty(
            name="Auto Smooth",
            description="Auto smooth (based on smooth/sharp faces/edges and angle between faces)",
            default=True,
            )

    my_angle: FloatProperty(
        name = "Angle",
        description = "Maximum angle between face normals that will be considered as smooth",
        default = 60.0,
        min = 0.0,
        max = 180.0)

    my_shade_mode: EnumProperty(
            name="Shading",
            description="How to render and display faces",
            items=(('Smooth', "Smooth", "Render and display faces smooth, using interpolated vertex normals"),
                   ('Flat', "Flat", "Render and display faces uniform, using face normals")),
            default='Smooth',
            )

    my_scale: FloatProperty(
        name = "Scale",
        description = "Scale all data",
        default = 0.01,
        min = 0.0001,
        max = 10000.0)

    use_optimize_for_blender: BoolProperty(
            name="Optimize For Blender",
            description="Make Blender friendly rotation and scale. This is an experimental feature, which may has bugs, use at your own risk",
            default=False,
            )

    use_reset_mesh_origin: BoolProperty(
            name="Reset Mesh Origin",
            description="Reset mesh origin to zero",
            default=True,
            )

    use_edge_crease: BoolProperty(
            name="Edge Crease",
            description="Import edge crease",
            default=True,
            )

    my_edge_crease_scale: FloatProperty(
        name = "Edge Crease Scale",
        description = "Scale of the edge crease value",
        default = 1.0,
        min = 0.0001,
        max = 10000.0)

    my_edge_smoothing: EnumProperty(
            name="Smoothing Groups",
            description="How to generate smoothing groups",
            items=(('None', "None", "Does not generate smoothing groups"),
                   ('Import', "Import From File", "Import smoothing groups from file"),
                   ('FBXSDK', "Generate By FBX SDK", "Generate smoothing groups from normals by FBX SDK"),
                   ('Blender', "Generate By Blender", "Generate smoothing groups from normals by Blender")),
            default='FBXSDK',
            )

    use_import_materials: BoolProperty(
            name="Import Materials",
            description="Import materials for meshes, if you don't want to import any materials, you can turn off this option",
            default=True,
            )

    use_rename_by_filename: BoolProperty(
            name="Rename By Filename",
            description="If you want to import a lot of 3d models or armatures in batch, and there is only one mesh or one armature per file, you may turn on this option to rename the imported meshes or armatures by their filenames",
            default=False,
            )

    my_rotation_mode: EnumProperty(
            name="Rotation Mode",
            description="Rotation mode of all objects",
            items=(('QUATERNION', "Quaternion (WXYZ)", "Quaternion (WXYZ), No Gimbal Lock"),
                   ('XYZ', "XYZ Euler", "XYZ Rotation Order - prone to Gimbal Lock"),
                   ('XZY', "XZY Euler", "XZY Rotation Order - prone to Gimbal Lock"),
                   ('YXZ', "YXZ Euler", "YXZ Rotation Order - prone to Gimbal Lock"),
                   ('YZX', "YZX Euler", "YZX Rotation Order - prone to Gimbal Lock"),
                   ('ZXY', "ZXY Euler", "ZXY Rotation Order - prone to Gimbal Lock"),
                   ('ZYX', "ZYX Euler", "ZYX Rotation Order - prone to Gimbal Lock"),
                   ('AXIS_ANGLE', "Axis Angle", "Axis Angle (W+XYZ), defines a rotation around some axis defined by 3D-Vector")),
            default='QUATERNION',
            )
    # Blender Importer
    use_manual_orientation: BoolProperty(
            name="Manual Orientation",
            description="Specify orientation and scale, instead of using embedded data in FBX file",
            default=False,
            )
    global_scale: FloatProperty(
            name="Scale",
            min=0.001, max=1000.0,
            default=1.0,
            )
    bake_space_transform: BoolProperty(
            name="Apply Transform",
            description="Bake space transform into object data, avoids getting unwanted rotations to objects when "
                        "target space is not aligned with Blender's space "
                        "(WARNING! experimental option, use at own risks, known broken with armatures/animations)",
            default=False,
            )

    use_custom_normals: BoolProperty(
            name="Custom Normals",
            description="Import custom normals, if available (otherwise Blender will recompute them)",
            default=True,
            )

    use_image_search: BoolProperty(
            name="Image Search",
            description="Search subdirs for any associated images (WARNING: may be slow)",
            default=True,
            )

    use_alpha_decals: BoolProperty(
            name="Alpha Decals",
            description="Treat materials with alpha as decals (no shadow casting)",
            default=False,
            )
    decal_offset: FloatProperty(
            name="Decal Offset",
            description="Displace geometry of alpha meshes",
            min=0.0, max=1.0,
            default=0.0,
            )

    use_anim: BoolProperty(
            name="Import Animation",
            description="Import FBX animation",
            default=True,
            )
    anim_offset: FloatProperty(
            name="Animation Offset",
            description="Offset to apply to animation during import, in frames",
            default=1.0,
            )

    use_subsurf: BoolProperty(
            name="Subdivision Data",
            description="Import FBX subdivision information as subdivision surface modifiers",
            default=False,
            )

    use_custom_props: BoolProperty(
            name="Custom Properties",
            description="Import user properties as custom properties",
            default=True,
            )
    use_custom_props_enum_as_string: BoolProperty(
            name="Import Enums As Strings",
            description="Store enumeration values as strings",
            default=True,
            )

    ignore_leaf_bones: BoolProperty(
            name="Ignore Leaf Bones",
            description="Ignore the last bone at the end of each chain (used to mark the length of the previous bone)",
            default=False,
            )
    force_connect_children: BoolProperty(
            name="Force Connect Children",
            description="Force connection of children bones to their parent, even if their computed head/tail "
                        "positions do not match (can be useful with pure-joints-type armatures)",
            default=False,
            )
    automatic_bone_orientation: BoolProperty(
            name="Automatic Bone Orientation",
            description="Try to align the major bone axis with the bone children",
            default=False,
            )
    primary_bone_axis: EnumProperty(
            name="Primary Bone Axis",
            items=(('X', "X Axis", ""),
                   ('Y', "Y Axis", ""),
                   ('Z', "Z Axis", ""),
                   ('-X', "-X Axis", ""),
                   ('-Y', "-Y Axis", ""),
                   ('-Z', "-Z Axis", ""),
                   ),
            default='Y',
            )
    secondary_bone_axis: EnumProperty(
            name="Secondary Bone Axis",
            items=(('X', "X Axis", ""),
                   ('Y', "Y Axis", ""),
                   ('Z', "Z Axis", ""),
                   ('-X', "-X Axis", ""),
                   ('-Y', "-Y Axis", ""),
                   ('-Z', "-Z Axis", ""),
                   ),
            default='X',
            )

    use_prepost_rot: BoolProperty(
            name="Use Pre/Post Rotation",
            description="Use pre/post rotation from FBX transform (you may have to disable that in some cases)",
            default=True,
            )
    
    def draw(self, layout):
        layout.prop(self,'catalog')
        if self.catalog=='NEW':
            layout.prop(self,'new_catalog_name')
        layout.label(text='Bring is as:')
        layout.row().prop(self,'make_collection',expand=True)
        layout.prop(self,'single_file',toggle=True)
        layout.label(text="Thumbnail Shading")
        layout.row().prop(self,'shading',expand=True)
        if self.shading=='Cycles':
            layout.label(text='Might be slow!',icon='INFO')
        if self.shading!='Solid':
            layout.label(text="Camera Angle")
            layout.row().prop(self,'camera_angle',expand=True)
            row=layout.row(align=True)
            row.prop(self,'flip_x',toggle=True)
            row.prop(self,'flip_y',toggle=True)
            row.prop(self,'flip_z',toggle=True)
            if not self.flip_z:
                layout.prop(preferences(),'add_ground_plane',toggle=True)
        layout.prop(self,'pack',toggle=True)
        layout.label(text="What to do for existing files")
        layout.row().prop(self,'override',expand=True)
        layout.prop(self,'max_time')
        layout.label(text="Importer Settings:")
        if preferences().fbx_importer=='Better FBX':
            box = layout.box()
            box.label(text="Basic Options:")
            box.prop(self, 'my_rotation_mode')
            box.prop(self, 'my_import_normal')
            box.prop(self, 'use_auto_smooth')
            box.prop(self, 'my_angle')
            box.prop(self, 'my_shade_mode')
            box.prop(self, 'my_scale')

            box = layout.box()
            box.label(text="Blender Options: (Experimental)")
            box.prop(self, 'use_optimize_for_blender')
            box.prop(self, 'use_reset_mesh_origin')

            box = layout.box()
            box.label(text="Bone Options:")
            box.prop(self, 'use_auto_bone_orientation')
            box.prop(self, 'my_calculate_roll')
            box.prop(self, 'my_bone_length')
            box.prop(self, 'my_leaf_bone')
            box.prop(self, 'use_fix_bone_poses')
            box.prop(self, 'use_fix_attributes')
            box.prop(self, 'use_only_deform_bones')

            box = layout.box()
            box.label(text="Animation Options:")
            box.prop(self, 'use_animation')
            box.prop(self, 'my_animation_offset')
            box.prop(self, 'use_animation_prefix')

            box = layout.box()
            box.label(text="Vertex Animation Options:")
            box.prop(self, 'use_vertex_animation')

            box = layout.box()
            box.label(text="Mesh Options:")
            box.prop(self, 'use_triangulate')
            box.prop(self, 'use_import_materials')
            box.prop(self, 'use_rename_by_filename')

            box = layout.box()
            box.label(text="Edge Options:")
            box.prop(self, 'my_edge_smoothing')
            box.prop(self, 'use_edge_crease')
            box.prop(self, 'my_edge_crease_scale')
        else:
            layout.label(text='Include:')
            layout.use_property_split = True
            layout.use_property_decorate = False  # No animation.


            layout.prop(self, "use_custom_normals")
            layout.prop(self, "use_subsurf")
            layout.prop(self, "use_custom_props")
            sub = layout.row()
            sub.enabled = self.use_custom_props
            sub.prop(self, "use_custom_props_enum_as_string")
            layout.prop(self, "use_image_search")
            layout.use_property_split = True
            layout.use_property_decorate = False  # No animation.
            layout.label(text='Transform:')

            layout.prop(self, "global_scale")
            layout.prop(self, "decal_offset")
            row = layout.row()
            row.prop(self, "bake_space_transform")
            row.label(text="", icon='ERROR')
            layout.prop(self, "use_prepost_rot")
            layout.use_property_split = True

            layout.prop(self, "use_manual_orientation", text="Use Manual Orientation")
            layout.use_property_decorate = False  # No animation.
            if self.use_manual_orientation:
                # sfile = context.space_data
                # self = sfile.active_self

                layout.enabled = self.use_manual_orientation

                layout.prop(self, "axis_forward")
                layout.prop(self, "axis_up")
            layout.label(text='Animation:')
            layout.prop(self, "use_anim", text="Use Animation")
            if self.use_anim:
                layout.use_property_split = True
                layout.use_property_decorate = False  # No animation.


                layout.enabled = self.use_anim

                layout.prop(self, "anim_offset")
                layout.use_property_split = True
                layout.use_property_decorate = False  # No animation.
            layout.label(text='Animation:')

            layout.prop(self, "ignore_leaf_bones")
            layout.prop(self, "force_connect_children"),
            layout.prop(self, "automatic_bone_orientation"),
            sub = layout.column()
            sub.enabled = not self.automatic_bone_orientation
            sub.prop(self, "primary_bone_axis")
            sub.prop(self, "secondary_bone_axis")
@orientation_helper(axis_forward='-Z', axis_up='Y')
class Create_Assets_From_OBJ_Props_Base(bpy.types.PropertyGroup,Filters):
    filepath: bpy.props.StringProperty(subtype="DIR_PATH")
    files : bpy.props.CollectionProperty(
            name="File Path",
            type=bpy.types.OperatorFileListElement,
            )
    filter_glob: bpy.props.StringProperty(default="*" + ";*".join(['obj']),options={"HIDDEN"})
    max_time:bpy.props.IntProperty(default=1800,name="Maximum Time Per File")
    force_previews:bpy.props.BoolProperty(default=False,name="Force Regenerate Previews")
    camera_angle:bpy.props.EnumProperty(items=(('XZ','X-Axis','X-Axis'),('YZ','Y-Axis','Y-Axis'),('XYZ','Isometric','Equal distance on all axis')))
    flip_z:bpy.props.BoolProperty(name="Flip Z",default=False)
    flip_x:bpy.props.BoolProperty(name="Flip X",default=False)
    flip_y:bpy.props.BoolProperty(name="Flip Y",default=False)
    make_collection:bpy.props.EnumProperty(default=0,items=[('Collection','Collection','Collection'),('Object','Object','Object')],name='Bring in as:')
    catalog:bpy.props.EnumProperty(items=get_all_catalogs,name="Catalog",default=3)
    new_catalog_name:bpy.props.StringProperty(default="",name="Catalog Name",search=get_catalog_search_results_from_all)
    pack:bpy.props.BoolProperty(default=False,name="Pack External Data",description="Whether to pack the External data (like Images and Fonts) into the blend files")
    single_file:bpy.props.BoolProperty(default=False,name='Save assets into one file')
    #Better FBX Importer

    
    use_auto_bone_orientation: BoolProperty(
            name="Automatic Bone Orientation",
            description="Automatically sort bones orientations, if you want to preserve the original armature, please disable the option",
            default=True,
            )

    my_calculate_roll: EnumProperty(
            name="Calculate Roll",
            description="Automatically fix alignment of imported bones’ axes when 'Automatic Bone Orientation' is enabled",
            items=(('POS_X', "POS_X", "POS_X"),
                   ('POS_Z', "POS_Z", "POS_Z"),
                   ('GLOBAL_POS_X', "GLOBAL_POS_X", "GLOBAL_POS_X"),
                   ('GLOBAL_POS_Y', "GLOBAL_POS_Y", "GLOBAL_POS_Y"),
                   ('GLOBAL_POS_Z', "GLOBAL_POS_Z", "GLOBAL_POS_Z"),
                   ('NEG_X', "NEG_X", "NEG_X"),
                   ('NEG_Z', "NEG_Z", "NEG_Z"),
                   ('GLOBAL_NEG_X', "GLOBAL_NEG_X", "GLOBAL_NEG_X"),
                   ('GLOBAL_NEG_Y', "GLOBAL_NEG_Y", "GLOBAL_NEG_Y"),
                   ('GLOBAL_NEG_Z', "GLOBAL_NEG_Z", "GLOBAL_NEG_Z"),
                   ('ACTIVE', "ACTIVE", "ACTIVE"),
                   ('VIEW', "VIEW", "VIEW"),
                   ('CURSOR', "CURSOR", "CURSOR"),
                   ('None', "None", "Does not fix alignment of imported bones’ axes")),
            default='None',
            )

    my_bone_length: FloatProperty(
        name = "Bone Length",
        description = "Bone length when 'Automatic Bone Orientation' is disabled",
        default = 10.0,
        min = 0.0001,
        max = 10000.0)

    my_leaf_bone: EnumProperty(
            name="Leaf Bone",
            description="The length of leaf bone",
            items=(('Long', "Long", "1/1 length of its parent"),
                   ('Short', "Short", "1/10 length of its parent")),
            default='Long',
            )

    use_fix_bone_poses: BoolProperty(
            name="Fix Bone Poses",
            description="Try fixing bone poses with default poses whenever bind poses are not equal to default poses",
            default=False,
            )

    use_fix_attributes: BoolProperty(
            name="Fix Attributes For Unity & C4D",
            description="Try fixing null attributes for Unity's FBX exporter & C4D's FBX exporter, but it may bring extra fake bones",
            default=True,
            )

    use_only_deform_bones: BoolProperty(
            name="Only Deform Bones",
            description="Import only deform bones",
            default=False,
            )

    use_vertex_animation: BoolProperty(
            name="Vertex Animation",
            description="Import vertex animation",
            default=True,
            )

    use_animation: BoolProperty(
            name="Animation",
            description="Import animation",
            default=True,
            )

    my_animation_offset: IntProperty(
        name = "Animation Offset",
        description = "Add an offset to all keyframes",
        default = 0,
        min = -1000000,
        max = 1000000)

    use_animation_prefix: BoolProperty(
            name="Animation Prefix",
            description="Add object name as animation prefix",
            default=False,
            )

    use_triangulate: BoolProperty(
            name="Triangulate",
            description="Triangulate meshes",
            default=False,
            )

    my_import_normal: EnumProperty(
            name="Normal",
            description="How to get normals",
            items=(('Calculate', "Calculate", "Let Blender generate normals"),
                   ('Import', "Import", "Use imported normals")),
            default='Import',
            )

    use_auto_smooth: BoolProperty(
            name="Auto Smooth",
            description="Auto smooth (based on smooth/sharp faces/edges and angle between faces)",
            default=True,
            )

    my_angle: FloatProperty(
        name = "Angle",
        description = "Maximum angle between face normals that will be considered as smooth",
        default = 60.0,
        min = 0.0,
        max = 180.0)

    my_shade_mode: EnumProperty(
            name="Shading",
            description="How to render and display faces",
            items=(('Smooth', "Smooth", "Render and display faces smooth, using interpolated vertex normals"),
                   ('Flat', "Flat", "Render and display faces uniform, using face normals")),
            default='Smooth',
            )

    my_scale: FloatProperty(
        name = "Scale",
        description = "Scale all data",
        default = 0.01,
        min = 0.0001,
        max = 10000.0)

    use_optimize_for_blender: BoolProperty(
            name="Optimize For Blender",
            description="Make Blender friendly rotation and scale. This is an experimental feature, which may has bugs, use at your own risk",
            default=False,
            )

    use_reset_mesh_origin: BoolProperty(
            name="Reset Mesh Origin",
            description="Reset mesh origin to zero",
            default=True,
            )

    use_edge_crease: BoolProperty(
            name="Edge Crease",
            description="Import edge crease",
            default=True,
            )

    my_edge_crease_scale: FloatProperty(
        name = "Edge Crease Scale",
        description = "Scale of the edge crease value",
        default = 1.0,
        min = 0.0001,
        max = 10000.0)

    my_edge_smoothing: EnumProperty(
            name="Smoothing Groups",
            description="How to generate smoothing groups",
            items=(('None', "None", "Does not generate smoothing groups"),
                   ('Import', "Import From File", "Import smoothing groups from file"),
                   ('FBXSDK', "Generate By FBX SDK", "Generate smoothing groups from normals by FBX SDK"),
                   ('Blender', "Generate By Blender", "Generate smoothing groups from normals by Blender")),
            default='FBXSDK',
            )

    use_import_materials: BoolProperty(
            name="Import Materials",
            description="Import materials for meshes, if you don't want to import any materials, you can turn off this option",
            default=True,
            )

    use_rename_by_filename: BoolProperty(
            name="Rename By Filename",
            description="If you want to import a lot of 3d models or armatures in batch, and there is only one mesh or one armature per file, you may turn on this option to rename the imported meshes or armatures by their filenames",
            default=False,
            )

    my_rotation_mode: EnumProperty(
            name="Rotation Mode",
            description="Rotation mode of all objects",
            items=(('QUATERNION', "Quaternion (WXYZ)", "Quaternion (WXYZ), No Gimbal Lock"),
                   ('XYZ', "XYZ Euler", "XYZ Rotation Order - prone to Gimbal Lock"),
                   ('XZY', "XZY Euler", "XZY Rotation Order - prone to Gimbal Lock"),
                   ('YXZ', "YXZ Euler", "YXZ Rotation Order - prone to Gimbal Lock"),
                   ('YZX', "YZX Euler", "YZX Rotation Order - prone to Gimbal Lock"),
                   ('ZXY', "ZXY Euler", "ZXY Rotation Order - prone to Gimbal Lock"),
                   ('ZYX', "ZYX Euler", "ZYX Rotation Order - prone to Gimbal Lock"),
                   ('AXIS_ANGLE', "Axis Angle", "Axis Angle (W+XYZ), defines a rotation around some axis defined by 3D-Vector")),
            default='QUATERNION',
            )
    use_edges: BoolProperty(
        name="Lines",
        description="Import lines and faces with 2 verts as edge",
        default=True,
    )
    use_smooth_groups: BoolProperty(
        name="Smooth Groups",
        description="Surround smooth groups by sharp edges",
        default=True,
    )

    use_split_objects: BoolProperty(
        name="Object",
        description="Import OBJ Objects into Blender Objects",
        default=True,
    )
    use_split_groups: BoolProperty(
        name="Group",
        description="Import OBJ Groups into Blender Objects",
        default=False,
    )

    use_groups_as_vgroups: BoolProperty(
        name="Poly Groups",
        description="Import OBJ groups as vertex groups",
        default=False,
    )

    use_image_search: BoolProperty(
        name="Image Search",
        description="Search subdirs for any associated images "
        "(Warning, may be slow)",
        default=True,
    )

    split_mode: EnumProperty(
        name="Split",
        items=(
            ('ON', "Split", "Split geometry, omits vertices unused by edges or faces"),
            ('OFF', "Keep Vert Order", "Keep vertex order from file"),
        ),
    )

    global_clamp_size: FloatProperty(
        name="Clamp Size",
        description="Clamp bounds under this value (zero to disable)",
        min=0.0, max=1000.0,
        soft_min=0.0, soft_max=1000.0,
        default=0.0,
    )
    clamp_size:FloatProperty(name="Clamp Bounding Box",default=0.520,min=0.0,max=1000)
    global_scale:FloatProperty(name="Scale",description="Scale all data",default=1.0,min=0.0001,max=10000.0)
    validate_meshes:BoolProperty(name="Validate Meshes",description="Validate mesh data",default=False)
    import_vertex_groups:BoolProperty(name="Vertex Groups",default=False)
    
    def draw(self, layout):
        layout.prop(self,'catalog')
        if self.catalog=='NEW':
            layout.prop(self,'new_catalog_name')
        layout.label(text='Bring is as:')
        layout.row().prop(self,'make_collection',expand=True)
        layout.prop(self,'single_file',toggle=True)
        layout.label(text="Thumbnail Shading")
        layout.row().prop(self,'shading',expand=True)
        if self.shading=='Cycles':
            layout.label(text='Might be slow!',icon='INFO')
        if self.shading!='Solid':
            layout.label(text="Camera Angle")
            layout.row().prop(self,'camera_angle',expand=True)
            row=layout.row(align=True)
            row.prop(self,'flip_x',toggle=True)
            row.prop(self,'flip_y',toggle=True)
            row.prop(self,'flip_z',toggle=True)
            if not self.flip_z:
                layout.prop(preferences(),'add_ground_plane',toggle=True)
        layout.prop(self,'pack',toggle=True)
        layout.label(text="What to do for existing files")
        layout.row().prop(self,'override',expand=True)
        layout.prop(self,'max_time')
        layout.label(text="Importer Settings:")
        if preferences().fbx_importer=='Better FBX':
            box = layout.box()
            box.label(text="Basic Options:")
            box.prop(self, 'my_rotation_mode')
            box.prop(self, 'my_import_normal')
            box.prop(self, 'use_auto_smooth')
            box.prop(self, 'my_angle')
            box.prop(self, 'my_shade_mode')
            box.prop(self, 'my_scale')

            box = layout.box()
            box.label(text="Blender Options: (Experimental)")
            box.prop(self, 'use_optimize_for_blender')
            box.prop(self, 'use_reset_mesh_origin')

            box = layout.box()
            box.label(text="Bone Options:")
            box.prop(self, 'use_auto_bone_orientation')
            box.prop(self, 'my_calculate_roll')
            box.prop(self, 'my_bone_length')
            box.prop(self, 'my_leaf_bone')
            box.prop(self, 'use_fix_bone_poses')
            box.prop(self, 'use_fix_attributes')
            box.prop(self, 'use_only_deform_bones')

            box = layout.box()
            box.label(text="Animation Options:")
            box.prop(self, 'use_animation')
            box.prop(self, 'my_animation_offset')
            box.prop(self, 'use_animation_prefix')

            box = layout.box()
            box.label(text="Vertex Animation Options:")
            box.prop(self, 'use_vertex_animation')

            box = layout.box()
            box.label(text="Mesh Options:")
            box.prop(self, 'use_triangulate')
            box.prop(self, 'use_import_materials')
            box.prop(self, 'use_rename_by_filename')

            box = layout.box()
            box.label(text="Edge Options:")
            box.prop(self, 'my_edge_smoothing')
            box.prop(self, 'use_edge_crease')
            box.prop(self, 'my_edge_crease_scale')
        else:
            if bpy.app.version>=(3,5,0):
                layout.use_property_split = True
                layout.use_property_decorate = False  # No animation.

                
                box=layout.box()
                box.label(text="Transform",icon='OBJECT_DATA')
                box.prop(self,'global_scale')
                box.prop(self, "clamp_size")
                box.prop(self, "axis_forward")
                box.prop(self, "axis_up")
                box=layout.box()
                box.label(text='Options',icon='EXPORT')
                box.prop(self, "use_split_objects", text="Split by Object")
                box.prop(self, "use_split_groups", text="Split by Group")
                box.prop(self,'import_vertex_groups')
                box.prop(self,'validate_meshes')
            else:
                layout.label(text="Include:")
                layout.use_property_split = True
                layout.use_property_decorate = False  # No animation.


                layout.prop(self, 'use_image_search')
                layout.prop(self, 'use_smooth_groups')
                layout.prop(self, 'use_edges')
                layout.label(text="Transform:")
                layout.use_property_split = True
                layout.use_property_decorate = False  # No animation.

                layout.prop(self, "global_clamp_size")
                layout.prop(self, "axis_forward")
                layout.prop(self, "axis_up")
                layout.label(text="Geometry")
                layout.row().prop(self, "split_mode", expand=True)

                layout.use_property_split = True
                layout.use_property_decorate = False  # No animation.

                col = layout.column()
                if self.split_mode == 'ON':
                    col.prop(self, "use_split_objects", text="Split by Object")
                    col.prop(self, "use_split_groups", text="Split by Group")
                else:
                    col.prop(self, "use_groups_as_vgroups")
class Create_Assets_From_USD_Props_Base(bpy.types.PropertyGroup,Filters):
    filepath: bpy.props.StringProperty(subtype="DIR_PATH")
    files : bpy.props.CollectionProperty(
            name="File Path",
            type=bpy.types.OperatorFileListElement,
            )
    filter_glob: bpy.props.StringProperty(default="*" + ";*".join(['usd','usdc','usda','usdz']),options={"HIDDEN"})
    max_time:bpy.props.IntProperty(default=1800,name="Maximum Time Per File")
    force_previews:bpy.props.BoolProperty(default=False,name="Force Regenerate Previews")
    camera_angle:bpy.props.EnumProperty(items=(('XZ','X-Axis','X-Axis'),('YZ','Y-Axis','Y-Axis'),('XYZ','Isometric','Equal distance on all axis')))
    flip_z:bpy.props.BoolProperty(name="Flip Z",default=False)
    flip_x:bpy.props.BoolProperty(name="Flip X",default=False)
    flip_y:bpy.props.BoolProperty(name="Flip Y",default=False)
    catalog:bpy.props.EnumProperty(items=get_all_catalogs,name="Catalog",default=3)
    new_catalog_name:bpy.props.StringProperty(default="",name="Catalog Name",search=get_catalog_search_results_from_all)
    make_collection:bpy.props.EnumProperty(default=0,items=[('Collection','Collection','Collection'),('Object','Object','Object')],name='Bring in as:')
    pack:bpy.props.BoolProperty(default=False,name="Pack External Data",description="Whether to pack the External data (like Images and Fonts) into the blend files")
    scale:bpy.props.FloatProperty(default=1.0,name='Scale')
    set_frame_range:bpy.props.BoolProperty(default=True,name='Set frame range')
    import_cameras:bpy.props.BoolProperty(default=True,name='Cameras')
    import_curves:bpy.props.BoolProperty(default=True,name='Curves')
    import_lights:bpy.props.BoolProperty(default=True,name='Lights')
    import_materials:bpy.props.BoolProperty(default=True,name='Materials')
    import_meshes:bpy.props.BoolProperty(default=True,name='Meshes')
    import_volumes:bpy.props.BoolProperty(default=True,name='Volumes')
    import_subdiv:bpy.props.BoolProperty(default=False,name='Subdivision')
    import_instance_proxies:bpy.props.BoolProperty(default=True,name='Import Instance Proxies')
    import_visible_only:bpy.props.BoolProperty(default=True,name='Visible Primitives Only')
    create_collection:bpy.props.BoolProperty(default=False,name='Create collection')
    read_mesh_uvs:bpy.props.BoolProperty(default=True,name='UV Coordinates')
    read_mesh_colors:bpy.props.BoolProperty(default=False,name='Color Attributes')
    prim_path_mask:bpy.props.StringProperty(default='',name='Path mask')
    import_guide:bpy.props.BoolProperty(default=False,name='Guide')
    import_proxy:bpy.props.BoolProperty(default=True,name='Proxy')
    import_render:bpy.props.BoolProperty(default=True,name='Render')
    relative_path:bpy.props.BoolProperty(default=False,name='Relative Path')
    import_usd_preview:bpy.props.BoolProperty(default=True,name='Import USD Preview')
    set_material_blend:bpy.props.BoolProperty(default=True,name='Set material blend')
    light_intensity_scale:bpy.props.FloatProperty(default=1.0,name='Light intensity scale')
    mtl_name_collision_mode:bpy.props.EnumProperty(default='MAKE_UNIQUE',name='Material name collision',items=[('MAKE_UNIQUE','Make Unique','Make Unique'),('REFERENCE_EXISTING','Reference Existing','Reference Existing')])
    import_all_materials:bpy.props.BoolProperty(default=False,name='Import All Materials')
    import_textures_mode:bpy.props.EnumProperty(default='IMPORT_PACK',items=(('IMPORT_NONE','None','None'),('IMPORT_COPY','Copy','Copy'),('IMPORT_PACK','Packed','Packed')),name='Import Textures')
    import_textures_dir:bpy.props.StringProperty(default='//textures/',name='Textures Directory')
    tex_name_collision_mode:bpy.props.EnumProperty(default='USE_EXISTING',name='File name collision',items=[('USE_EXISTING','Use Existing','Use Existing'),('OVERWRITE','Overwrite','Overwrite')])
    
    def draw(self, layout):
        layout.prop(self,'catalog')
        if self.catalog=='NEW':
            layout.prop(self,'new_catalog_name')
        layout.label(text='Bring is as:')
        layout.row().prop(self,'make_collection',expand=True)
        layout.label(text="Thumbnail Shading")
        layout.row().prop(self,'shading',expand=True)
        if self.shading=='Cycles':
            layout.label(text='Might be slow!',icon='INFO')
        if self.shading!='Solid':
            layout.label(text="Camera Angle")
            layout.row().prop(self,'camera_angle',expand=True)
            row=layout.row(align=True)
            row.prop(self,'flip_x',toggle=True)
            row.prop(self,'flip_y',toggle=True)
            row.prop(self,'flip_z',toggle=True)
            if not self.flip_z:
                layout.prop(preferences(),'add_ground_plane',toggle=True)
        layout.prop(self,'pack',toggle=True)
        layout.label(text="What to do for existing files")
        layout.row().prop(self,'override',expand=True)
        layout.prop(self,'max_time')
        layout.label(text="Importer Settings:")
        layout.use_property_split = True
        layout.use_property_decorate = False 
        box=layout.box()
        box.label(text='Data Types:')
        box.prop(self,'import_cameras')
        box.prop(self,'import_curves')
        box.prop(self,'import_lights')
        box.prop(self,'import_materials')
        box.prop(self,'import_meshes')
        box.prop(self,'import_volumes')
        box.prop(self,'prim_path_mask')
        box.prop(self,'scale')
        box=layout.box()
        box.label(text='Mesh Data:')
        box.prop(self,'read_mesh_uvs')
        box.prop(self,'read_mesh_colors')
        box.label(text='Include:')
        box.prop(self,'import_subdiv')
        box.prop(self,'import_instance_proxies')
        box.prop(self,'import_visible_only')
        box.prop(self,'import_guide')
        box.prop(self,'import_proxy')

        box.prop(self,'import_render')
        box.label(text='Options:')
        box.prop(self,'set_frame_range')
        box.prop(self,'relative_path')
        box.prop(self,'create_collection')
        box.prop(self,'light_intensity_scale')
        
        box=layout.box()
        box.prop(self,'import_all_materials')
        box.prop(self,'import_usd_preview')
        row=box.row()
        row.prop(self,'set_material_blend')
        box.prop(self,'mtl_name_collision_mode')
        if not self.import_usd_preview:
            row.enabled=False
        box=layout.box()
        box.prop(self,'import_textures_mode')
        box.prop(self,'import_textures_dir')
        box.prop(self,'tex_name_collision_mode')
class Create_Materials_Props_Base(bpy.types.PropertyGroup):
    bsdf_settings=['ao_amount','emission_strength','enable_opacity_mask','dx_to_ogl_normals']
    bsdf_mapping=['location','rotation','scale']
    uber_settings=['enable_colour_changes',
        'metal_colour',
        'metal_lighness',
        'non_metal_colour',
        'non_metal_lightness',
        'metallic_ness',
        'metal_spots',
        'none_metal_spots',
        'roughness',
        'shiny_spots',
        'rough_spots',
        'anisotropic',
        'anisotropic_rotation',
        'enable_opacity_mask','dx_to_ogl_normals','ao_strength']
    organic_settings=['albedo_hsl',
        'hue',
        'saturation',
        'lightness',
        'sss_hue_and_saturation',
        'sss_hue',
        'sss_saturation',
        'make_sss_human',
        'use_translucent',
        'translucent_amount',
        'use_subsurface_scattering',
        'sss_amount',
        'subsurface_ior',
        'subsurface_anisotropy',
        'roughness',
        'shiny_spots',
        'rough_spots','enable_opacity_mask','dx_to_ogl_normals','ao_strength']
    plastic_settings=['enable_colour_changes_plastic',
        'new_colour',
        'enable_surface_hsl',
        'surface_hue',
        'surface_saturation',
        'surface_lightness',
        'subsurface_hue',
        'subsurface_saturation',
        'roughness',
        'shiny_spots',
        'rough_spots',
        'enable_translucent',
        'translucent_amount',
        'translucent_angle',
        'translucent_frosting',
        'translucent_ior',
        'enable_sss_plastic',
        'sss_amount_plastic','enable_opacity_mask','dx_to_ogl_normals','ao_strength']
    surface_settings=['normal_map_strength',
        'bump_strength',
        'bump_distance',
        'displacement_scale',
        'displacement_midlevel']
    mapping_settings=['tri_planar_mapping',
        'seam_blending',
        'rotate_texture',
        'scale_textures_up',
        'scale_x',
        'scale_y',
        'scale_z',
        'x_location',
        'y_location',
        'z_location']
    only_basecolor:bpy.props.BoolProperty(default=False,name="Create materials even if only Color texture is found!")
    #useBump:bpy.props.BoolProperty(default=False,name="Use Bump Node instead of Displacement Node")
    mark_asset:bpy.props.BoolProperty(default=True,name="Mark as Asset")
    catalog:bpy.props.EnumProperty(items=lambda self, context: get_catalogs_to_move_to_for_mats.catalogs,default=2,name="Catalog")
    new_catalog_name:bpy.props.StringProperty(default="",name="Catalog Name",search=get_catalog_search_results_for_mats)
    dir_path:bpy.props.StringProperty(default="")
    shader_type:bpy.props.EnumProperty(items=(("True-Uber","Uber","True-Uber"),("True-Plastic","Plastic","True-Plastic"),("True-Organic","Organic","True-Organic"),("True-BSDF","BSDF","True-BSDF"),("Standard","Standard","Standard")),name="Shader Type")
    pack:bpy.props.BoolProperty(default=False,name="Pack Images",description="Whether to pack the images into the blend files or not. Packing will use more space")
    show_settings:bpy.props.BoolProperty(default=False,name="Default Values")
    ### Standard Shader Settings
    ao_amount:bpy.props.FloatProperty(name='AO Amount',default=0.0,min=0,max=1)
    emission_strength:bpy.props.FloatProperty(name='Emission Strength',default=0.0,min=0)
    rotation:bpy.props.FloatVectorProperty(name='Rotation',default=(0.0,0.0,0.0),subtype='EULER',size=3)
    scale:bpy.props.FloatVectorProperty(name='Scale',default=(1.0,1.0,1.0),size=3)
    location:bpy.props.FloatVectorProperty(name='Location',default=(0.0,0.0,0.0),size=3)
    ### Uber Shader Settins
    enable_colour_changes:bpy.props.FloatProperty(name='Enable Colour Changes',default=0.0,min=0,max=1)
    metal_colour:bpy.props.FloatVectorProperty(name='Metal Colour',default=(0,0,0,1),size=4,subtype ='COLOR',min=0,max=1)
    metal_lighness:bpy.props.FloatProperty(name='Metal Lighness',default=1.0,min=0,max=1)
    non_metal_colour:bpy.props.FloatVectorProperty(name='Non-Metal Colour',default=(0,0,0,1),size=4,subtype ='COLOR',min=0,max=1)
    non_metal_lightness:bpy.props.FloatProperty(name='Non-Metal Lightness',default=1.0,min=0,max=1)
    metallic_ness:bpy.props.FloatProperty(name='Metallic-ness',default=0.5,min=0,max=1)
    metal_spots:bpy.props.FloatProperty(name='Metal Spots',default=0.0,min=0,max=1)
    none_metal_spots:bpy.props.FloatProperty(name='None Metal Spots',default=0.0,min=0,max=1)
    roughness:bpy.props.FloatProperty(name='Roughness',default=0.5,min=0,max=1)
    shiny_spots:bpy.props.FloatProperty(name='Shiny Spots',default=0.0,min=0,max=1)
    rough_spots:bpy.props.FloatProperty(name='Rough Spots',default=0.0,min=0,max=1)
    anisotropic:bpy.props.FloatProperty(name='Anisotropic',default=0.0,min=0,max=1)
    anisotropic_rotation:bpy.props.FloatProperty(name='Anisotropic Rotation',default=0.0,min=0,max=1)
    enable_opacity_mask:bpy.props.FloatProperty(name='Enable opacity mask',default=1.0,min=0,max=1)
    
    
    ### Organic Shader Settings

    albedo_hsl:bpy.props.FloatProperty(name='Albedo HSL (Affects Trans and SSS)',default=0.0,min=0,max=1)
    hue:bpy.props.FloatProperty(name='Hue',default=0.5,min=0,max=1)
    saturation:bpy.props.FloatProperty(name='Saturation',default=1.0,min=0,max=1)
    lightness:bpy.props.FloatProperty(name='Lightness',default=1.0,min=0,max=1)
    sss_hue_and_saturation:bpy.props.FloatProperty(name='SSS Hue and Saturation',default=0.0,min=0,max=1)
    sss_hue:bpy.props.FloatProperty(name='SSS Hue',default=0.5,min=0,max=1)
    sss_saturation:bpy.props.FloatProperty(name='SSS Saturation',default=1.0,min=0,max=1)
    make_sss_human:bpy.props.FloatProperty(name='Make SSS Human (Blood)',default=0.0,min=0,max=1)
    use_translucent:bpy.props.FloatProperty(name='Use Translucent (Alpha-masked mats only)',default=0.0,min=0,max=1)
    translucent_amount:bpy.props.FloatProperty(name='Translucent amount',default=1.0,min=0,max=1)
    use_subsurface_scattering:bpy.props.FloatProperty(name='Use Subsurface Scattering (Manifold Only)',default=0.0,min=0,max=1)
    sss_amount:bpy.props.FloatProperty(name='SSS Amount',default=1.0,min=0,max=1)
    subsurface_ior:bpy.props.FloatProperty(name='Subsurface IOR',default=1.399999976158142,min=0,max=1)
    subsurface_anisotropy:bpy.props.FloatProperty(name='Subsurface Anisotropy',default=0.0,min=0,max=1)
    ### Plastic Shader Setttings

    enable_colour_changes_plastic:bpy.props.FloatProperty(name='Enable Colour Changes',default=1.0,min=0,max=1)
    new_colour:bpy.props.FloatVectorProperty(name='New Colour',default=(0,0,0,1),size=4,subtype ='COLOR',min=0,max=1)
    enable_surface_hsl:bpy.props.FloatProperty(name='Enable surface HSL',default=1.0,min=0,max=1)
    surface_hue:bpy.props.FloatProperty(name='Surface Hue',default=0.5,min=0,max=1)
    surface_saturation:bpy.props.FloatProperty(name='Surface Saturation',default=1.0,min=0,max=1)
    surface_lightness:bpy.props.FloatProperty(name='Surface Lightness',default=1.0,min=0,max=1)
    subsurface_hue:bpy.props.FloatProperty(name='SubSurface Hue',default=0.5,min=0,max=1)
    subsurface_saturation:bpy.props.FloatProperty(name='SubSurface Saturation',default=1.0,min=0,max=1)
    enable_translucent:bpy.props.FloatProperty(name='Enable Translucent',default=0.0,min=0,max=1)
    translucent_angle:bpy.props.FloatProperty(name='Translucent Angle',default=0.5,min=0,max=1)
    translucent_frosting:bpy.props.FloatProperty(name='Translucent Frosting',default=0.0,min=0,max=1)
    translucent_ior:bpy.props.FloatProperty(name='Translucent IOR',default=1.2999999523162842,min=0,max=1)
    enable_sss_plastic:bpy.props.FloatProperty(name='Enable SSS Plastic',default=0.0,min=0,max=1)
    sss_amount_plastic:bpy.props.FloatProperty(name='SSS Amount',default=0.23000001907348633,min=0,max=1)
    ### Common Settings
    ao_strength:bpy.props.FloatProperty(name='AO Map Strength',default=0.5,min=0,max=1)
    dx_to_ogl_normals:bpy.props.FloatProperty(name='DX to OGL Normals',default=0.0,min=0,max=1)
    normal_map_strength:bpy.props.FloatProperty(name='Normal map strength',default=1.0,min=0,max=10)
    bump_strength:bpy.props.FloatProperty(name='Bump Strength',default=0.10000000149011612,min=0,max=1)
    bump_distance:bpy.props.FloatProperty(name='Bump Distance',default=0.25,min=0,max=1)
    displacement_scale:bpy.props.FloatProperty(name='Displacement Scale',default=0.0,min=0,max=1000)
    displacement_midlevel:bpy.props.FloatProperty(name='Displacement Midlevel',default=0.5,min=0,max=1000)
    tri_planar_mapping:bpy.props.FloatProperty(name='Tri-Planar mapping',default=0.0,min=0,max=1)
    seam_blending:bpy.props.FloatProperty(name='Seam Blending (only for Tri-planar)',default=0.02592591941356659,min=0,max=1)
    rotate_texture:bpy.props.FloatProperty(name='Rotate texture',default=0.0)
    scale_textures_up:bpy.props.FloatProperty(name='Scale textures up',default=1.0)
    scale_x:bpy.props.FloatProperty(name='Scale X',default=1.0)
    scale_y:bpy.props.FloatProperty(name='Scale Y',default=1.0)
    scale_z:bpy.props.FloatProperty(name='Scale Z (Only works with Tri-Planar)',default=1.0)
    x_location:bpy.props.FloatProperty(name='X Location',default=0.0)
    y_location:bpy.props.FloatProperty(name='Y Location',default=0.0)
    z_location:bpy.props.FloatProperty(name='Z Location (Only works with Tri-Planar)',default=0.0)
    
    def draw(self, layout):
        # layout.prop(self,'mark_asset',toggle=True)
        if hasattr(self,'filter_search'):
            layout.prop(self,'catalog')
            if self.catalog=='NEW':
                layout.prop(self,'new_catalog_name')
        layout.prop(self,'pack',toggle=True)
        layout.row().prop(self,'shader_type',expand=True)
        if hasattr(self,'filter_search'):
            layout.prop(self,'filter_search')
        layout.prop(self,'show_settings',icon='TRIA_DOWN',toggle=True)
        if self.shader_type!='Standard':
            if self.show_settings:
                layout.label(text="Material Settings")
                if self.shader_type =='True-Uber':
                    for prop in self.uber_settings:
                        layout.prop(self,prop)
                elif self.shader_type =='True-BSDF':
                    for prop in self.bsdf_settings:
                        layout.prop(self,prop)
                elif self.shader_type =='True-Organic':
                    for prop in self.organic_settings:
                        layout.prop(self,prop)
                else:
                    for prop in self.plastic_settings:
                        layout.prop(self,prop)
                layout.label(text="Surface Settings")
                for prop in self.surface_settings:
                    layout.prop(self,prop)
                layout.label(text="Mapping")
                if self.shader_type=='True-BSDF':
                    for prop in self.bsdf_mapping:
                        layout.prop(self,prop)
                else:
                    for prop in self.mapping_settings:
                        layout.prop(self,prop)
class Create_Worlds_Props_Base(bpy.types.PropertyGroup):
    settings=['tonned_mapped','light_intensity', 'exposure','rotate', 'clean_horizon', 'ev_compensation', 'background_blur', 'hsl_on_off', 'hue', 'saturation','contrast', 'atmosphere_effector', 'glow_effector', 'hsl_uniform', 'sky_tint', 'tint_colour', 'drop_horizon', 'horizon_position', 'light_source_matcher']  
    standard_settings=['strength','rotation']
    use_hdr:bpy.props.BoolProperty(default=True,name="HDR")
    use_exr:bpy.props.BoolProperty(default=True,name="EXR")
    use_jpg:bpy.props.BoolProperty(default=False,name="JPG")
    use_jpeg:bpy.props.BoolProperty(default=False,name="JPEG")
    use_png:bpy.props.BoolProperty(default=False,name="PNG")
    search_subfolders:bpy.props.BoolProperty(default=True,name="Search Subfolders")
    mark_asset:bpy.props.BoolProperty(default=True,name="Mark as Asset")
    catalog:bpy.props.EnumProperty(items=lambda self, context: get_catalogs_to_move_to_for_worlds.catalogs,name="Catalog")
    new_catalog_name:bpy.props.StringProperty(default="",name="Catalog Name",search=get_catalog_search_results_for_worlds)
    dir_path:bpy.props.StringProperty(default="")
    pack:bpy.props.BoolProperty(default=False,name="Pack Images",description="Whether to pack the images into the blend files or not. Packing will use more space")
    show_world_settings:bpy.props.BoolProperty(default=False,name="Default Values")
    shader_type:bpy.props.EnumProperty(items=(("True-HDRI","True-HDRI","True-HdRi"),("True-HDRI (Standard)","Standard","True-HDRI Standard")),name="Shader Type")
    
    ##Node Group Settings
    tonned_mapped:bpy.props.FloatProperty(name='Tonemapped Toggle',default=0,min=0,max=1)
    light_intensity:bpy.props.FloatProperty(name='Light Intensity',default=1.0,min=0,max=1)
    exposure:bpy.props.FloatProperty(name='Exposure',default=5.0,min=0,max=1)
    clean_horizon:bpy.props.FloatProperty(name='Clean Horizon',default=0.0,min=0,max=1)
    ev_compensation:bpy.props.FloatProperty(name='EV Compensation',default=5.0,min=0,max=1)
    rotate:bpy.props.FloatProperty(name='Rotate',default=0.0,min=0,max=1)
    background_blur:bpy.props.FloatProperty(name='Background Blur',default=0.0,min=0,max=1)
    hsl_on_off:bpy.props.FloatProperty(name='HSL ON/ OFF',default=0.0,min=0,max=1)
    hue:bpy.props.FloatProperty(name='Hue',default=1.0,min=0,max=1)
    saturation:bpy.props.FloatProperty(name='Saturation',default=1.0,min=0,max=1)
    atmosphere_effector:bpy.props.FloatProperty(name='Atmosphere Effector',default=0.0,min=0,max=1)
    glow_effector:bpy.props.FloatProperty(name='Glow Effector',default=0.25,min=0,max=1)
    hsl_uniform:bpy.props.FloatProperty(name='HSL uniform',default=0.0,min=0,max=1)
    sky_tint:bpy.props.FloatProperty(name='Sky Tint',default=0.0,min=0,max=1)
    tint_colour:bpy.props.FloatVectorProperty(name='Tint colour',default=(1,1,1,1),size=4,subtype='COLOR',min=0,max=1)
    drop_horizon:bpy.props.FloatProperty(name='Drop Horizon',default=0.5,min=0,max=1)
    horizon_position:bpy.props.FloatProperty(name='Horizon Position',default=0.5,min=0,max=1)
    light_source_matcher:bpy.props.FloatProperty(name='Light source matcher',default=0.5,min=0,max=1)
    contrast:bpy.props.FloatProperty(name='Contrast',default=0,min=-10,max=10)
    ### Standard Settings
    strength:bpy.props.FloatProperty(name='Strength',default=1.0,min=0)
    rotation:bpy.props.FloatVectorProperty(name='Rotation',default=(0,0,0))
    
    def draw(self, layout):
        # layout.prop(self,'mark_asset',toggle=True)
        if hasattr(self,'filter_search'):
            layout.prop(self,'catalog')
            if self.catalog=='NEW':
                layout.prop(self,'new_catalog_name')
        layout.label(text="Filters:")
        row=layout.row(align=True)
        row.prop(self,'use_hdr',toggle=True)
        row.prop(self,'use_exr',toggle=True)
        row.prop(self,'use_jpg',toggle=True)
        row.prop(self,'use_jpeg',toggle=True)
        row.prop(self,'use_png',toggle=True)
        layout.separator(factor=1)
        layout.prop(self,'search_subfolders',toggle=True)
        if hasattr(self,'filter_search'):
            layout.prop(self,'filter_search')
        layout.prop(self,'pack',toggle=True)
        layout.row().prop(self,'shader_type',expand=True)
        layout.prop(self,'show_world_settings',toggle=True,icon='TRIA_DOWN')
        if self.show_world_settings:
            if self.shader_type=='True-HDRI':
                for a in self.settings:
                    layout.prop(self,a)
                    if a=='background_blur':
                        layout.label(text="Colour Values")
                    elif a=='tint_colour':
                        layout.label(text="Advanced Settings")
            else:
                for a in self.standard_settings:
                    layout.prop(self,a)
        #layout.prop(self,'useBump',toggle=True)
class TA_Prefs(bpy.types.AddonPreferences,AddonUpdateChecker):
    def _update_use_studio_background(self,context):
        if self.use_studio_background:
            self.add_ground_plane=False
    def _update_add_ground_plane(self,context):
        if self.add_ground_plane:
            self.use_studio_background=False
    def _draw_modules_ui(self,layout):
        box=layout.box()
        row=box.row()
        row=row.split(factor=0.8,align=True)
        row.alignment='CENTER'
        row.scale_y=2
        row.operator("ta.installmodule")
        row.operator("wm.url_open",text='Get Modules',icon='URL').url="https://store.true-vfx.xyz/"
        installed_modules={mod.__name__.split('.')[-1]:str(getattr(mod,'VERSION',"")).strip('(').strip(')').replace(',','.') for mod in EXTRA_MODULES}
        for mod in ['ShareTextures','AmbientCG']:
            
            box.label(text=f"{mod} {'(Version:'+installed_modules[mod]+')' if mod in installed_modules.keys() else 'Not Installed'}",icon='CHECKMARK' if mod in installed_modules.keys() else 'PANEL_CLOSE')
    bl_idname = __package__
    ex:bpy.props.BoolProperty(name='Experimental Options',description='Show Experimental settings',default=False)
    thumbnail_crash_fix:bpy.props.BoolProperty(name='Blender 3.5 Crash Fix',description='Fix crashing when using batch fbx/obj/usd asset creation by using Custom Solid shading mode',default=False)
    auto_clear_search:bpy.props.BoolProperty(name='Auto Clear Search',description="Auto clear the search field when switching libraries",default=True,update=update_auto_clear)
    ta_directory:bpy.props.StringProperty(default=os.path.join(os.path.dirname(os.path.dirname(__file__)), 'True Assets'),subtype='DIR_PATH',name="True Assets Directory",update=master_dir_changed)
    objects_directory:bpy.props.StringProperty(default=os.path.join(os.path.dirname(os.path.dirname(__file__)), 'TA-Objects'),subtype='DIR_PATH',name="Objects Directory")
    worlds_directory:bpy.props.StringProperty(default=os.path.join(os.path.dirname(os.path.dirname(__file__)), 'TA-Materials'),subtype='DIR_PATH',name="Worlds Directory")
    materials_directory:bpy.props.StringProperty(default=os.path.join(os.path.dirname(os.path.dirname(__file__)), 'TA-Worlds'),subtype='DIR_PATH',name="Materials Directory")
    node_groups_directory:bpy.props.StringProperty(default=os.path.join(os.path.dirname(os.path.dirname(__file__)), 'TA-NodeGroups'),subtype='DIR_PATH',name='Node Groups Directory')
    online_assets_directory:bpy.props.StringProperty(default=os.path.join(os.path.dirname(os.path.dirname(__file__)), 'TA-AmbientCG'),subtype='DIR_PATH',name='Ambient Directory')
    downloads_directory:bpy.props.StringProperty(default=os.path.join(os.path.dirname(os.path.dirname(__file__)), 'TA-Downloaded Assets'),subtype='DIR_PATH',name='Downloaded Assets',update=update_add_to_ab)
    add_downloads_directory_to_ab:bpy.props.BoolProperty(default=False,name='Add as library',description='Add downloads directory as a library in the asset browser to access downloaded assets',update=update_add_to_ab)
    blends_directory:bpy.props.StringProperty(default=os.path.join(os.path.dirname(os.path.dirname(__file__)), 'TA-Blend Files'),subtype='DIR_PATH',name='Blends Directory')
    object_library_name:bpy.props.StringProperty(default="True-Assets (Objects)",name="Object Library Name",update=lib_name_changed)
    material_library_name:bpy.props.StringProperty(default="True-Assets (Materials)",name="Material Library Name",update=lib_name_changed)
    world_library_name:bpy.props.StringProperty(default="True-Assets (Worlds)",name="World Library Name",update=lib_name_changed)
    node_group_library_name:bpy.props.StringProperty(default="True-Assets (Node Groups)",name="Node Group Library Name",update=lib_name_changed)
    online_library_name:bpy.props.StringProperty(default="True-Assets (Online Assets)",name="Online Library Name",update=lib_name_changed)
    blends_library_name:bpy.props.StringProperty(default="True-Assets (Blend Files)",name="Blend Files Library Name",update=lib_name_changed)
    
    non_blocking:bpy.props.BoolProperty(default=True,name="Use Separate threads for tasks (Non-Blocking)")
    add_ground_plane:bpy.props.BoolProperty(default=False,name="Add Ground Plane",description="Add ground plane for thumbnails",update=_update_add_ground_plane)
    use_studio_background:bpy.props.BoolProperty(default=False,name="Use Studio Background",description="Use Studio Background",update=_update_use_studio_background)
    world_strength:bpy.props.FloatProperty(default=5,name="World Strength (For Rendered Thumbnails)")
    fbx_importer:bpy.props.EnumProperty(items=get_available_importers,name="Importer to use")
    show_default_settings:bpy.props.BoolProperty(default=False,name="Default Settings")
    show_default_settings_this_file:bpy.props.BoolProperty(default=False,name="Mark Assets from This File")
    default_settings_this_file:bpy.props.PointerProperty(type=Create_Assets_From_This_File_Props_Base)
    show_default_settings_from_directory:bpy.props.BoolProperty(default=False,name="Create Assets From Directory")
    default_settings_from_directory:bpy.props.PointerProperty(type=Create_Assets_From_Directory_Props_Base)
    show_default_settings_fbx:bpy.props.BoolProperty(default=False,name="Create Assets From FBX files")
    default_settings_fbx:bpy.props.PointerProperty(type=Create_Assets_FBX_Props_Base)
    show_default_settings_obj:bpy.props.BoolProperty(default=False,name="Create Assets From OBJ files")
    default_settings_obj:bpy.props.PointerProperty(type=Create_Assets_From_OBJ_Props_Base)
    show_default_settings_usd:bpy.props.BoolProperty(default=False,name="Create Assets From USD files")
    default_settings_usd:bpy.props.PointerProperty(type=Create_Assets_From_USD_Props_Base)
    show_default_settings_materials:bpy.props.BoolProperty(default=False,name="Create Materials")
    default_settings_materials:bpy.props.PointerProperty(type=Create_Materials_Props_Base)
    show_default_settings_worlds:bpy.props.BoolProperty(default=False,name="Create Worlds")
    default_settings_worlds:bpy.props.PointerProperty(type=Create_Worlds_Props_Base)
    def draw_default_settings(self,layout,setting_name):
        box=layout.box()
        box.prop(self,"show_"+setting_name,toggle=True,icon="DOWNARROW_HLT" if getattr(self,"show_"+setting_name,False) else "RIGHTARROW_THIN",emboss=False)
        if getattr(self,"show_"+setting_name):
            getattr(self,setting_name).draw(box)
    def draw(self, context):
        layout=self.layout
        draw_update_section_for_prefs(layout)
        self._draw_modules_ui(layout)
        draw_hotkeys(layout,"3D View")
        
        layout.prop(self,'ta_directory')
        layout.prop(self,'downloads_directory')
        row=layout.row()
        row=row.split(factor=0.6)
        row.label(text="")
        row.prop(self,'add_downloads_directory_to_ab',icon='CHECKBOX_HLT' if self.add_downloads_directory_to_ab else 'CHECKBOX_DEHLT',toggle=True)
        box=layout.box()
        box.label(text="Library Names")
        # same_path=[a for a in bpy.context.preferences.filepaths.asset_libraries if a.path==self.objects_directory]
        # if same_path:
        #     box.prop(same_path[0],'name',text="Object Library Name")
        # same_path=[a for a in bpy.context.preferences.filepaths.asset_libraries if a.path==self.materials_directory]
        # if same_path:
        #     box.prop(same_path[0],'name',text="Material Library Name")
        # same_path=[a for a in bpy.context.preferences.filepaths.asset_libraries if a.path==self.worlds_directory]
        # if same_path:
        #     box.prop(same_path[0],'name',text="World Library Name")
        # same_path=[a for a in bpy.context.preferences.filepaths.asset_libraries if a.path==self.node_groups_directory]
        # if same_path:
        #     box.prop(same_path[0],'name',text="Node Group Library Name")
        box.prop(self,'object_library_name')
        box.prop(self,'material_library_name')
        box.prop(self,'world_library_name')
        box.prop(self,'node_group_library_name')
        box.prop(self,'blends_library_name')
        box.prop(self,'online_library_name')
        
        layout.separator()
        layout.prop(self,'world_strength')
        layout.prop(self,'auto_clear_search',toggle=False)
        layout.prop(self,'fbx_importer')
        layout.separator()

        # layout.label(text="Performance:")
        # layout.prop(self,'non_blocking',toggle=True)
        
        layout.label(text="Windows Context Menu:")
        layout.operator('ta.createcontextmenuentries')
        layout.operator('ta.removecontextmenuentries')
        box=layout.box()
        box.prop(self,'show_default_settings',toggle=True,icon="DOWNARROW_HLT" if self.show_default_settings else "RIGHTARROW_THIN",emboss=False)
        if self.show_default_settings:
            
            self.draw_default_settings(box,'default_settings_this_file')
            self.draw_default_settings(box,'default_settings_from_directory')
            self.draw_default_settings(box,'default_settings_fbx')
            self.draw_default_settings(box,'default_settings_obj')
            self.draw_default_settings(box,'default_settings_usd')
            self.draw_default_settings(box,'default_settings_materials')
            self.draw_default_settings(box,'default_settings_worlds')
        # box=layout.box()
        # box.prop(self,'ex',toggle=True,icon="DOWNARROW_HLT" if self.ex else "RIGHTARROW_THIN",emboss=False)
        # if self.ex:
        #     box.prop(self,'thumbnail_crash_fix')
        # layout.prop(self,'objects_directory')
        # layout.prop(self,'materials_directory')
        # layout.prop(self,'worlds_directory')
        # layout.prop(self,'node_groups_directory')
class TA_PT_ASSET_BROWSER(bpy.types.Panel):
    bl_idname = "TA_PT_ASSET_BROWSER"
    bl_label = "True-Assets"
    bl_space_type = "FILE_BROWSER"
    bl_region_type = "TOOLS"
    bl_category = "True-VFX"
    bl_order=1000
    bl_options={'HIDE_HEADER',}
    @classmethod
    def poll(cls,context):
        if SpaceAssetInfo.is_asset_browser(context.space_data) and context.space_data.params.asset_library_reference == preferences().online_library_name:
            return True
        return False
    def draw(self, context):
        layout = self.layout
        pcoll = icon_collection["icons"]
        ambientcg_icon = pcoll["ambientcg"]
        sharetextures_icon = pcoll["sharetextures"]
        layout.label(text="Support Creators",icon='FUND')

        layout.operator('wm.url_open', text="Support AmbientCG on Patreon",
                         icon_value=ambientcg_icon.icon_id).url = "https://www.patreon.com/ambientCG"
        if get_ta_module("ShareTextures"):
            layout.operator('wm.url_open', text="Support ShareTextures on Patreon",
                         icon_value=sharetextures_icon.icon_id).url = "https://www.patreon.com/sharetextures"


class TA_PT_Help(bpy.types.Panel):
    bl_idname = "TA_PT_Help"
    bl_label = "Help"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_category = "True-VFX"
    bl_parent_id='TRUE_ASSETS_PT_TA'
    bl_options={'DEFAULT_CLOSED',}
    def draw(self, context):
        layout = self.layout
        scene = context.scene
        col=layout.column()
        pcoll = icon_collection["icons"]
        youtube_icon = pcoll["youtube"]
        discord_icon=pcoll["discord"]
        col.operator('wm.url_open',text="Documentation",icon="HELP").url="https://docs.true-vfx.xyz/docs/true-assets"
        row=col.row(align=True)
        row.operator('wm.url_open',text="Chat Support",icon_value=discord_icon.icon_id).url="https://discord.gg/true-vfx"
        row.operator('wm.url_open',text="Tutorials",icon_value=youtube_icon.icon_id).url="https://www.youtube.com/watch?v=oZFfgtgIPng&list=PLTEbsU2fMZm9bqHG9jp_0DiSKGhiIMPFK"
        
class TA_PT_TA(bpy.types.Panel):
    bl_label = ""
    bl_idname = "TRUE_ASSETS_PT_TA"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_category='True-VFX'
    
    def draw_header(self,context):
        self.layout.label(text="True Assets")
        self.layout.operator("ta.resetprogress",icon='FILE_REFRESH',text="")
    def draw(self, context):
        layout=self.layout
        draw_update_section_for_panel(layout,context)
        pcoll = icon_collection["icons"]
        ambientcg_icon=pcoll['ambientcg']
        box=layout.box()
        box.label(text='Libraries:')
        row=box.row(align=False)
        
        row.operator('ta.opentalibrary',text='',icon='OBJECT_DATA',emboss=False).name=preferences().object_library_name
        row.operator('ta.opentalibrary',text='',icon='MATERIAL',emboss=False).name=preferences().material_library_name
        row.operator('ta.opentalibrary',text='',icon='WORLD',emboss=False).name=preferences().world_library_name
        row.operator('ta.opentalibrary',text='',icon='NODETREE',emboss=False).name=preferences().node_group_library_name
        row.operator('ta.opentalibrary',text='',icon='BLENDER',emboss=False).name=preferences().blends_library_name
        row.operator('ta.opentalibrary',text='',icon_value=ambientcg_icon.icon_id,emboss=False).name=preferences().online_library_name
        layout.operator("ta.createassetsfromthisfile",text="Mark assets from this file")
        layout.operator("ta.createassetsfromdirectory",text=TA_OT_Assets_From_Directory.bl_label if context.scene.ta_progress=="" else context.scene.ta_progress)
        layout.operator("ta.createfbxassetsfromdirectory",text=TA_OT_FBX_Assets_From_Directory.bl_label if context.scene.ta_progress=="" else context.scene.ta_progress)
        layout.operator("ta.createobjassetsfromdirectory",text=TA_OT_OBJ_Assets_From_Directory.bl_label if context.scene.ta_progress=="" else context.scene.ta_progress)
        layout.operator("ta.createusdassetsfromdirectory",text=TA_OT_USD_Assets_From_Directory.bl_label if context.scene.ta_progress=="" else context.scene.ta_progress)
        
        layout.operator("ta.creatematerialsfromdirectory",text=TA_OT_Materials_From_Directory.bl_label if context.scene.ta_progress=="" else context.scene.ta_progress)
        layout.operator("ta.createworldsfromdirectory",text=TA_OT_WORLDS_From_Directory.bl_label if context.scene.ta_progress=="" else context.scene.ta_progress)
        layout.operator("ta.switchshader")
        #print(["True-" in a.label for a in context.active_object.active_material.node_tree.nodes])
        layout.separator()
        layout.prop(context.scene,'show_material_settings',toggle=True,icon='TRIA_DOWN')

        if context.scene.show_material_settings:
            
            box= layout.box()
            
            if context.active_object and context.active_object.active_material and context.active_object.active_material.use_nodes and (any([a.is_ta_node for a in context.active_object.active_material.node_tree.nodes]) or any(["True-" in a.label for a in context.active_object.active_material.node_tree.nodes])):
                if any([a.is_ta_node for a in context.active_object.active_material.node_tree.nodes]):
                    for node in context.active_object.active_material.node_tree.nodes:
                        count=0
                        toggle_count=0
                        adv_drawen=False
                        if node.is_ta_node:
                            
                            for inp in node.inputs:
                                if '*t' in inp.name:
                                    toggle_count+=1
                                if '--' in inp.name:
                                    box2=box.column()
                                    box2.use_property_split = True
                                    # box2.use_property_decorate = False 
                                    count+=1
                                    adv_drawen=False
                                    adv_props=[]
                                    drop_down_col=box2.column()
                                    drop_down_col.use_property_split = False
                                    drop_down_col.prop(context.scene.sub_sections,f'show_material_section_{count}',text=inp.name.replace('-',''),icon="DOWNARROW_HLT" if getattr(context.scene.sub_sections,f"show_material_section_{count}",False) else "RIGHTARROW_THIN",emboss=False)
                                    if getattr(context.scene.sub_sections,f"show_material_section_{count}"):
                                        box2=box2.box().column(align=True)
                                    
                                else:
                                    if getattr(context.scene.sub_sections,f"show_material_section_{count}"):
                                        
                                        if '*a' in inp.name:
                                            if not adv_drawen:
                                                adv_drawen=True
                                                box3=box2.column()
                                                drop_down_col=box3.column()
                                                drop_down_col.use_property_split = False
                                                drop_down_col.prop(context.scene.sub_sections,f'show_material_adv_section_{count}',text="Show Advanced Settings",icon="DOWNARROW_HLT" if getattr(context.scene.sub_sections,f"show_material_adv_section_{count}",False) else "RIGHTARROW_THIN",emboss=False)
                                                if getattr(context.scene.sub_sections,f"show_material_adv_section_{count}"):
                                                    box3=box3.box()
                                            if getattr(context.scene.sub_sections,f"show_material_adv_section_{count}"):
                                                if '*t' in inp.name:

                                                    box3.prop(node.ta_toggles,f'toggle_{toggle_count}',toggle=True,text=inp.name[:inp.name.index('*') if '*' in inp.name else len(inp.name)])
                                                    
                                                else:
                                                    row=box3.row()
                                                    row.prop(inp,'default_value',text=inp.name[:inp.name.index('*') if '*' in inp.name else len(inp.name)])
                                                    if '*d' in inp.name:
                                                        target=int(inp.name[inp.name.index('*d')+2:inp.name.index('*d')+5])
                                                        if node.inputs[target].default_value!=1:
                                                            row.enabled=False
                                        else:
                                                if '*t' in inp.name:
                                                    box2.prop(node.ta_toggles,f'toggle_{toggle_count}',toggle=True,text=inp.name[:inp.name.index('*') if '*' in inp.name else len(inp.name)])
                                                else:
                                                    row=box2.row()
                                                    row.prop(inp,'default_value',text=inp.name[:inp.name.index('*') if '*' in inp.name else len(inp.name)])
                                                    if '*d' in inp.name:
                                                        target=int(inp.name[inp.name.index('*d')+2:inp.name.index('*d')+5])
                                                        if node.inputs[target].default_value!=1:
                                                            row.enabled=False
                else:
                    count=1
                    for node in context.active_object.active_material.node_tree.nodes:
                        for inp in node.inputs:
                            if '--' in inp.name:
                                count+=1
                                box.prop(context.scene.sub_sections,f'show_material_section_{count}',text=inp.name.replace('-',''),icon="DOWNARROW_HLT" if getattr(context.scene.sub_sections,f"show_material_section_{count}",False) else "RIGHTARROW_THIN",emboss=False)
                                
                            else:
                                if getattr(context.scene.sub_sections,f"show_material_section_{count}"):
                                    box.prop(inp,'default_value',text=inp.name)
            else:   
                box.label(text="Make sure True-Asset material is active!",icon='ERROR')
        layout.prop(context.scene,'show_hdri_settings',toggle=True,icon='TRIA_DOWN')
        if context.scene.show_hdri_settings:
            
            box= layout.box()
            if context.scene.world and context.scene.world.node_tree  and (any([a.is_ta_node for a in context.scene.world.node_tree.nodes]) or any(["True" in a.name for a in context.scene.world.node_tree.nodes])):
                if any([a.is_ta_node for a in context.scene.world.node_tree.nodes]):
                    for node in context.scene.world.node_tree.nodes:
                        count=0
                        toggle_count=0
                        if node.is_ta_node:
                            for inp in node.inputs:
                                if '*t' in inp.name:
                                    toggle_count+=1
                                if '--' in inp.name:
                                    box2=box.column()
                                    box2.use_property_split = True
                                    # box2.use_property_decorate = False 
                                    count+=1
                                    adv_drawen=False
                                    adv_props=[]
                                    drop_down_col=box2.column()
                                    drop_down_col.use_property_split = False
                                    drop_down_col.prop(context.scene.sub_sections,f'show_world_section_{count}',text=inp.name.replace('-',''),icon="DOWNARROW_HLT" if getattr(context.scene.sub_sections,f"show_material_section_{count}",False) else "RIGHTARROW_THIN",emboss=False)
                                    if getattr(context.scene.sub_sections,f"show_world_section_{count}"):
                                        box2=box2.box().column(align=True)
                                    
                                else:
                                    if getattr(context.scene.sub_sections,f"show_world_section_{count}"):
                                        
                                        if '*a' in inp.name:
                                            if not adv_drawen:
                                                adv_drawen=True
                                                box3=box2.column()
                                                drop_down_col=box3.column()
                                                drop_down_col.use_property_split = False
                                                drop_down_col.prop(context.scene.sub_sections,f'show_world_adv_section_{count}',text="Show Advanced Settings",icon="DOWNARROW_HLT" if getattr(context.scene.sub_sections,f"show_material_adv_section_{count}",False) else "RIGHTARROW_THIN",emboss=False)
                                                if getattr(context.scene.sub_sections,f"show_world_adv_section_{count}"):
                                                    box3=box3.box()
                                            if getattr(context.scene.sub_sections,f"show_world_adv_section_{count}"):
                                                if '*t' in inp.name:

                                                    box3.prop(node.ta_toggles,f'toggle_{toggle_count}',toggle=True,text=inp.name[:inp.name.index('*') if '*' in inp.name else len(inp.name)])
                                                    
                                                else:
                                                    row=box3.row()
                                                    row.prop(inp,'default_value',text=inp.name[:inp.name.index('*') if '*' in inp.name else len(inp.name)])
                                                    if '*d' in inp.name:
                                                        target=int(inp.name[inp.name.index('*d')+2:inp.name.index('*d')+5])
                                                        if node.inputs[target].default_value!=1:
                                                            row.enabled=False
                                        else:
                                                if '*t' in inp.name:
                                                    box2.prop(node.ta_toggles,f'toggle_{toggle_count}',toggle=True,text=inp.name[:inp.name.index('*') if '*' in inp.name else len(inp.name)])
                                                else:
                                                    row=box2.row()
                                                    row.prop(inp,'default_value',text=inp.name[:inp.name.index('*') if '*' in inp.name else len(inp.name)])
                                                    if '*d' in inp.name:
                                                        target=int(inp.name[inp.name.index('*d')+2:inp.name.index('*d')+5])
                                                        if node.inputs[target].default_value!=1:
                                                            row.enabled=False
                else:
                    count=1
                    for node in context.scene.world.node_tree.nodes:
                        for inp in node.inputs:
                            if '--' in inp.name:
                                count+=1
                                box.prop(context.scene.sub_sections,f'show_world_section_{count}',text=inp.name.replace('-',''),icon="DOWNARROW_HLT" if getattr(context.scene.sub_sections,f"show_world_section_{count}",False) else "RIGHTARROW_THIN",emboss=False)
                            else:
                                if getattr(context.scene.sub_sections,f"show_world_section_{count}"):
                                    box.prop(inp,'default_value',text=inp.name)
            else:
                box.label(text="World does not have any True-HDRI node!",icon='ERROR')
        
class TA_OT_Save_Assets_From_Current_File(bpy.types.Operator,Create_Assets_From_This_File_Props_Base):
    bl_idname = "ta.createassetsfromthisfile"
    bl_label = "Mark Assets"
    bl_description="Create Assets from this File"
    bl_options = {'REGISTER', 'UNDO'}
    
    def draw(self, context):
        layout=self.layout
        super().draw(layout)
    def execute(self, context):

        filters=[]
        if not self.show_filters:
            filters.append('OBJECTS')
            filters.append('MATS')
            filters.append('WORLDS')
            filters.append('MESH')
            filters.append('CURVE')
            filters.append('SURFACE')
            filters.append('CAMERA')
            filters.append('LIGHT')
            filters.append('FONT')
            filters.append('VOLUME')
            filters.append('GPENCIL')
            filters.append('META')
            filters.append('NODEGROUPS')
        else:
            if self.mark_objects:
                filters.append('OBJECTS')
            if self.mark_materials:
                filters.append('MATS')
            if self.mark_worlds:
                filters.append('WORLDS')
            if self.mark_objects:
                if self.mark_collections:
                    filters.append('COLLECTIONS')
                if self.mark_meshes:
                    filters.append('MESH')
                if self.mark_curves:
                    filters.append('CURVE')
                if self.mark_surfaces:
                    filters.append('SURFACE')
                if self.mark_cameras:
                    filters.append('CAMERA')
                if self.mark_lights:
                    filters.append('LIGHT')
                if self.mark_texts:
                    filters.append('FONT')
                if self.mark_volumes:
                    filters.append('VOLUME')
                if self.mark_gpencils:
                    filters.append('GPENCIL')
                if self.mark_metaballs:
                    filters.append('META')
            if self.mark_nodegroups:
                filters.append('NODEGROUPS')
        catalog=self.catalog
        if self.catalog=='NEW':
            catalog=self.new_catalog_name
        if self.pack:
            try:
                bpy.ops.file.pack_all()
            except Exception as e:
                print(e)
        if bpy.data.is_saved:
            create_assets(context,filters,force_previews=self.force_previews,min_nodes=self.min_nodes,clear_materials=self.clear_materials,quit=self.quit,shading=self.shading,engine=self.engine,override=self.override,angle=resolve_angle(self.camera_angle,self.flip_x,self.flip_y,self.flip_z),add_plane=preferences().add_ground_plane and not self.flip_z,what_to_save=self.what_to_save,catalog=catalog,origin_to_bottom=self.origin_to_bottom)
            self.report({'INFO'},'Assets created successfully!')
        return {'FINISHED'}
    def invoke(self,context,event):
        if not ensure_directory_is_set():
            self.report({'WARNING'},'Set directory path in the addon preferences!')
            return {'CANCELLED'}
        if not bpy.data.is_saved:
            bpy.ops.wm.save_mainfile('INVOKE_DEFAULT')
        load_default_settings(self,'default_settings_this_file')
        return context.window_manager.invoke_props_dialog(self)

    
class TA_OT_Assets_From_Directory(bpy.types.Operator,Create_Assets_From_Directory_Props_Base):
    bl_idname = "ta.createassetsfromdirectory"
    bl_label = "Create Assets From Directory"
    bl_description="Create Assets from Directory"
    bl_options = {'REGISTER', 'UNDO'}
    
    @classmethod
    def poll(cls,context):
        return context.scene.ta_progress==""
    def draw(self, context):
        layout=self.layout
        super().draw(layout)
        
    def modal(self, context,event):
        if event.type == 'TIMER':
            context.scene.ta_progress=f"Creating Assets..."
            if context.area:
                context.area.tag_redraw()
            if self.thread and not self.thread.is_alive():
                self.report({'INFO'},"Assets Creation Completed!")
                context.scene.ta_progress=""
                return {'FINISHED'}
        
        else:
            return {'PASS_THROUGH'}

        return {'RUNNING_MODAL'}
    def execute(self, context):
        
        dir_path=self.filepath
        filters=[]
        if not self.show_filters:
            filters.append('OBJECTS')
            filters.append('MATS')
            filters.append('WORLDS')
            filters.append('MESH')
            filters.append('CURVE')
            filters.append('SURFACE')
            filters.append('CAMERA')
            filters.append('LIGHT')
            filters.append('FONT')
            filters.append('VOLUME')
            filters.append('GPENCIL')
            filters.append('META')
            filters.append('NODEGROUPS')
        else:
            
            if self.mark_objects:
                filters.append('OBJECTS')
            if self.mark_materials:
                filters.append('MATS')
            if self.mark_worlds:
                filters.append('WORLDS')
            if self.mark_objects:
                if self.mark_collections:
                    filters.append('COLLECTIONS')
                if self.mark_meshes:
                    filters.append('MESH')
                if self.mark_curves:
                    filters.append('CURVE')
                if self.mark_surfaces:
                    filters.append('SURFACE')
                if self.mark_cameras:
                    filters.append('CAMERA')
                if self.mark_lights:
                    filters.append('LIGHT')
                if self.mark_texts:
                    filters.append('FONT')
                if self.mark_volumes:
                    filters.append('VOLUME')
                if self.mark_gpencils:
                    filters.append('GPENCIL')
                if self.mark_metaballs:
                    filters.append('META')
            if self.mark_nodegroups:
                    filters.append('NODEGROUPS')
        catalog=self.catalog if self.catalog!='NEW' else self.new_catalog_name
        objects_lib_path, mats_lib_path, worlds_lib_path, node_groups_lib_path,blends_lib_path,_ = ensure_all_libraries()
        files=[]
        dir_path=self.filepath
        if self.files:
            dir_path=os.path.dirname(self.filepath)
        if len(self.files)>0 and all([a.name and os.path.exists(os.path.join(dir_path,a.name)) for a in self.files]):
            files=[os.path.join(dir_path,a.name) for a in self.files]
        else:
            files=get_all_blend_files(dir_path)
        if self.add_lib_to_ab:
            ensure_library(os.path.basename(self.export_dir.strip('\\')),self.export_dir)
        
        if filters and  os.path.isdir(dir_path) and objects_lib_path and mats_lib_path and worlds_lib_path:
            
            if preferences().non_blocking:

                t1=threading.Thread(target=functools.partial(create_assets_from_path,files,filters,objects_lib_path,mats_lib_path,worlds_lib_path,node_groups_lib_path,blends_lib_path,self.override,self.min_nodes,shading=self.shading,engine=self.engine,max_time=self.max_time,force_previews=self.force_previews,clear_materials=self.clear_materials,angle=resolve_angle(self.camera_angle,self.flip_x,self.flip_y,self.flip_z),join_children=self.kitbash,clear_parent=self.clear_parent,apply_rotation=self.apply_rotation,catalog=catalog,add_plane=preferences().add_ground_plane and not self.flip_z,join_method=self.join_method,what_to_save=self.what_to_save,pack=self.pack,change_displacement_type=self.change_displacement_type,remove_subdiv=self.remove_subdiv,origin_to_bottom=self.origin_to_bottom,save_as_blend=self.save_as_blend,skip_hidden=self.skip_hidden))
                t1.start()
                self.thread=t1
            else:
                create_assets_from_path(files,filters,objects_lib_path,mats_lib_path,worlds_lib_path,node_groups_lib_path,blends_lib_path,self.override,self.min_nodes,shading=self.shading,engine=self.engine,max_time=self.max_time,force_previews=self.force_previews,clear_materials=self.clear_materials,angle=resolve_angle(self.camera_angle,self.flip_x,self.flip_y,self.flip_z),join_children=self.kitbash,clear_parent=self.clear_parent,apply_rotation=self.apply_rotation,catalog=catalog,add_plane=preferences().add_ground_plane and not self.flip_z,join_method=self.join_method,what_to_save=self.what_to_save,pack=self.pack,change_displacement_type=self.change_displacement_type,remove_subdiv=self.remove_subdiv,origin_to_bottom=self.origin_to_bottom,save_as_blend=self.save_as_blend,skip_hidden=self.skip_hidden)
        if preferences().non_blocking:
            wm=context.window_manager
            self.timer = wm.event_timer_add(0.2,window= context.window)
            wm.modal_handler_add(self) 
            return {'RUNNING_MODAL'}
        else:
            return {'FINISHED'}

    
    def invoke(self,context,event):
        self.thread=None
        if not ensure_directory_is_set():
            self.report({'WARNING'},'Set directory path in the addon preferences!')
            return {'CANCELLED'}
        self.filename=''
        context.window_manager.fileselect_add(self)
        # print(preferences().default_settings_from_directory.keys()[:])
        load_default_settings(self,'default_settings_from_directory')
        return {'RUNNING_MODAL'}


class TA_OT_USD_Assets_From_Directory(bpy.types.Operator,Create_Assets_From_USD_Props_Base):
    bl_idname = "ta.createusdassetsfromdirectory"
    bl_label = "Create Assets From USD Files"
    bl_description="Create Assets from Directory"
    bl_options = {'REGISTER', 'UNDO'}
    @classmethod
    def poll(cls,context):
        return context.scene.ta_progress==""
    def draw(self,context):
        super().draw(self.layout)
    def modal(self, context,event):
        if event.type == 'TIMER':
            context.scene.ta_progress=f"Creating Assets..."
            if context.area:
                context.area.tag_redraw()
            if self.thread and not self.thread.is_alive():
                self.report({'INFO'},"Assets Creation Completed!")
                context.scene.ta_progress=""
                return {'FINISHED'}
        else:
            return {'PASS_THROUGH'}
        return {'RUNNING_MODAL'}
    def execute(self, context):
        
        dir_path=self.filepath
        filters=[]
        filters.append('OBJECTS')
        filters.append('MESH')
        
        catalog=self.catalog if self.catalog!='NEW' else self.new_catalog_name
        objects_lib_path, mats_lib_path, worlds_lib_path, node_groups_lib_path,_,_ = ensure_all_libraries()
        files=[]
        dir_path=self.filepath
        if self.files:
            dir_path=os.path.dirname(self.filepath)
        if len(self.files)>0 and all([a.name and os.path.exists(os.path.join(dir_path,a.name)) for a in self.files]):
            files=[os.path.join(dir_path,a.name) for a in self.files]
        else:
            files=get_all_files_with_extension(dir_path,'usd')
            files.extend(get_all_files_with_extension(dir_path,'usdc'))
            files.extend(get_all_files_with_extension(dir_path,'usda'))
            files.extend(get_all_files_with_extension(dir_path,'usdz'))
        args_dict={'scale':self.scale,'set_frame_range':self.set_frame_range,'import_cameras':self.import_cameras,'import_curves':self.import_curves,'import_lights':self.import_lights,'import_materials':self.import_materials,'import_meshes':self.import_meshes,'import_volumes':self.import_volumes,'import_subdiv':self.import_subdiv,'import_instance_proxies':self.import_instance_proxies,'import_visible_only':self.import_visible_only,'create_collection':self.create_collection,'read_mesh_uvs':self.read_mesh_uvs,'read_mesh_colors':self.read_mesh_colors,'prim_path_mask':self.prim_path_mask,'import_guide':self.import_guide,'import_proxy':self.import_proxy,'import_render':self.import_render,'import_usd_preview':self.import_usd_preview,'set_material_blend':self.set_material_blend,'light_intensity_scale':self.light_intensity_scale,'mtl_name_collision_mode':self.mtl_name_collision_mode,'import_all_materials':self.import_all_materials,'import_textures_mode':self.import_textures_mode,'import_textures_dir':self.import_textures_dir,'tex_name_collision_mode':self.tex_name_collision_mode}
        if filters and  os.path.isdir(dir_path) and objects_lib_path and mats_lib_path and worlds_lib_path:
            
            if preferences().non_blocking:

                t1=threading.Thread(target=functools.partial(create_usd_assets_from_path,files,objects_lib_path,self.override,shading=self.shading,engine=self.engine,max_time=self.max_time,force_previews=self.force_previews,angle=resolve_angle(self.camera_angle,self.flip_x,self.flip_y,self.flip_z),catalog=catalog,add_plane=preferences().add_ground_plane and not self.flip_z,pack=self.pack,make_collection=self.make_collection=='Collection',**args_dict))
                t1.start()
                self.thread=t1
            else:
                create_usd_assets_from_path(files,objects_lib_path,self.override,shading=self.shading,engine=self.engine,max_time=self.max_time,force_previews=self.force_previews,angle=resolve_angle(self.camera_angle,self.flip_x,self.flip_y,self.flip_z),catalog=catalog,add_plane=preferences().add_ground_plane and not self.flip_z,pack=self.pack,make_collection=self.make_collection=='Collection',**args_dict)
        if preferences().non_blocking:
            wm=context.window_manager
            self.timer = wm.event_timer_add(0.2,window= context.window)
            wm.modal_handler_add(self) 
            return {'RUNNING_MODAL'}
        else:
            return {'FINISHED'}

    
    def invoke(self,context,event):
        self.thread=None
        if not ensure_directory_is_set():
            self.report({'WARNING'},'Set directory path in the addon preferences!')
            return {'CANCELLED'}
        context.window_manager.fileselect_add(self)
        load_default_settings(self,'default_settings_usd')
        return {'RUNNING_MODAL'}


class TA_OT_FBX_Assets_From_Directory(bpy.types.Operator,Create_Assets_FBX_Props_Base):
    bl_idname = "ta.createfbxassetsfromdirectory"
    bl_label = "Create Assets From FBX Files"
    bl_description="Create Assets from Directory"
    bl_options = {'REGISTER', 'UNDO'}
    @classmethod
    def poll(cls,context):
        return context.scene.ta_progress==""
    def draw(self,context):
        super().draw(self.layout)
    def modal(self, context,event):
        if event.type == 'TIMER':
            context.scene.ta_progress=f"Creating Assets..."
            if context.area:
                context.area.tag_redraw()
            if self.thread and not self.thread.is_alive():
                self.report({'INFO'},"Assets Creation Completed!")
                context.scene.ta_progress=""
                return {'FINISHED'}
        else:
            return {'PASS_THROUGH'}
        return {'RUNNING_MODAL'}
    def execute(self, context):
        
        dir_path=self.filepath
        filters=[]
        filters.append('OBJECTS')
        filters.append('MESH')
        
        catalog=self.catalog if self.catalog!='NEW' else self.new_catalog_name
        objects_lib_path, mats_lib_path, worlds_lib_path, node_groups_lib_path,_,_ = ensure_all_libraries()
        files=[]
        dir_path=self.filepath
        if self.files:
            dir_path=os.path.dirname(self.filepath)
        if len(self.files)>0 and all([a.name and os.path.exists(os.path.join(dir_path,a.name)) for a in self.files]):
            files=[os.path.join(dir_path,a.name) for a in self.files]
        else:
            files=get_all_files_with_extension(dir_path,'fbx')
        if filters and  os.path.isdir(dir_path) and objects_lib_path and mats_lib_path and worlds_lib_path:
            
            if preferences().non_blocking:

                t1=threading.Thread(target=functools.partial(create_fbx_assets_from_path,files,objects_lib_path,self.override,shading=self.shading,engine=self.engine,max_time=self.max_time,force_previews=self.force_previews,angle=resolve_angle(self.camera_angle,self.flip_x,self.flip_y,self.flip_z),catalog=catalog,add_plane=preferences().add_ground_plane and not self.flip_z,use_auto_bone_orientation= self.use_auto_bone_orientation, my_calculate_roll= self.my_calculate_roll, my_bone_length= self.my_bone_length, my_leaf_bone= self.my_leaf_bone, use_fix_bone_poses= self.use_fix_bone_poses, use_fix_attributes= self.use_fix_attributes, use_only_deform_bones= self.use_only_deform_bones, use_vertex_animation= self.use_vertex_animation, use_animation= self.use_animation, my_animation_offset= self.my_animation_offset, use_animation_prefix= self.use_animation_prefix, use_triangulate= self.use_triangulate, my_import_normal= self.my_import_normal, use_auto_smooth= self.use_auto_smooth, my_angle= self.my_angle, my_shade_mode= self.my_shade_mode, my_scale= self.my_scale, use_optimize_for_blender= self.use_optimize_for_blender, use_reset_mesh_origin= self.use_reset_mesh_origin, use_edge_crease= self.use_edge_crease, my_edge_crease_scale= self.my_edge_crease_scale, my_edge_smoothing= self.my_edge_smoothing, use_import_materials= self.use_import_materials, use_rename_by_filename= self.use_rename_by_filename, my_rotation_mode= self.my_rotation_mode, use_manual_orientation= self.use_manual_orientation, global_scale= self.global_scale, bake_space_transform= self.bake_space_transform, use_custom_normals= self.use_custom_normals, use_image_search= self.use_image_search, use_alpha_decals= self.use_alpha_decals, decal_offset= self.decal_offset, use_anim= self.use_anim, anim_offset= self.anim_offset, use_subsurf= self.use_subsurf, use_custom_props= self.use_custom_props, use_custom_props_enum_as_string= self.use_custom_props_enum_as_string, ignore_leaf_bones= self.ignore_leaf_bones, force_connect_children= self.force_connect_children, automatic_bone_orientation= self.automatic_bone_orientation, primary_bone_axis= self.primary_bone_axis, secondary_bone_axis= self.secondary_bone_axis, use_prepost_rot= self.use_prepost_rot, axis_forward= self.axis_forward, axis_up= self.axis_up,importer=preferences().fbx_importer,pack=self.pack,single_file=self.single_file,make_collection=self.make_collection=='Collection'))
                t1.start()
                self.thread=t1
            else:
                create_fbx_assets_from_path(files,objects_lib_path,self.override,shading=self.shading,engine=self.engine,max_time=self.max_time,force_previews=self.force_previews,angle=resolve_angle(self.camera_angle,self.flip_x,self.flip_y,self.flip_z),catalog=catalog,add_plane=preferences().add_ground_plane and not self.flip_z,use_auto_bone_orientation= self.use_auto_bone_orientation, my_calculate_roll= self.my_calculate_roll, my_bone_length= self.my_bone_length, my_leaf_bone= self.my_leaf_bone, use_fix_bone_poses= self.use_fix_bone_poses, use_fix_attributes= self.use_fix_attributes, use_only_deform_bones= self.use_only_deform_bones, use_vertex_animation= self.use_vertex_animation, use_animation= self.use_animation, my_animation_offset= self.my_animation_offset, use_animation_prefix= self.use_animation_prefix, use_triangulate= self.use_triangulate, my_import_normal= self.my_import_normal, use_auto_smooth= self.use_auto_smooth, my_angle= self.my_angle, my_shade_mode= self.my_shade_mode, my_scale= self.my_scale, use_optimize_for_blender= self.use_optimize_for_blender, use_reset_mesh_origin= self.use_reset_mesh_origin, use_edge_crease= self.use_edge_crease, my_edge_crease_scale= self.my_edge_crease_scale, my_edge_smoothing= self.my_edge_smoothing, use_import_materials= self.use_import_materials, use_rename_by_filename= self.use_rename_by_filename, my_rotation_mode= self.my_rotation_mode, use_manual_orientation= self.use_manual_orientation, global_scale= self.global_scale, bake_space_transform= self.bake_space_transform, use_custom_normals= self.use_custom_normals, use_image_search= self.use_image_search, use_alpha_decals= self.use_alpha_decals, decal_offset= self.decal_offset, use_anim= self.use_anim, anim_offset= self.anim_offset, use_subsurf= self.use_subsurf, use_custom_props= self.use_custom_props, use_custom_props_enum_as_string= self.use_custom_props_enum_as_string, ignore_leaf_bones= self.ignore_leaf_bones, force_connect_children= self.force_connect_children, automatic_bone_orientation= self.automatic_bone_orientation, primary_bone_axis= self.primary_bone_axis, secondary_bone_axis= self.secondary_bone_axis, use_prepost_rot= self.use_prepost_rot, axis_forward= self.axis_forward, axis_up= self.axis_up,importer=preferences().fbx_importer,pack=self.pack,single_file=self.single_file,make_collection=self.make_collection=='Collection')
        if preferences().non_blocking:
            wm=context.window_manager
            self.timer = wm.event_timer_add(0.2,window= context.window)
            wm.modal_handler_add(self) 
            return {'RUNNING_MODAL'}
        else:
            return {'FINISHED'}

    
    def invoke(self,context,event):
        #print(self.as_keywords())
        # temp={ 'use_auto_bone_orientation': True, 'my_calculate_roll': 'None', 'my_bone_length': 10.0, 'my_leaf_bone': 'Long', 'use_fix_bone_poses': False, 'use_fix_attributes': True, 'use_only_deform_bones': False, 'use_vertex_animation': True, 'use_animation': True, 'my_animation_offset': 0, 'use_animation_prefix': False, 'use_triangulate': False, 'my_import_normal': 'Import', 'use_auto_smooth': True, 'my_angle': 60.0, 'my_shade_mode': 'Smooth', 'my_scale': 0.009999999776482582, 'use_optimize_for_blender': False, 'use_reset_mesh_origin': True, 'use_edge_crease': True, 'my_edge_crease_scale': 1.0, 'my_edge_smoothing': 'FBXSDK', 'use_import_materials': True, 'use_rename_by_filename': False, 'my_rotation_mode': 'QUATERNION', 'use_manual_orientation': False, 'global_scale': 1.0, 'bake_space_transform': False, 'use_custom_normals': True, 'use_image_search': True, 'use_alpha_decals': False, 'decal_offset': 0.0, 'use_anim': True, 'anim_offset': 1.0, 'use_subsurf': False, 'use_custom_props': True, 'use_custom_props_enum_as_string': True, 'ignore_leaf_bones': False, 'force_connect_children': False, 'automatic_bone_orientation': False, 'primary_bone_axis': 'Y', 'secondary_bone_axis': 'X', 'use_prepost_rot': True, 'axis_forward': '-Z', 'axis_up': 'Y'}
        # i=10
        # for a,b in temp.items():
        #     i+=1
        #     print(f"{a}={a.upper()}")
        #     # if type(b)==str:
        #     #     print(f"{a.upper()}=argv[{i}]")
        #     # else:
        #     #     print(f"{a.upper()}=eval(argv[{i}])")
        #     temp[a]=None
        #print(temp)
        self.thread=None
        if not ensure_directory_is_set():
            self.report({'WARNING'},'Set directory path in the addon preferences!')
            return {'CANCELLED'}
        context.window_manager.fileselect_add(self)
        load_default_settings(self,'default_settings_fbx')
        return {'RUNNING_MODAL'}


class TA_OT_OBJ_Assets_From_Directory(bpy.types.Operator,Create_Assets_From_OBJ_Props_Base):
    bl_idname = "ta.createobjassetsfromdirectory"
    bl_label = "Create Assets From OBJ Files"
    bl_description="Create Assets from Directory"
    bl_options = {'REGISTER', 'UNDO'}
    @classmethod
    def poll(cls,context):
        return context.scene.ta_progress==""
    def draw(self,context):
        super().draw(self.layout)
    def modal(self, context,event):
        if event.type == 'TIMER':
            context.scene.ta_progress=f"Creating Assets..."
            if context.area:
                context.area.tag_redraw()
            if self.thread and not self.thread.is_alive():
                self.report({'INFO'},"Assets Creation Completed!")
                context.scene.ta_progress=""
                return {'FINISHED'}
        else:
            return {'PASS_THROUGH'}
        return {'RUNNING_MODAL'}
    def execute(self, context):
        
        dir_path=self.filepath
        filters=[]
        filters.append('OBJECTS')
        filters.append('MESH')
        
        catalog=self.catalog if self.catalog!='NEW' else self.new_catalog_name
        objects_lib_path, mats_lib_path, worlds_lib_path, node_groups_lib_path,_,_ = ensure_all_libraries()
        files=[]
        dir_path=self.filepath
        if self.files:
            dir_path=os.path.dirname(self.filepath)
        if len(self.files)>0 and all([a.name and os.path.exists(os.path.join(dir_path,a.name)) for a in self.files]):
            files=[os.path.join(dir_path,a.name) for a in self.files]
        else:
            files=get_all_files_with_extension(dir_path,'obj')
        
        if filters and  os.path.isdir(dir_path) and objects_lib_path and mats_lib_path and worlds_lib_path:
            
            if preferences().non_blocking:

                t1=threading.Thread(target=functools.partial(create_obj_assets_from_path,files,objects_lib_path,self.override,shading=self.shading,engine=self.engine,max_time=self.max_time,force_previews=self.force_previews,angle=resolve_angle(self.camera_angle,self.flip_x,self.flip_y,self.flip_z),catalog=catalog,add_plane=preferences().add_ground_plane and not self.flip_z,use_auto_bone_orientation= self.use_auto_bone_orientation, my_calculate_roll= self.my_calculate_roll, my_bone_length= self.my_bone_length, my_leaf_bone= self.my_leaf_bone, use_fix_bone_poses= self.use_fix_bone_poses, use_fix_attributes= self.use_fix_attributes, use_only_deform_bones= self.use_only_deform_bones, use_vertex_animation= self.use_vertex_animation, use_animation= self.use_animation, my_animation_offset= self.my_animation_offset, use_animation_prefix= self.use_animation_prefix, use_triangulate= self.use_triangulate, my_import_normal= self.my_import_normal, use_auto_smooth= self.use_auto_smooth, my_angle= self.my_angle, my_shade_mode= self.my_shade_mode, my_scale= self.my_scale, use_optimize_for_blender= self.use_optimize_for_blender, use_reset_mesh_origin= self.use_reset_mesh_origin, use_edge_crease= self.use_edge_crease, my_edge_crease_scale= self.my_edge_crease_scale, my_edge_smoothing= self.my_edge_smoothing, use_import_materials= self.use_import_materials, use_rename_by_filename= self.use_rename_by_filename, my_rotation_mode= self.my_rotation_mode,use_edges=self.use_edges,use_smooth_groups=self.use_smooth_groups,use_split_objects=self.use_split_objects,use_split_groups=self.use_split_groups,use_groups_as_vgroups=self.use_groups_as_vgroups,use_image_search=self.use_image_search,split_mode=self.split_mode,global_clamp_size=self.global_clamp_size,clamp_size=self.clamp_size,global_scale=self.global_scale,import_vertex_groups=self.import_vertex_groups,validate_meshes=self.validate_meshes,axis_forward=self.axis_forward,axis_up=self.axis_up,importer=preferences().fbx_importer,pack=self.pack,single_file=self.single_file,make_collection=self.make_collection=='Collection'))
                t1.start()
                self.thread=t1
            else:
                create_obj_assets_from_path(files,objects_lib_path,self.override,shading=self.shading,engine=self.engine,max_time=self.max_time,force_previews=self.force_previews,angle=resolve_angle(self.camera_angle,self.flip_x,self.flip_y,self.flip_z),catalog=catalog,add_plane=preferences().add_ground_plane and not self.flip_z,use_auto_bone_orientation= self.use_auto_bone_orientation, my_calculate_roll= self.my_calculate_roll, my_bone_length= self.my_bone_length, my_leaf_bone= self.my_leaf_bone, use_fix_bone_poses= self.use_fix_bone_poses, use_fix_attributes= self.use_fix_attributes, use_only_deform_bones= self.use_only_deform_bones, use_vertex_animation= self.use_vertex_animation, use_animation= self.use_animation, my_animation_offset= self.my_animation_offset, use_animation_prefix= self.use_animation_prefix, use_triangulate= self.use_triangulate, my_import_normal= self.my_import_normal, use_auto_smooth= self.use_auto_smooth, my_angle= self.my_angle, my_shade_mode= self.my_shade_mode, my_scale= self.my_scale, use_optimize_for_blender= self.use_optimize_for_blender, use_reset_mesh_origin= self.use_reset_mesh_origin, use_edge_crease= self.use_edge_crease, my_edge_crease_scale= self.my_edge_crease_scale, my_edge_smoothing= self.my_edge_smoothing, use_import_materials= self.use_import_materials, use_rename_by_filename= self.use_rename_by_filename, my_rotation_mode= self.my_rotation_mode,use_edges=self.use_edges,use_smooth_groups=self.use_smooth_groups,use_split_objects=self.use_split_objects,use_split_groups=self.use_split_groups,use_groups_as_vgroups=self.use_groups_as_vgroups,use_image_search=self.use_image_search,split_mode=self.split_mode,global_clamp_size=self.global_clamp_size,clamp_size=self.clamp_size,global_scale=self.global_scale,import_vertex_groups=self.import_vertex_groups,validate_meshes=self.validate_meshes,axis_forward=self.axis_forward,axis_up=self.axis_up,importer=preferences().fbx_importer,pack=self.pack,single_file=self.single_file,make_collection=self.make_collection=='Collection')
        if preferences().non_blocking:
            wm=context.window_manager
            self.timer = wm.event_timer_add(0.2,window= context.window)
            wm.modal_handler_add(self) 
            return {'RUNNING_MODAL'}
        else:
            return {'FINISHED'}

    
    def invoke(self,context,event):
        # print(self.as_keywords())
        # temp={ 'use_edges': True, 'use_smooth_groups': True, 'use_split_objects': True, 'use_split_groups': False, 'use_groups_as_vgroups': False, 'use_image_search': True, 'split_mode': 'ON', 'global_clamp_size': 0.0}
        # i=35
        # for a,b in temp.items():
        #     i+=1
        #     print(f"cmd.append(str({a}))")
        #     #print(f"{a}=self.{a}")
        #     # if type(b)==str:
        #     #     print(f"{a.upper()}=argv[{i}]")
        #     # else:
        #     #     print(f"{a.upper()}=eval(argv[{i}])")
        #     temp[a]=None
        # print(temp)
        self.thread=None
        if not ensure_directory_is_set():
            self.report({'WARNING'},'Set directory path in the addon preferences!')
            return {'CANCELLED'}
        context.window_manager.fileselect_add(self)
        load_default_settings(self,'default_settings_obj')
        return {'RUNNING_MODAL'}
import inspect

class TA_OT_Create_Node_Group_Asset(bpy.types.Operator):
    bl_idname = "ta.marknodegroupasset"
    bl_label = "Mark Asset (True Assets)"
    bl_description="Mark Asset"
    bl_options = {'REGISTER', 'UNDO'}
    name:bpy.props.StringProperty(default="",name="Name",update=recheck_if_node_group_file_exists)
    description:bpy.props.StringProperty(default="",name="Description")
    tags:bpy.props.StringProperty(default="",name="Tags")
    engine:bpy.props.EnumProperty(name="Render Engine",default=0,items=(("CYCLES","Cycles","Cycles"),("BLENDER_EEVEE","Eevee","Eevee")))
    override:bpy.props.EnumProperty(name="If File Already Exists",items=(('OVERRIDE','Override','Override'),('COPY','Create New','Create New')))
    exists:bpy.props.BoolProperty(default=False,options={'SKIP_SAVE'})
    @classmethod
    def poll(cls, context):
        return context.area.type =='NODE_EDITOR' and context.active_node and context.active_node.type=='GROUP'
    def draw(self, context):
        layout=self.layout
        layout.prop(self,'name')
        layout.prop(self,'description')
        layout.prop(self,'tags')
        if self.type!='GEOMETRY':
            layout.prop(self,'engine',expand=True)
        if self.exists:
            layout.label(text="File with same name already exists!")
            layout.prop(self,'override',expand=True)
    def execute(self, context):
        ensure_all_libraries()
        #ensure_library("True-Assets (Node Groups)",preferences().node_groups_directory)
        if not os.path.isdir(os.path.join(preferences().node_groups_directory,"Thumbnails")):
            os.mkdir(os.path.join(preferences().node_groups_directory,"Thumbnails"))
        node_group=context.active_node.node_tree
        if node_group.type=='GEOMETRY':
            node_group.name=self.name
            node_group.asset_mark()
            for t in self.tags.split(","):
                node_group.asset_data.tags.new(t)    
            node_group.asset_data.description=self.description
            node_group.asset_data.tags.new(node_group.name)
            with context.temp_override(id=node_group):

                bpy.ops.ed.lib_id_load_custom_preview( filepath=os.path.join(os.path.dirname(
                        os.path.abspath(__file__)), "Assets","GeometryNodeGroup.png"))
        else:
            Create_Node_Group_Asset(context,context.active_node,node_group,self.engine,tags=self.tags,description=self.description)
        write_to_blend("NODEGROUPS",[node_group,],"COPY",self.name)
        return {'FINISHED'}
    def invoke(self,context,event):
        self.type='SHADER'
        if not ensure_directory_is_set():
            self.report({'WARNING'},'Set directory path in the addon preferences!')
            return {'CANCELLED'}
        if context.active_node.node_tree:
            self.name=context.active_node.node_tree.name
        self.tags=self.name
        
        # for a in context.active_node.inputs:
        #     if "--" not in a.name:
        #         print(f"{a.name.replace(' ','_').replace('-','_').lower()}:bpy.props.FloatProperty(name='{a.name}',default={a.default_value})")
        if context.active_node.node_tree.type=='GEOMETRY':
            self.type='GEOMETRY'
        return context.window_manager.invoke_props_dialog(self)
def create_dict(list1,list2):
    result={}
    if len(list1)==len(list2):
        for i,a in enumerate(list1):
            result[a]=list2[i]
    return result
class TA_OT_Materials_From_Directory(bpy.types.Operator,Create_Materials_Props_Base):
    bl_idname = "ta.creatematerialsfromdirectory"
    bl_label = "Create Materials"
    bl_description="Create Materials from Directory"
    bl_options = {'REGISTER', 'UNDO'}
    filepath: bpy.props.StringProperty(subtype="DIR_PATH")
    files : bpy.props.CollectionProperty(
            name="File Path",
            type=bpy.types.OperatorFileListElement,
            )

    directory:bpy.props.StringProperty(name="Directory")
    filter_glob: bpy.props.StringProperty(default="*" + ";*".join(['jpg','jpeg','png']),options={"HIDDEN"})
    filter_search:bpy.props.StringProperty(default="",name="Search Filter",description="File names must include this")
    @classmethod
    def poll(cls,context):
        return context.scene.ta_progress==""
    def draw(self,context):
        super().draw(self.layout)
    def modal(self, context,event):
        if event.type == 'TIMER':
            context.scene.ta_progress=f"Creating Materials..."
            if context.area:
                context.area.tag_redraw()
            if self.thread and not self.thread.is_alive():
                self.report({'INFO'},"Material Creation Completed!")
                context.scene.ta_progress=""
                return {'FINISHED'}
        else:
            return {'PASS_THROUGH'}
        return {'RUNNING_MODAL'}
    def execute(self, context):
        ensure_all_libraries()
        self.mark_asset=True
        files=None
        if len([a.name for a in self.files if a.name])>0:
            files=[a.name for a in self.files]
        dir_path=self.directory
        # if self.files:
        #     dir_path=os.path.dirname(self.filepath)
        catalog=self.catalog
        if self.catalog=="NEW":
            catalog=create_new_catalog(self.dir_path,self.new_catalog_name)
        else:
            catalog=self.catalog
        if files:
            files=[a for a in files if os.path.isfile(os.path.join(dir_path,a))]
        # print("Directory:\n",self.dir_path)
        
        if self.shader_type=='True-Uber':
            shader_values=create_dict([self.rna_type.properties[a].name for a in self.uber_settings+self.surface_settings+self.mapping_settings],[getattr(self,a) if not is_sequence(getattr(self,a)) else getattr(self,a)[:] for a in self.uber_settings+self.surface_settings+self.mapping_settings])
        elif self.shader_type=='True-BSDF':
            shader_values=create_dict([self.rna_type.properties[a].name for a in self.bsdf_settings+self.surface_settings+self.bsdf_mapping],[getattr(self,a) if not is_sequence(getattr(self,a)) else getattr(self,a)[:] for a in self.bsdf_settings+self.surface_settings+self.bsdf_mapping])
        elif self.shader_type=='True-Organic':
            shader_values=create_dict([self.rna_type.properties[a].name for a in self.organic_settings+self.surface_settings+self.mapping_settings],[getattr(self,a) if not is_sequence(getattr(self,a)) else getattr(self,a)[:] for a in self.organic_settings+self.surface_settings+self.mapping_settings])
        else:
            shader_values=create_dict([self.rna_type.properties[a].name for a in self.plastic_settings+self.surface_settings+self.mapping_settings],[getattr(self,a) if not is_sequence(getattr(self,a)) else getattr(self,a)[:] for a in self.plastic_settings+self.surface_settings+self.mapping_settings])
        
        if os.path.isdir(dir_path):
            if preferences().non_blocking and self.mark_asset:
                t=threading.Thread(target=functools.partial(create_materials_in_separate_thread,self.shader_type,dir_path,files,
                        catalog=catalog,pack=self.pack,search=self.filter_search,shader_values=shader_values))
                t.start()
                self.thread=t
                wm=context.window_manager
                self.timer = wm.event_timer_add(1,window= context.window)
                wm.modal_handler_add(self) 
                return {'RUNNING_MODAL'}
            else:
                setup_materials(self.shader_type,dir_path,files,
                        mark_asset=self.mark_asset,only_basecolor=self.only_basecolor,catalog=catalog,pack=self.pack,search=self.filter_search,shader_values=shader_values)
        
                return {'FINISHED'}
    def invoke(self,context,event):
        
        if not ensure_directory_is_set():
            self.report({'WARNING'},'Set directory path in the addon preferences!')
            return {'CANCELLED'}
        self.dir_path=preferences().materials_directory
        self.use_image_filter=True
        context.window_manager.fileselect_add(self)
        get_catalogs_to_move_to_for_mats(self,context)
        load_default_settings(self,'default_settings_materials')
        return {'RUNNING_MODAL'}

class TA_OT_WORLDS_From_Directory(bpy.types.Operator,Create_Worlds_Props_Base):
    bl_idname = "ta.createworldsfromdirectory"
    bl_label = "Create Worlds"
    bl_description="Create Worlds from Directory"
    bl_options = {'REGISTER', 'UNDO'}
    filepath: bpy.props.StringProperty(subtype="DIR_PATH")
    files : bpy.props.CollectionProperty(
            name="File Path",
            type=bpy.types.OperatorFileListElement,
            )
    directory:bpy.props.StringProperty(name="Directory")
    filter_glob: bpy.props.StringProperty(default="*" + ";*".join(['jpg','jpeg','png','hdr','exr']),options={"HIDDEN"})
    filter_search:bpy.props.StringProperty(default="",name="Search Filter",description="File names must include this")
    
    @classmethod
    def poll(cls,context):
        return context.scene.ta_progress==""
    def draw(self,context):
        super().draw(self.layout)
    def modal(self, context,event):
        if event.type == 'TIMER':
            context.scene.ta_progress=f"Creating Worlds..."
            if context.area:
                context.area.tag_redraw()
            if self.thread and not self.thread.is_alive():
                self.report({'INFO'},"World Creation Completed!")
                context.scene.ta_progress=""
                return {'FINISHED'}
        else:
            return {'PASS_THROUGH'}
        return {'RUNNING_MODAL'}
    def execute(self, context):
        self.mark_asset=True
        ensure_all_libraries()
        files=None
        if len([a.name for a in self.files if a.name])>0:
            files=[a.name for a in self.files]
        dir_path=self.directory
        image_filters=[]
        if self.use_exr:
            image_filters.append("exr")
        if self.use_hdr:
            image_filters.append("hdr")
        if self.use_jpeg:
            image_filters.append("jpeg")
        if self.use_jpg:
            image_filters.append("jpg")
        if self.use_png:
            image_filters.append("png")
        catalog=self.catalog
        if self.catalog=="NEW":
            catalog=create_new_catalog(self.dir_path,self.new_catalog_name)
        else:
            catalog=self.catalog
        if files:
            files=[a for a in files if os.path.isfile(os.path.join(dir_path,a))]
        if self.shader_type=='True-HDRI':
            shader_values=create_dict([self.rna_type.properties[a].name for a in self.settings],[getattr(self,a) if not is_sequence(getattr(self,a)) else getattr(self,a)[:] for a in self.settings])    
        else:
            shader_values=create_dict([self.rna_type.properties[a].name for a in self.standard_settings],[getattr(self,a) if not is_sequence(getattr(self,a)) else getattr(self,a)[:] for a in self.standard_settings])    
        if os.path.isdir(dir_path):
            if preferences().non_blocking and self.mark_asset:
                t=threading.Thread(target=functools.partial(create_worlds_in_separate_thread,dir_path,files,image_filters=image_filters,subfolders=self.search_subfolders,catalog=catalog,pack=self.pack,search=self.filter_search,shader_values=shader_values,shader_type=self.shader_type))
                t.start()
                self.thread=t
                wm=context.window_manager
                self.timer = wm.event_timer_add(1,window= context.window)
                wm.modal_handler_add(self) 
                return {'RUNNING_MODAL'}
            else:
                setup_hdris(context,dir_path,files,
                        mark_asset=self.mark_asset,image_filters=image_filters,subfolders=self.search_subfolders,catalog=catalog,pack=self.pack,search=self.filter_search,shader_values=shader_values,shader_type=self.shader_type)
                return {'FINISHED'}
        
    def invoke(self,context,event):
        # inputs=[]
        # for a in context.scene.world.node_tree.nodes:
        #     if "True" in a.name:
        #         for inp in a.inputs:
        #             if '--' not in inp.name:
        #                 inputs.append(inp.name.replace(' ','_').replace('-','_').lower())
        #                 print(f"{inp.name.replace(' ','_').replace('-','_').lower()}:bpy.props.FloatProperty(name='{inp.name}',default_value={inp.default_value},min=0,max=1)")
        
        if not ensure_directory_is_set():
            self.report({'WARNING'},'Set directory path in the addon preferences!')
            return {'CANCELLED'}
        self.dir_path=preferences().worlds_directory
        context.window_manager.fileselect_add(self)
        get_catalogs_to_move_to_for_worlds(self,context)
        load_default_settings(self,'default_settings_worlds')
        return {'RUNNING_MODAL'}
# class TA_OT_Assets_From_Directory(bpy.types.Operator):
#     bl_idname = "ta.createassetsfromdirectory"
#     bl_label = "Create Assets From Directory"
#     bl_description="Create Assets from Directory"
#     bl_options = {'REGISTER', 'UNDO'}
#     filepath: bpy.props.StringProperty(subtype="DIR_PATH",options={'SKIP_SAVE','HIDDEN'})
#     def modal(self, context,event):
#         if event.type == 'TIMER' and not self.completed:
#             if self.paths:
#                 path=self.paths.pop(0)
#                 create_assets_from_path(path,self.objects_lib_path,self.mats_lib_path,self.worlds_lib_path)
#             return {'RUNNING_MODAL'}
#         if not self.paths:
#             print("Done!")
#             return {'FINISHED'}
#         return {'RUNNING_MODAL'}
#     def execute(self, context):
#         dir_path=self.filepath
#         self.objects_lib_path=ensure_library("True-Assets (Objects)",path=preferences().objects_directory)
#         self.mats_lib_path=ensure_library("True-Assets (Materials)",path=preferences().materials_directory)
#         self.worlds_lib_path=ensure_library("True-Assets (Worlds)",path=preferences().worlds_directory)
#         self.paths=get_all_blend_files(dir_path)
#         wm=context.window_manager
#         self.timer = wm.event_timer_add(1,window= context.window)
#         wm.modal_handler_add(self)
#         self.completed = False
#         #for path in get_all_blend_files(dir_path):
#             #create_assets_from_path(path,objects_lib_path,mats_lib_path,worlds_lib_path)
        
#         return {'RUNNING_MODAL'}
#     def invoke(self,context,event):
#         context.window_manager.fileselect_add(self)
#         return {'RUNNING_MODAL'}
class Asset_Names(bpy.types.PropertyGroup):
    filepath: bpy.props.StringProperty()
    og_name:bpy.props.StringProperty(default="")
    name:bpy.props.StringProperty(default="",name="Name")
def append_object(path,data_type,name,instance:bool=True):
    if data_type=='COLLECTION':
        init_col=bpy.data.collections[:]
        bpy.ops.wm.append(directory=os.path.join(path,data_type),filename=name,instance_collections=instance)
        col=[a for a in bpy.data.collections if a not in init_col]
        if col:
            return col[0]
    else:
        init_objs=bpy.data.objects[:]
        bpy.ops.wm.append(directory=os.path.join(path,data_type.capitalize()),filename=name)
        obj=[a for a in bpy.data.objects if a not in init_objs]
        if obj:
            return obj[0]
class TA_OT_Replace_In_Scene(bpy.types.Operator):
    bl_idname = "ta.replaceinscene"
    bl_label = "Quick Swap"
    bl_description="Replace Active object with this asset"
    bl_options = {'REGISTER', 'UNDO'}
    instance_collections:bpy.props.BoolProperty(default=True,name='Instance')
    def draw(self, context):
        layout=self.layout
        layout.prop(self,'instance_collections')
    @classmethod
    def poll(cls,context):
        return context.area.ui_type=='ASSETS' and len(context.selected_assets[:])==1 and context.selected_assets[0].id_type in ['OBJECT','COLLECTION'] and context.space_data.params.asset_library_reference!='LOCAL' and (context.active_object or context.selected_objects)
    
    def execute(self, context):
        for a in context.selected_assets:
            path=(a.full_library_path,a.id_type,a.name)
        active=context.active_object
        if not active:
            active=context.selected_objects[0]
        new_data=append_object(path[0],path[1],path[2],self.instance_collections)
        for a in new_data.users_collection:
            a.objects.unlink(new_data)
        for a in active.users_collection:
            a.objects.link(new_data)
        for a in ['location','rotation_euler','scale']:
            setattr(new_data,a,getattr(active,a))
        if True:
            new_data.dimensions=new_data.dimensions*max(active.dimensions)/max(new_data.dimensions)
        bpy.data.objects.remove(active)
        return {'FINISHED'}

    
class TA_OT_Rename_Assets(bpy.types.Operator):
    bl_idname = "ta.renameassets"
    bl_label = "Rename Assets"
    bl_description="Rename Selected Assets"
    bl_options = {'REGISTER', 'UNDO'}
    names:bpy.props.CollectionProperty(type=Asset_Names)
    dir_path:bpy.props.StringProperty(default="")
    prefix:bpy.props.StringProperty(default="",name="Prefix",options={'SKIP_SAVE'})
    suffix:bpy.props.StringProperty(default="",name="Suffix",options={'SKIP_SAVE'})
    def draw(self, context):
        layout=self.layout
        layout.ui_units_x=18
        layout.prop(self,'prefix')
        layout.prop(self,'suffix')
        row=layout.row(align=True)
        row.label(text="Old Names")
        row.label(text="New Names")
        for a in self.names:
            row=layout.row()
            row.label(text=a.og_name+":")
            row.prop(a,'name',text="")
        
    @classmethod
    def poll(cls,context):
        return context.area.ui_type=='ASSETS' and context.selected_assets and context.selected_assets[:] and context.space_data.params.asset_library_reference!='LOCAL'
    def modal(self, context,event):
        if event.type == 'TIMER':
            completed=len([a for a in self.threads if not a.is_alive()])
            context.scene.ta_progress=f"Renaming Assets...({completed}/{len(self.threads)} files completed)"
            if context.area:
                context.area.tag_redraw()
            if all([not a.is_alive() for a in self.threads]):
                self.report({'INFO'},"Renaming Complete!")
                bpy.ops.asset.library_refresh()
                context.scene.ta_progress=""
                return {'FINISHED'}
        else:
            return {'PASS_THROUGH'}
        return {'RUNNING_MODAL'}
    def execute(self, context):
        blends={}
        for a in context.selected_assets:
            blends[a.full_library_path]=blends[a.full_library_path]+[(a.name,a.id_type),] if blends.get(a.full_library_path) else [(a.name,a.id_type),]
        
        self.threads=[]
        for b,f in blends.items():
            self.threads.append(rename_assets(b,f,[a.name for a in self.names if a.filepath in [os.path.join(b,n) for n,c in f]],self.prefix,self.suffix))
        if preferences().non_blocking:
            wm=context.window_manager
            self.timer = wm.event_timer_add(0.2,window= context.window)
            wm.modal_handler_add(self) 
            return {'RUNNING_MODAL'}
        else:
            bpy.ops.asset.library_refresh() 
            return {'FINISHED'}
    def invoke(self,context,event):
        
        for a in context.selected_assets:
            active_lib_name=context.area.spaces.active.params.asset_library_reference
            active_lib_path=context.preferences.filepaths.asset_libraries.get(active_lib_name).path
            # self.dir_path=os.path.dirname(a.full_library_path)
            self.dir_path=active_lib_path
            break
        self.names.clear()
        for a in context.selected_assets:
            t=self.names.add()
            t.og_name=a.name
            t.filepath=os.path.join(a.full_library_path,a.name)
        return context.window_manager.invoke_props_dialog(self)
class TA_OT_Update_Asset_Data(bpy.types.Operator):
    bl_idname = "ta.updateassetdata"
    bl_label = "Update Asset Data"
    bl_description="Update Asset Data(Tags,Author and Description)"
    bl_options = {'REGISTER', 'UNDO'}
    tags:bpy.props.StringProperty(default="",name="Tags")
    author:bpy.props.StringProperty(default="",name="Author")
    description:bpy.props.StringProperty(default="",name="Description")
    add_or_replace:bpy.props.EnumProperty(items=(("Replace","Replace","Replace"),("Add","Add","Add")),name="Add Or Replace")
    dir_path:bpy.props.StringProperty(default="")
    def draw(self, context):
        layout=self.layout
        layout.ui_units_x=18
        layout.prop(self,'tags')
        layout.prop(self,'add_or_replace',expand=True)
        layout.prop(self,'author')
        layout.prop(self,'description')
    @classmethod
    def poll(cls,context):
        return context.area.ui_type=='ASSETS' and context.selected_assets and context.selected_assets[:] and context.space_data.params.asset_library_reference!='LOCAL'
    def modal(self, context,event):
        if event.type == 'TIMER':
            completed=len([a for a in self.threads if not a.is_alive()])
            context.scene.ta_progress=f"Updating Asset Data...({completed}/{len(self.threads)} files completed)"
            if context.area:
                context.area.tag_redraw()
            if all([not a.is_alive() for a in self.threads]):
                self.report({'INFO'},"Asset Data Updated!")
                bpy.ops.asset.library_refresh()
                for f,id in self.new_blend_names:
                    if os.path.isfile(f):
                        os.rename(f,f.replace(f"Temp_Name{id}",""))
                bpy.ops.asset.library_refresh()
                context.scene.ta_progress=""
                return {'FINISHED'}
        else:
            return {'PASS_THROUGH'}
        return {'RUNNING_MODAL'}
    def execute(self, context):
        blends={}
        for a in context.selected_assets:
            blends[a.full_library_path]=blends[a.full_library_path]+[(a.name,a.id_type),] if blends.get(a.full_library_path) else [(a.name,a.id_type),]
        
        self.threads=[]
        self.blend_names=[b for b,f in blends.items()]
        self.new_blend_names=[]
        id=get_id(3)
        for f in self.blend_names:
            if os.path.isfile(f) :
                
                os.rename(f,os.path.join(os.path.dirname(f),f"Temp_Name{id}"+os.path.basename(f)))
                
                self.new_blend_names.append((os.path.join(os.path.dirname(f),f"Temp_Name{id}"+os.path.basename(f)),id))
        bpy.ops.asset.library_refresh()
        for b,f in blends.items():
            self.threads.append(update_asset_data(os.path.join(os.path.dirname(b),f"Temp_Name{id}"+os.path.basename(b)),f,self.tags,self.author,self.description,self.add_or_replace))
        if preferences().non_blocking:
            wm=context.window_manager
            self.timer = wm.event_timer_add(0.2,window= context.window)
            wm.modal_handler_add(self) 
            return {'RUNNING_MODAL'}
        else:
            bpy.ops.asset.library_refresh() 
            return {'FINISHED'}
    def invoke(self,context,event):
        
        if not ensure_directory_is_set():
            self.report({'WARNING'},'Set directory path in the addon preferences!')
            return {'CANCELLED'}
        for a in context.selected_assets:
            active_lib_name=context.area.spaces.active.params.asset_library_reference
            active_lib_path=context.preferences.filepaths.asset_libraries.get(active_lib_name).path
            # self.dir_path=os.path.dirname(a.full_library_path)
            self.dir_path=active_lib_path
            break
        get_catalogs_to_move_to(self,context)
        return context.window_manager.invoke_props_dialog(self)

class TA_OT_Move_To_Catalog(bpy.types.Operator):
    bl_idname = "ta.movetocatalog"
    bl_label = "Move to Catalog (True Assets)"
    bl_description="Move selected assets to some catalog"
    bl_options = {'REGISTER', 'UNDO'}
    catalog:bpy.props.EnumProperty(items=lambda self, context: get_catalogs_to_move_to.catalogs,name="Catalog")
    new_catalog_name:bpy.props.StringProperty(default="",name="Catalog Name",search=get_catalog_search_results)
    dir_path:bpy.props.StringProperty(default="")
    def draw(self, context):
        layout=self.layout
        layout.ui_units_x=18
        layout.prop(self,'catalog')
        if self.catalog=="NEW":
            layout.prop(self,'new_catalog_name')
    @classmethod
    def poll(cls,context):
        return context.area.ui_type=='ASSETS' and context.selected_assets and context.selected_assets[:] and context.space_data.params.asset_library_reference!='LOCAL'
    def modal(self, context,event):
        if event.type == 'TIMER':
            completed=len([a for a in self.threads if not a.is_alive()])
            context.scene.ta_progress=f"Moving Assets...({completed}/{len(self.threads)} files completed)"
            if context.area:
                context.area.tag_redraw()
            if all([not a.is_alive() for a in self.threads]):
                self.report({'INFO'},"Moved to Catalog!")
                bpy.ops.asset.library_refresh()
                for f,id in self.new_blend_names:
                    if os.path.isfile(f):
                        os.rename(f,f.replace(f"Temp_Name{id}",""))
                bpy.ops.asset.library_refresh()
                context.scene.ta_progress=""
                return {'FINISHED'}
        else:
            return {'PASS_THROUGH'}
        return {'RUNNING_MODAL'}
    def execute(self, context):
        blends={}
        for a in context.selected_assets:
            blends[a.full_library_path]=blends[a.full_library_path]+[(a.name,a.id_type),] if blends.get(a.full_library_path) else [(a.name,a.id_type),]
        catalog="00000000-0000-0000-0000-000000000000"
        if self.catalog=="NEW":
            catalog=create_new_catalog(self.dir_path,self.new_catalog_name)
        else:
            catalog=self.catalog
        self.threads=[]
        self.blend_names=[b for b,f in blends.items()]
        self.new_blend_names=[]
        id=get_id(3)
        for f in self.blend_names:
            if os.path.isfile(f) :
                
                os.rename(f,os.path.join(os.path.dirname(f),f"Temp_Name{id}"+os.path.basename(f)))
                
                self.new_blend_names.append((os.path.join(os.path.dirname(f),f"Temp_Name{id}"+os.path.basename(f)),id))
        bpy.ops.asset.library_refresh()
        for b,f in blends.items():
            self.threads.append(move_assets_to_catalog(os.path.join(os.path.dirname(b),f"Temp_Name{id}"+os.path.basename(b)),f,catalog))
        if preferences().non_blocking:
            wm=context.window_manager
            self.timer = wm.event_timer_add(0.2,window= context.window)
            wm.modal_handler_add(self) 
            return {'RUNNING_MODAL'}
        else:
            bpy.ops.asset.library_refresh() 
            return {'FINISHED'}
    def invoke(self,context,event):
        
        if not ensure_directory_is_set():
            self.report({'WARNING'},'Set directory path in the addon preferences!')
            return {'CANCELLED'}
        for a in context.selected_assets:
            active_lib_name=context.area.spaces.active.params.asset_library_reference
            active_lib_path=context.preferences.filepaths.asset_libraries.get(active_lib_name).path
            # self.dir_path=os.path.dirname(a.full_library_path)
            self.dir_path=active_lib_path
            break
        get_catalogs_to_move_to(self,context)
        return context.window_manager.invoke_props_dialog(self)
class TA_OT_Clear_K3D_Asset(bpy.types.Operator):
    bl_idname = "ta.fixkitbashassets"
    bl_label = "Fix Kitbash Assets"
    bl_description="Fix some shading issues with kitbash assets by changing their material's displacement type from 'Displacement and Bump' to 'Only Bump'"
    bl_options = {'REGISTER', 'UNDO'}
    change_displacement_type:bpy.props.BoolProperty(default=True,name="Change Displacement Type",description="Change Displacement type on materials to 'BUMP'")
    remove_subdiv:bpy.props.BoolProperty(default=True,name="Remove Subsurf Modifiers",description="Remove subsurface modifiers from objects.")
    @classmethod
    def poll(cls,context):
        return context.area.ui_type=='ASSETS' and context.selected_assets and context.selected_assets[:]
    def draw(self, context):
        if context.space_data.params.asset_library_reference!='LOCAL':
            layout=self.layout
            layout.prop(self,'change_displacement_type',toggle=True)
            layout.prop(self,'remove_subdiv',toggle=True)
    def modal(self, context,event):
        if event.type == 'TIMER':
            completed=len([a for a in self.threads if not a.is_alive()])
            context.scene.ta_progress=f"Fixing Kitbash Assets...({completed}/{len(self.threads)} files completed)"
            if context.area:
                context.area.tag_redraw()
            if all([not a.is_alive() for a in self.threads]):
                self.report({'INFO'},"Task Completed!")
                bpy.ops.asset.library_refresh()
                context.scene.ta_progress=""
                if context.area:
                    context.area.tag_redraw()
                return {'FINISHED'}
        else:
            return {'PASS_THROUGH'}
        return {'RUNNING_MODAL'}
    def execute(self, context):
        if context.space_data.params.asset_library_reference!='LOCAL':
            blends={}
            for a in context.selected_assets:
                blends[a.full_library_path]=blends[a.full_library_path]+[(a.name,a.id_type),] if blends.get(a.full_library_path) else [(a.name,a.id_type),]
            self.threads=[]
            for b,f in blends.items():
                self.threads.append(fix_kitbash_assets(b,f,self.change_displacement_type,self.remove_subdiv))
            if preferences().non_blocking:
                wm=context.window_manager
                self.timer = wm.event_timer_add(1,window= context.window)
                wm.modal_handler_add(self) 
                return {'RUNNING_MODAL'}
            else:
                bpy.ops.asset.library_refresh() 
                return {'FINISHED'}
        else:
            for a in context.selected_assets:
                a.local_id.asset_clear()
                #bpy.ops.asset.clear({'asset':a,'active_file':a,'selected_assets':[a,],'selected_files':[a,]},set_fake_user=False)
            return {'FINISHED'}
    def invoke(self,context,event):
        return context.window_manager.invoke_props_dialog(self)
class TA_OT_Delete_Catalog(bpy.types.Operator):
    bl_idname = "ta.deletecatalog"
    bl_label = "Delete Catalog"
    bl_description="Delete Catalog"
    bl_options = {'REGISTER', 'UNDO'}
   
    @classmethod
    def poll(cls,context):
        return context.area.ui_type=='ASSETS' 
    
    def execute(self, context):
        catalog_path='asmjansvmiaenavjnnlxvkmssdoifinauodnamfoankfmozc'
        if context.space_data.params.asset_library_reference!='LOCAL':
            catalog_id=context.space_data.params.catalog_id
            cats_file=get_library_cat_file(context.space_data.params.asset_library_reference)
            if cats_file:
                file_data=[]
                with open(cats_file, 'r') as filedata:
                    file_data=filedata.readlines()
                for line in file_data:
                    if catalog_id in line:
                        catalog_path=line.split(':')
                        if len(catalog_path)==3:
                            catalog_path=catalog_path[1].replace("\n","")+'/'
                if file_data:
                    with open(cats_file,mode='w+',newline='\n') as file:
                        for line in file_data:
                            if catalog_id not in line and catalog_path not in line:
                                file.write(line)
        try:
            bpy.ops.asset.library_refresh() 
        except:
            pass
        return {'FINISHED'}
    def invoke(self,context,event):
        return context.window_manager.invoke_confirm(self,event)
class TA_OT_Rename_Catalog(bpy.types.Operator):
    bl_idname = "ta.renamecatalog"
    bl_label = "Rename Catalog"
    bl_description="Rename Catalog"
    bl_options = {'REGISTER', 'UNDO'}
    bl_property='name'
    name:bpy.props.StringProperty(default="",name="New Name")
    @classmethod
    def poll(cls,context):
        return context.area.ui_type=='ASSETS' and context.space_data.params.asset_library_reference!='LOCAL'
    
    def execute(self, context):
        catalog_path='asmjansvmiaenavjnnlxvkmssdoifinauodnamfoankfmozc'
        if context.space_data.params.asset_library_reference!='LOCAL':
            catalog_id=context.space_data.params.catalog_id
            cats_file=get_library_cat_file(context.space_data.params.asset_library_reference)
            if cats_file:
                file_data=[]
                with open(cats_file, 'r') as filedata:
                    file_data=filedata.readlines()
                for line in file_data:
                    if catalog_id in line:
                        catalog_path=line.split(':')
                        if len(catalog_path)==3:
                            catalog_path=catalog_path[1].replace("\n","")
                if file_data:
                    with open(cats_file,mode='w+',newline='\n') as file:
                        for line in file_data:
                            if catalog_path in line:
                                if '/' not in catalog_path:
                                    line=line.replace(catalog_path,self.name,1)
                                else:
                                    # print(catalog_path,os.path.dirname(catalog_path),os.path.dirname(catalog_path)+"/"+self.name)
                                    line=line.replace(catalog_path,os.path.dirname(catalog_path)+"/"+self.name,1)
                            file.write(line)
        try:
            bpy.ops.asset.library_refresh() 
        except:
            pass
        return {'FINISHED'}
    def invoke(self,context,event):
        return context.window_manager.invoke_props_dialog(self)
class TA_OT_Clear_Broken_Materials(bpy.types.Operator):
    bl_idname = "ta.clearbrokenmaterials"
    bl_label = "Clear Broken Materials (True Assets)"
    bl_description="Clear Broken Materials"
    bl_options = {'REGISTER', 'UNDO'}
    @classmethod
    def poll(cls,context):
        return context.area.ui_type=='ASSETS' and context.selected_assets and context.selected_assets[:]
    
    def modal(self, context,event):
        if event.type == 'TIMER':
            completed=len([a for a in self.threads if not a.is_alive()])
            context.scene.ta_progress=f"Clearing Assets...({completed}/{len(self.threads)} files completed)"
            if context.area:
                context.area.tag_redraw()
            if all([not a.is_alive() for a in self.threads]):
                self.report({'INFO'},"Assets Cleared!")
                bpy.ops.asset.library_refresh()
                context.scene.ta_progress=""
                if context.area:
                    context.area.tag_redraw()
                return {'FINISHED'}
        else:
            return {'PASS_THROUGH'}
        return {'RUNNING_MODAL'}
    def execute(self, context):
        if context.space_data.params.asset_library_reference!='LOCAL':
            blends={}
            for a in context.selected_assets:
                if a.id_type=='MATERIAL':
                    blends[a.full_library_path]=blends[a.full_library_path]+[(a.name,a.id_type),] if blends.get(a.full_library_path) else [(a.name,a.id_type),]
            self.threads=[]
            for b,f in blends.items():
                self.threads.append(clear_broken_materials(b,f))
            if preferences().non_blocking:
                wm=context.window_manager
                self.timer = wm.event_timer_add(1,window= context.window)
                wm.modal_handler_add(self) 
                return {'RUNNING_MODAL'}
            else:
                bpy.ops.asset.library_refresh() 
                return {'FINISHED'}
        else:
            for a in context.selected_assets:
                a.local_id.asset_clear()
                #bpy.ops.asset.clear({'asset':a,'active_file':a,'selected_assets':[a,],'selected_files':[a,]},set_fake_user=False)
            return {'FINISHED'}
    def invoke(self,context,event):
        return context.window_manager.invoke_props_dialog(self)

class TA_OT_Clear_Asset(bpy.types.Operator):
    bl_idname = "ta.clearasset"
    bl_label = "Clear Assets (True Assets)"
    bl_description="Clear Assets"
    bl_options = {'REGISTER', 'UNDO'}
    delete:bpy.props.BoolProperty(default=False,name="Delete Blend file if no other asset is found!")
    remove_data:bpy.props.BoolProperty(default=False,name="Delete Data",options={'SKIP_SAVE'})
    remove_linked_data:bpy.props.BoolProperty(default=False,name="Delete Linked Data (Materials and Images)",options={'SKIP_SAVE'})
    duplicates:bpy.props.BoolProperty(default=False,name="Only Duplicates",options={'SKIP_SAVE'})
    @classmethod
    def poll(cls,context):
        return context.area.ui_type=='ASSETS' and context.selected_assets and context.selected_assets[:]
    def draw(self, context):
        if context.space_data.params.asset_library_reference!='LOCAL':
            layout=self.layout
            layout.prop(self,'duplicates',toggle=True)
            layout.prop(self,'remove_data',toggle=True)
            layout.prop(self,'remove_linked_data',toggle=True)
            layout.prop(self,'delete',toggle=True)
    def modal(self, context,event):
        if event.type == 'TIMER':
            completed=len([a for a in self.threads if not a.is_alive()])
            context.scene.ta_progress=f"Clearing Assets...({completed}/{len(self.threads)} files completed)"
            if context.area:
                context.area.tag_redraw()
            if all([not a.is_alive() for a in self.threads]):
                self.report({'INFO'},"Assets Cleared!")
                bpy.ops.asset.library_refresh()
                context.scene.ta_progress=""
                if context.area:
                    context.area.tag_redraw()
                return {'FINISHED'}
        else:
            return {'PASS_THROUGH'}
        return {'RUNNING_MODAL'}
    def execute(self, context):
        if context.space_data.params.asset_library_reference!='LOCAL':
            blends={}
            
            name_counter = defaultdict(int)
            for a in context.selected_assets:
                if self.duplicates:
                    name_counter[a.name+a.id_type] += 1
                    if name_counter[a.name+a.id_type] > 1:
                        blends[a.full_library_path]=blends[a.full_library_path]+[(a.name,a.id_type),] if blends.get(a.full_library_path) else [(a.name,a.id_type),]
                else:
                    blends[a.full_library_path]=blends[a.full_library_path]+[(a.name,a.id_type),] if blends.get(a.full_library_path) else [(a.name,a.id_type),]
            # print(blends)
            self.threads=[]
            for b,f in blends.items():
                self.threads.append(clear_assets_from_external_file(b,f,self.delete,self.remove_data,self.remove_linked_data))
            if preferences().non_blocking:
                wm=context.window_manager
                self.timer = wm.event_timer_add(1,window= context.window)
                wm.modal_handler_add(self) 
                return {'RUNNING_MODAL'}
            else:
                bpy.ops.asset.library_refresh() 
                return {'FINISHED'}
        else:
            for a in context.selected_assets:
                a.local_id.asset_clear()
                #bpy.ops.asset.clear({'asset':a,'active_file':a,'selected_assets':[a,],'selected_files':[a,]},set_fake_user=False)
            return {'FINISHED'}
    def invoke(self,context,event):
        # context.space_data.active_asset_by_id("")
        return context.window_manager.invoke_props_dialog(self)
class TA_OT_Edit_Asset(bpy.types.Operator):
    bl_idname = "ta.editasset"
    bl_label = "Edit Asset (True Assets)"
    bl_description="Open Blend file containing Selected Asset"
    bl_options = {'REGISTER', 'UNDO'}
    @classmethod
    def poll(cls,context):
        return context.area.ui_type=='ASSETS' and context.selected_assets and context.selected_assets[:] and context.space_data.params.asset_library_reference!='LOCAL'
    def execute(self, context):
        active_asset=context.asset
        blend_path=active_asset.full_library_path
        edit_assets_from_external_file(blend_path,[(active_asset.name,active_asset.id_type),],context.window.workspace.name)
        bk = blend_path + "1"
        if os.path.exists(bk):
            os.remove(bk)
        bpy.ops.asset.library_refresh()
        return {'FINISHED'}
class TA_OT_Reset_Progress(bpy.types.Operator):
    bl_idname = "ta.resetprogress"
    bl_label = "Reset Status in case of a crash"
    bl_description="Reset Status and re-enable buttons in case a crash happens"
    bl_options = {'REGISTER', 'UNDO'}
    def execute(self, context):
        for scene in bpy.data.scenes:
            scene.ta_progress=""
        return {'FINISHED'}
class TA_OT_Clear_Asset_By_Type(bpy.types.Operator):
    bl_idname = "ta.clearassetsbytype"
    bl_label = "Clear Assets By Type(True Assets)"
    bl_description="Clear Assets by asset type"
    bl_options = {'REGISTER', 'UNDO'}
    type:bpy.props.EnumProperty(name="Type",items=(("OBJECT","Objects","Objects"),("MATERIAL","Materials","Materials"),("WORLD","Worlds","Worlds"),("NODETREE","Node Groups","Node Groups")))
    delete:bpy.props.BoolProperty(default=False,name="Delete Blend file if no other asset is found!")
    remove_data:bpy.props.BoolProperty(default=False,name="Delete Data",options={'SKIP_SAVE'})
    remove_linked_data:bpy.props.BoolProperty(default=False,name="Delete Linked Data (Materials and Images)",options={'SKIP_SAVE'})
    @classmethod
    def poll(cls,context):
        return context.area.ui_type=='ASSETS' and context.selected_assets and context.selected_assets[:]
    def draw(self, context):
        layout=self.layout
        layout.prop(self,'type')
        if context.space_data.params.asset_library_reference!='LOCAL':
            layout.prop(self,'remove_data',toggle=True)
            layout.prop(self,'remove_linked_data',toggle=True)
            layout.prop(self,'delete',toggle=True)
    def modal(self, context,event):
        if event.type == 'TIMER':
            completed=len([a for a in self.threads if not a.is_alive()])
            context.scene.ta_progress=f"Clearing Assets...({completed}/{len(self.threads)} files completed)"
            if context.area:
                context.area.tag_redraw()
            if all([not a.is_alive() for a in self.threads]):
                self.report({'INFO'},"Assets Cleared!")
                bpy.ops.asset.library_refresh()
                context.scene.ta_progress=""
                if context.area:
                    context.area.tag_redraw()
                return {'FINISHED'}
        else:
            return {'PASS_THROUGH'}
        return {'RUNNING_MODAL'}
    def execute(self, context):
        if context.space_data.params.asset_library_reference!='LOCAL':
            blends={}
            
            for a in context.selected_assets:
                if a.id_type==self.type:
                    blends[a.full_library_path]=blends[a.full_library_path]+[(a.name,a.id_type),] if blends.get(a.full_library_path) else [(a.name,a.id_type),]
        # print([(a.name,a.id_type,a.full_library_path) for a in context.selected_assets])
            #print(context.space_data.params.asset_library_reference)
            #print(context.asset.id_type)
            self.threads=[]
            for b,f in blends.items():
                self.threads.append(clear_assets_from_external_file(b,f,self.delete,self.remove_data,self.remove_linked_data))
            if preferences().non_blocking:
                wm=context.window_manager
                self.timer = wm.event_timer_add(1,window= context.window)
                wm.modal_handler_add(self) 
                return {'RUNNING_MODAL'}
            else:
                bpy.ops.asset.library_refresh() 
                return {'FINISHED'}
        else:
            for a in context.selected_assets:
                if a.id_type==self.type:
                    a.local_id.asset_clear()
                #bpy.ops.asset.clear({'asset':a,'active_file':a,'selected_assets':[a,],'selected_files':[a,]},set_fake_user=False)
            return {'FINISHED'}
    def invoke(self,context,event):
        return context.window_manager.invoke_props_dialog(self)
class TA_OT_Rerender_Thumbnail(bpy.types.Operator):
    bl_idname = "ta.rerenderthumbnail"
    bl_label = "Re-Render Thumbnail (True Assets)"
    bl_description=""
    bl_options = {'REGISTER', 'UNDO'}
    shading:bpy.props.EnumProperty(name="Thumbnail Shading",default=0,items=(("Material","Material","Material"),("Eevee","Eevee","Eevee"),("Cycles","Cycles","Cycles")))
    camera_angle:bpy.props.EnumProperty(items=(('XZ','X-Axis','X-Axis'),('YZ','Y-Axis','Y-Axis'),('XYZ','Isometric','Equal distance on all axis')))
    flip_z:bpy.props.BoolProperty(name="Flip Z",default=False)
    flip_x:bpy.props.BoolProperty(name="Flip X",default=False)
    flip_y:bpy.props.BoolProperty(name="Flip Y",default=False)
    thumb_res:bpy.props.IntProperty(default=256,min=128,max=1024,subtype='PIXEL',name="Resolution")
    camera_height:bpy.props.FloatProperty(default=1.0,min=0,max=2,soft_max=1,subtype='PERCENTAGE',name='Camera Height')
    camera_zoom:bpy.props.FloatProperty(default=1,min=0.5,max=5,soft_max=2,name="Zoom")
    @classmethod
    def poll(cls,context):
        return context.area.ui_type=='ASSETS' and context.selected_assets and context.selected_assets[:] and context.space_data.params.asset_library_reference!='LOCAL'
    def draw(self, context):
        layout=self.layout
        layout.prop(self,'shading',expand=True)
        if self.shading!='Solid':
            layout.prop(self,'camera_angle',expand=True)
            row=layout.row(align=True)
            # layout.prop(self,'camera_height')
            # layout.prop(self,'camera_zoom')
            row.prop(self,'flip_x',toggle=True)
            row.prop(self,'flip_y',toggle=True)
            row.prop(self,'flip_z',toggle=True)

            if not self.flip_z:
                row=layout.row(align=True)
                row.prop(preferences(),'add_ground_plane',toggle=True)
                # row.prop(preferences(),'use_studio_background',toggle=True)
            # if preferences().add_ground_plane:
            #     layout.prop(context.scene,'ta_material')
            # if preferences().use_studio_background:
            #     layout.template_icon_view(
            #         context.scene, "ta_studio", show_labels=True, scale=8, scale_popup=6)
            # layout.prop(self,'thumb_res')
    def modal(self, context,event):
        if event.type == 'TIMER':
            context.scene.ta_progress=f"Regenerating Thumbnails..."
            if context.area:
                context.area.tag_redraw()
            if all([not a.is_alive() for a in self.threads]):
                self.report({'INFO'},"Thumbnails Regenerated!")
                bpy.ops.asset.library_refresh()
                context.scene.ta_progress=""
                if context.area:
                    context.area.tag_redraw()
                return {'FINISHED'}
        else:
            return {'PASS_THROUGH'}
        return {'RUNNING_MODAL'}
    def execute(self, context):
        blends={}
        for a in context.selected_assets:
            if a.id_type=='OBJECT' or a.id_type=='COLLECTION' or a.id_type=='MATERIAL':
                blends[a.full_library_path]=blends[a.full_library_path]+[(a.name,a.id_type),] if blends.get(a.full_library_path) else [(a.name,a.id_type),]
        self.threads=[]
        lib_path=[a for a in bpy.context.preferences.filepaths.asset_libraries if a.name==context.space_data.params.asset_library_reference]
        if lib_path:
            lib_path=lib_path[0].path
        #for b,f in blends.items():
        print(blends)
        self.threads.append(rerender_thumbnail([b for b in blends.keys()],lib_path if lib_path else preferences().objects_directory,[f for f in blends.values()],self.shading,angle=resolve_angle(self.camera_angle,self.flip_x,self.flip_y,self.flip_z),add_plane=preferences().add_ground_plane and not self.flip_z))
        if preferences().non_blocking:
            wm=context.window_manager
            self.timer = wm.event_timer_add(1,window= context.window)
            wm.modal_handler_add(self) 
            return {'RUNNING_MODAL'}
        else:
            bpy.ops.asset.library_refresh() 
            return {'FINISHED'}
    def invoke(self,context,event):
        return context.window_manager.invoke_props_dialog(self)

class TA_OT_Create_Asset(bpy.types.Operator):
    bl_idname = "ta.markobjectasset"
    bl_label = "Mark Asset (True Assets)"
    bl_description="Mark Asset"
    bl_options = {'REGISTER', 'UNDO'}
    #bl_property = "tags"
    name:bpy.props.StringProperty(default="",name="File Name",update=recheck_if_object_file_exists)
    description:bpy.props.StringProperty(default="",name="Description")
    tags:bpy.props.StringProperty(default="",name="Tags")
    mark_object:bpy.props.BoolProperty(default=True,name="Mesh")
    mark_materials:bpy.props.BoolProperty(default=False,name="Materials")
    mark_collection:bpy.props.BoolProperty(default=False,name='Collection')
    shading:bpy.props.EnumProperty(name="Thumbnail Shading",default=0,items=(("Material","Material","Material"),("Eevee","Eevee","Eevee"),("Cycles","Cycles","Cycles")))
    override:bpy.props.EnumProperty(name="If File Already Exists",items=(('OVERRIDE','Override','Override'),('COPY','Create New','Create New')))
    exists:bpy.props.BoolProperty(default=False,options={'SKIP_SAVE'})
    catalog:bpy.props.EnumProperty(items=get_object_catalogs,name="Catalog",default=1)
    new_catalog_name:bpy.props.StringProperty(default="",name="Catalog Name",search=get_catalog_search_results)
    camera_angle:bpy.props.EnumProperty(items=(('XZ','X-Axis','X-Axis'),('YZ','Y-Axis','Y-Axis'),('XYZ','Isometric','Equal distance on all axis'),('VIEW','View','Render from Viewing Angle')))
    flip_z:bpy.props.BoolProperty(name="Flip Z",default=False)
    flip_x:bpy.props.BoolProperty(name="Flip X",default=False)
    flip_y:bpy.props.BoolProperty(name="Flip Y",default=False)
    apply_rotation:bpy.props.BoolProperty(default=False,name="Apply Rotation")
    apply_scale:bpy.props.BoolProperty(default=False,name="Apply Scale")
    move_origin_to_bottom:bpy.props.BoolProperty(default=False,name="Move origin to bottom")
    @classmethod
    def poll(cls,context):
        return context.active_object and context.selected_objects

    def draw(self, context):
        layout=self.layout
        row=layout.row(align=True)
        row.prop(self,'mark_object',toggle=True)
        row2=row.row()
        row2.prop(self,'mark_materials',toggle=True)

        if not self.mark_object:
            row2.enabled=False
        
        row.prop(self,'mark_collection',toggle=True)
        
        #    row.enabled=False
        layout.prop(self,'name')
        layout.prop(self,'description')
        layout.prop(self,'tags')
        row=layout.row(align=True)
        row.prop(self,'apply_rotation',toggle=True)
        row.prop(self,'apply_scale',toggle=True)
        if self.mark_object:
            layout.prop(self,'move_origin_to_bottom')
        layout.prop(self,'shading',expand=True)
        if self.shading=='Cycles':
            layout.label(text="Might be Slow",icon='INFO')
        if self.shading!="Solid":
            layout.label(text="Camera Angle")
            layout.prop(self,'camera_angle',expand=True)
            row=layout.row(align=True)
            row.prop(self,'flip_x',toggle=True)
            row.prop(self,'flip_y',toggle=True)
            row.prop(self,'flip_z',toggle=True)
            if not self.flip_z:
                layout.prop(preferences(),'add_ground_plane',toggle=True)
        layout.prop(self,'catalog')
        if self.catalog=='NEW':
            layout.prop(self,'new_catalog_name')
        if self.exists:
            layout.label(text="File with same name already exists!")
            layout.prop(self,'override',expand=True)
    def execute(self, context):
        objects_to_mark=[]
        objects_to_check_previews=[]
        catalog=None
        location_backup=[]
        bpy.ops.object.transform_apply(location=False, rotation=self.apply_rotation, scale=self.apply_scale)
        col=None
        if context.active_object and self.mark_collection:
            col=context.active_object.users_collection[0]
            bottom_left,top_right=get_bounding_box_of_collection(col)
            bbox_center=(bottom_left+top_right)/2
            col.og_instance_offset=col.instance_offset.copy()
            col.instance_offset=mathutils.Vector((bbox_center.x,bbox_center.y,0))
            col.asset_mark()
            
            if self.catalog!="NONE":
                if self.catalog=="NEW":
                    catalog=create_new_catalog(preferences().objects_directory,self.new_catalog_name)
                else:
                    catalog=self.catalog
                col.asset_data.catalog_id=catalog
            for t in self.tags.split(","):
                col.asset_data.tags.new(t)
            create_thumbnail(context,col,self.shading,angle=resolve_angle(self.camera_angle,self.flip_x,self.flip_y,self.flip_z),add_plane=preferences().add_ground_plane and not self.flip_z)
            col.asset_data.description=self.description
            objects_to_mark.append(col)
            if [a for a in col.all_objects if a.type!='EMPTY' and a.type!='LIGHT' and a.display_type!='WIRE'and a.display_type!='BOUNDS' and a.hide_render==False]:
                            objects_to_check_previews.append(col)
        if self.mark_object:
            for ob in context.selected_objects:
                if ob.type in {'MESH','CURVE','SURFACE','TEXT','VOLUME'}:
                    objects_to_mark.append(ob)
                    if self.move_origin_to_bottom:
                        set_new_origin_of_object(ob)
                        
                    ob.asset_mark()
                    if self.catalog!="NONE":
                        if self.catalog=="NEW":
                            catalog=create_new_catalog(preferences().objects_directory,self.new_catalog_name)
                        else:
                            catalog=self.catalog
                        ob.asset_data.catalog_id=catalog
                    for t in self.tags.split(","):
                        ob.asset_data.tags.new(t)
                    if does_support_thumbnails(ob):
                        objects_to_check_previews.append(ob)
                        create_thumbnail(context,ob,self.shading,angle=resolve_angle(self.camera_angle,self.flip_x,self.flip_y,self.flip_z),add_plane=preferences().add_ground_plane and not self.flip_z)
                    ob.asset_data.description=self.description
        bpy.app.timers.register(functools.partial(
            ensure_preview_and_save_asset, 'OBJECTS',objects_to_mark,objects_to_check_previews,self.mark_materials,name=self.name,override=self.override,location_backup=location_backup), first_interval=0)
        
        return {'FINISHED'}

    
    def invoke(self, context,event):
        if not ensure_directory_is_set():
            self.report({'WARNING'},'Set directory path in the addon preferences!')
            return {'CANCELLED'}
        ensure_all_libraries()
        self.dir_path=preferences().objects_directory
        get_catalogs_to_move_to(self,context)
        if context.active_object:
            self.name=context.active_object.name
        self.tags=self.name
        return context.window_manager.invoke_props_dialog(self)
class TA_OT_Fetch_Online_Asset(bpy.types.Operator):
    bl_idname = "ta.fetchonlineassets"
    bl_label = "Fetch Online Assets"
    bl_description="Fetch new assets from the online libraries\nCTRL+SHIFT+ALT+LMB: Erase existing data and perform a complete reload."
    bl_options = {'REGISTER', 'UNDO'}
    running=False
    @classmethod
    def start_op(cls):
        cls.running=True
    @classmethod
    def stop_op(cls):
        cls.running=False
    @classmethod
    def poll(cls,context):
        return context.area.ui_type=='ASSETS'
    def modal(self, context,event):
        if event.type == 'TIMER':
            if os.path.exists(os.path.join(os.path.dirname(os.path.dirname(__file__)),"progress.txt")):
                with open(os.path.join(os.path.dirname(os.path.dirname(__file__)),"progress.txt"), "r",encoding='utf-8') as f:
                    status=f.read().splitlines()
                    try:
                        context.scene.ta_progress=status[0]
                        context.scene.ta_progress_bar=min(99,eval(status[1])*100)
                    except:
                        pass
            if context.area:
                context.area.tag_redraw()
            if self.thread and not self.thread.is_alive() :
                
                if self.remaining_modules:
                    module=self.remaining_modules.pop(0)
                    if get_ta_module(module):
                        t=threading.Thread(target=functools.partial(fetch_assets_for_module,preferences().online_assets_directory,module))
                        t.start()
                        self.thread=t
                        with open(os.path.join(os.path.dirname(os.path.dirname(__file__)),"progress.txt"), "w+",encoding='utf-8') as f:
                            f.writelines(["\n",str(0)])
                        context.scene.ta_progress_bar=0
                    return {'RUNNING_MODAL'}
                else:
                    self.report({'INFO'},"Fetching Completed!")
                    self.stop_op()
                    context.scene.ta_progress=""
                return {'FINISHED'}
        else:
            return {'PASS_THROUGH'}
        return {'RUNNING_MODAL'}
    def execute(self, context):
        self.remaining_modules=['ShareTextures',"AmbientCG"]
        module=self.remaining_modules.pop(0)
        while not get_ta_module(module) and self.remaining_modules:
            module=self.remaining_modules.pop(0)
        if get_ta_module(module):
            
            t=threading.Thread(target=functools.partial(fetch_assets_for_module,preferences().online_assets_directory,module))
            t.start()
            self.thread=t
            with open(os.path.join(os.path.dirname(os.path.dirname(__file__)),"progress.txt"), "w+",encoding='utf-8') as f:
                f.writelines(["\n",str(0)])
            context.scene.ta_progress_bar=0
            wm=context.window_manager
            self.timer = wm.event_timer_add(1,window= context.window)
            wm.modal_handler_add(self) 
            self.start_op()
            return {'RUNNING_MODAL'}
        
        return {'FINISHED'}
    def invoke(self,context,event):
        if event.ctrl and event.alt and event.shift:
            if os.path.isdir(preferences().online_assets_directory):
                delete_folder_data(preferences().online_assets_directory)
        return self.execute(context)

    
class TA_OT_Switch_Shader_Type(bpy.types.Operator):
    bl_idname = "ta.switchshader"
    bl_label = "Switch Shader"
    bl_description=""
    bl_options = {'REGISTER', 'UNDO'}
    type:bpy.props.EnumProperty(items=[('Material','Active Material','Material'),('World','Scene World','World')])
    shader:bpy.props.EnumProperty(items=(("True-Uber","Uber","True-Uber"),("True-Plastic","Plastic","True-Plastic"),("True-Organic","Organic","True-Organic"),("True-BSDF","BSDF","True-BSDF"),("Standard","Standard","Standard"),("Ambient CG (Uber)","AmbientCG-Uber","Ambient CG (Uber)"),("Ambient CG (Standard)","AmbientCG Standard","Ambient CG (Standard)")),name="Shader Type")
    world_shader:bpy.props.EnumProperty(items=(("True-HDRI","True-HDRI","True-HdRi"),("True-HDRI (Standard)","Standard","True-HDRI Standard")),name="Shader Type")
    @classmethod
    def poll(cls,context):
        
        return (context.active_object and context.active_object.active_material) or context.scene.world
    def draw(self,context):
        layout=self.layout
        layout.row().prop(self,'type',expand=True)
        if self.type=='Material':
            layout.prop(self,'shader')
        else:
            layout.prop(self,'world_shader')
    def execute(self, context):
        if context.active_object and context.active_object.active_material and self.type=='Material':
            switch_material_shader(context.active_object.active_material,type=self.shader)
        elif context.scene.world and self.type=='World':
            context.scene.world=switch_world_shader(context.scene.world,self.world_shader,shader_values={'Tonemapped Toggle':0},context=context)
        return {'FINISHED'}
    def invoke(self,context,event):

        return context.window_manager.invoke_props_dialog(self)
def drawIntoNodeGroupMenu(self, context):
    layout = self.layout
    layout.operator_context = "INVOKE_DEFAULT"
    layout.operator("ta.marknodegroupasset")
def drawIntoOutliner(self, context):
    layout = self.layout
    layout.operator_context = "INVOKE_DEFAULT"
    layout.operator("ta.markobjectasset")
def drawIntoAssetContextMenu(self, context):
    layout=self.layout
    active_lib_name=context.area.spaces.active.params.asset_library_reference
    if active_lib_name not in {'LOCAL','ALL','ESSENTIALS'}:
        layout.operator_context= "INVOKE_DEFAULT"
        layout.separator(factor=0)
        layout.label(text="True Assets",icon='ASSET_MANAGER')
        layout.separator(factor=0)
        if context.selected_assets and len(context.selected_assets[:])==1 and context.selected_assets[0].id_type in ['OBJECT','COLLECTION']:
            layout.operator("ta.replaceinscene")
        layout.operator("ta.editasset")
        layout.operator("ta.renameassets")
        layout.operator("ta.updateassetdata")
        layout.operator("ta.clearasset")
        layout.operator("ta.clearassetsbytype")
        layout.operator("ta.fixkitbashassets")
        layout.operator("ta.movetocatalog")
        layout.operator("ta.rerenderthumbnail")
        layout.operator("ta.deletecatalog")
        layout.operator("ta.renamecatalog")
        
    
def drawIntoAssetHeader(self, context):
    layout=self.layout
    row=layout.row(align=True)
    pcoll = icon_collection["icons"]
    ambientcg_icon=pcoll['ambientcg']
    active_lib_name=context.area.spaces.active.params.asset_library_reference
    if active_lib_name not in {'LOCAL','ALL','ESSENTIALS'}:
        row.menu("TA_MT_Header_Menu")
    row.operator('ta.opentalibrary',text='',icon='OBJECT_DATA').name=preferences().object_library_name
    row.operator('ta.opentalibrary',text='',icon='MATERIAL').name=preferences().material_library_name
    row.operator('ta.opentalibrary',text='',icon='WORLD').name=preferences().world_library_name
    row.operator('ta.opentalibrary',text='',icon='NODETREE').name=preferences().node_group_library_name
    row.operator('ta.opentalibrary',text='',icon='BLENDER').name=preferences().blends_library_name
    if EXTRA_MODULES:
        row.operator('ta.opentalibrary',text='',icon_value=ambientcg_icon.icon_id).name=preferences().online_library_name
    if SpaceAssetInfo.is_asset_browser(context.space_data) and context.space_data.params.asset_library_reference == preferences().online_library_name and EXTRA_MODULES:
        layout.separator()    
        row=layout.row()
        row.operator("ta.fetchonlineassets",icon='FILE_REFRESH',text="Fetch")
        row.enabled=not TA_OT_Fetch_Online_Asset.running
    layout=layout.row(align=True)
    if context.scene.ta_progress:
        if context.scene.ta_progress in ['Fetching Materials...','Fetching HDRIs...','Downloading Previews...'] or 'Downloading Material' in context.scene.ta_progress or 'Downloading HDRI' in context.scene.ta_progress :
            row=layout.row()
            row.prop(context.scene,'ta_progress_bar',slider=True,text=context.scene.ta_progress)
            row.enabled=False
            if 'Downloading Material' in context.scene.ta_progress or 'Downloading HDRI' in context.scene.ta_progress:
                row2=layout.row()
                row2.operator("ta.canceldownload",icon='CANCEL',text='')
                row2.alert=True
        else:
            layout.label(text=context.scene.ta_progress)
def draw_into_header_right(self,context):
        

        layout = self.layout

        space_data = context.space_data

        if space_data.active_operator is None:
            layout.template_header()

        if SpaceAssetInfo.is_asset_browser(space_data) and context.space_data.params.asset_library_reference == preferences().online_library_name:
            if "true_terrain" in bpy.context.preferences.addons.keys():
                layout.prop(context.scene,'use_as_tt_layer')
                
            layout.label(text="Options:")
            layout.prop(context.scene,'ta_material_quality',text='Quality')
            if context.asset and context.asset.id_type=='MATERIAL':
                layout.prop(context.scene,'ta_shader_type',text='Shader')
            if context.asset and context.asset.id_type=='WORLD':
                layout.prop(context.scene,'ta_world_shader_type',text='Shader')
            layout.operator("ta.showdownloaded",icon='IMPORT',text='Show Downloaded',depress=context.space_data.params.filter_search=='Downloaded')
            # layout.operator("ta.fetchonlineassets",icon='FILE_REFRESH',text="Fetch")
            
class TA_MT_Header_Menu(bpy.types.Menu):
    bl_label="True Assets"
    def draw(self, context):
        layout = self.layout
        layout.operator_context= "INVOKE_DEFAULT"
        active_lib_name=context.area.spaces.active.params.asset_library_reference
        if active_lib_name not in {'LOCAL','ALL','ESSENTIALS'}:
            if len(context.selected_assets[:])==1 and context.selected_assets[0].id_type in ['OBJECT','COLLECTION']:
                layout.operator("ta.replaceinscene")
            
            layout.operator("ta.editasset")
            layout.operator("ta.renameassets")
            layout.operator("ta.updateassetdata")
            layout.operator("ta.clearasset")
            layout.operator("ta.clearassetsbytype")
            layout.operator("ta.clearbrokenmaterials")
            layout.operator("ta.fixkitbashassets")
            layout.operator("ta.movetocatalog")
            
            layout.operator("ta.rerenderthumbnail")
            layout.operator("ta.deletecatalog")
            layout.operator("ta.batchdeletecatalogs")
            layout.operator("ta.renamecatalog")
            layout.operator("ta.reorderlibraries")
            # layout.operator("ta.exportjson")
    
class TA_Props(bpy.types.PropertyGroup):
    pass
class TA_Sub_Section(bpy.types.PropertyGroup):
    show_material_section_1:bpy.props.BoolProperty(default=False)
    show_material_section_2:bpy.props.BoolProperty(default=False)
    show_material_section_3:bpy.props.BoolProperty(default=False)
    show_material_section_4:bpy.props.BoolProperty(default=False)
    show_material_section_5:bpy.props.BoolProperty(default=False)
    show_material_section_6:bpy.props.BoolProperty(default=False)
    show_material_section_7:bpy.props.BoolProperty(default=False)
    show_material_section_8:bpy.props.BoolProperty(default=False)
    show_material_section_9:bpy.props.BoolProperty(default=False)
    show_material_section_10:bpy.props.BoolProperty(default=False)
    show_material_section_11:bpy.props.BoolProperty(default=False)
    show_material_section_12:bpy.props.BoolProperty(default=False)
    show_material_section_13:bpy.props.BoolProperty(default=False)
    show_material_section_14:bpy.props.BoolProperty(default=False)
    show_material_adv_section_1:bpy.props.BoolProperty(default=False)
    show_material_adv_section_2:bpy.props.BoolProperty(default=False)
    show_material_adv_section_3:bpy.props.BoolProperty(default=False)
    show_material_adv_section_4:bpy.props.BoolProperty(default=False)
    show_material_adv_section_5:bpy.props.BoolProperty(default=False)
    show_material_adv_section_6:bpy.props.BoolProperty(default=False)
    show_material_adv_section_7:bpy.props.BoolProperty(default=False)
    show_material_adv_section_8:bpy.props.BoolProperty(default=False)
    show_material_adv_section_9:bpy.props.BoolProperty(default=False)
    show_material_adv_section_10:bpy.props.BoolProperty(default=False)
    show_material_adv_section_11:bpy.props.BoolProperty(default=False)
    show_material_adv_section_12:bpy.props.BoolProperty(default=False)
    show_material_adv_section_13:bpy.props.BoolProperty(default=False)
    show_material_adv_section_14:bpy.props.BoolProperty(default=False)

    show_world_section_1:bpy.props.BoolProperty(default=False)
    show_world_section_2:bpy.props.BoolProperty(default=False)
    show_world_section_3:bpy.props.BoolProperty(default=False)
    show_world_section_4:bpy.props.BoolProperty(default=False)
    show_world_section_5:bpy.props.BoolProperty(default=False)
    show_world_section_6:bpy.props.BoolProperty(default=False)
    show_world_section_7:bpy.props.BoolProperty(default=False)
    show_world_section_8:bpy.props.BoolProperty(default=False)
    show_world_section_9:bpy.props.BoolProperty(default=False)
    show_world_section_10:bpy.props.BoolProperty(default=False)
    show_world_section_11:bpy.props.BoolProperty(default=False)
    show_world_section_12:bpy.props.BoolProperty(default=False)
    show_world_section_13:bpy.props.BoolProperty(default=False)
    show_world_section_14:bpy.props.BoolProperty(default=False)
    show_world_adv_section_1:bpy.props.BoolProperty(default=False)
    show_world_adv_section_2:bpy.props.BoolProperty(default=False)
    show_world_adv_section_3:bpy.props.BoolProperty(default=False)
    show_world_adv_section_4:bpy.props.BoolProperty(default=False)
    show_world_adv_section_5:bpy.props.BoolProperty(default=False)
    show_world_adv_section_6:bpy.props.BoolProperty(default=False)
    show_world_adv_section_7:bpy.props.BoolProperty(default=False)
    show_world_adv_section_8:bpy.props.BoolProperty(default=False)
    show_world_adv_section_9:bpy.props.BoolProperty(default=False)
    show_world_adv_section_10:bpy.props.BoolProperty(default=False)
    show_world_adv_section_11:bpy.props.BoolProperty(default=False)
    show_world_adv_section_12:bpy.props.BoolProperty(default=False)
    show_world_adv_section_13:bpy.props.BoolProperty(default=False)
    show_world_adv_section_14:bpy.props.BoolProperty(default=False)


class TA_OT_Cancel_Download(bpy.types.Operator):
    bl_idname = "ta.canceldownload"
    bl_label = "Cancel Download"
    bl_description = ""
    # bl_options = {}
    
    def execute(self, context):
        with open(os.path.join(ADDONS_DIR,"cancel.txt"), "w+") as f:
            f.write("CANCEL")
        bpy.ops.ed.undo()
        return {'FINISHED'}
class TA_OT_Show_Downloaded(bpy.types.Operator):
    bl_idname = "ta.showdownloaded"
    bl_label = "Show Downloaded Assets"
    bl_description = ""
    bl_options = {"REGISTER", "UNDO"}
    @classmethod
    def poll(cls,context):
        return context.area.ui_type=='ASSETS' 
    def execute(self, context):
        if context.space_data.params.filter_search=='Downloaded':
            context.space_data.params.filter_search=''
        else:
            context.space_data.params.filter_search='Downloaded'
        return {'FINISHED'}
class TA_OT_Delete_Key(bpy.types.Operator):
    bl_idname = "ta.removecontextmenuentries"
    bl_label = "Delete Windows Context Menu Entries"
    bl_description = ""
    bl_options = {"REGISTER", "UNDO"}
    
    def execute(self, context):
        delete_key(self,"fbx",['Create Assets',"Convert To Blend"])
        delete_key(self,"obj",['Create Assets',"Convert To Blend"])
        delete_key(self,"usdz",['Create Assets',"Convert To Blend"])
        delete_key(self,"usdc",['Create Assets',"Convert To Blend"])
        delete_key(self,"usda",['Create Assets',"Convert To Blend"])
        delete_key(self,"usd",['Create Assets',"Convert To Blend"])
        delete_key(self,"blendfile",["Create Assets","Move to True Assets Library",])
        return {'FINISHED'}
class TA_OT_Add_Key(bpy.types.Operator):
    bl_idname = "ta.createcontextmenuentries"
    bl_label = "Create Windows Context Menu Entries"
    bl_description = ""
    bl_options = {"REGISTER", "UNDO"}
    def execute(self, context):
        cwd = os.path.dirname(__file__)
        file_data=""
        with open(f"{cwd}\\call_script_with_multiple_files.py",mode="r") as file:
            file_data=file.readlines()
        
        if file_data:
            new_lines=[]
            for line in file_data:
                if "blender_executable=" in line:
                    new_lines.append(f"blender_executable=r'{bpy.app.binary_path}'\n")
                else:
                    new_lines.append(line)
            with open(f"{cwd}\\call_script_with_multiple_files.py",mode="w+") as file:
                
                file_data=file.writelines(new_lines)
        create_key(self,"fbx",['Create Assets',"Convert To Blend"],['create_fbx_and_obj_assets','converttoblendfile'])
        create_key(self,"usdz",['Create Assets',"Convert To Blend"],['create_fbx_and_obj_assets','converttoblendfile'])
        create_key(self,"usda",['Create Assets',"Convert To Blend"],['create_fbx_and_obj_assets','converttoblendfile'])
        create_key(self,"usdc",['Create Assets',"Convert To Blend"],['create_fbx_and_obj_assets','converttoblendfile'])
        create_key(self,"usd",['Create Assets',"Convert To Blend"],['create_fbx_and_obj_assets','converttoblendfile'])
        create_key(self,"obj",['Create Assets',"Convert To Blend"],['create_fbx_and_obj_assets','converttoblendfile'])
        create_key(self,"blendfile",["Create Assets","Move to True Assets Library",],["create_blend_assets",'movetoassetlibrary',])
        return {'FINISHED'}
class TA_UL_Library_Order_List(bpy.types.UIList):
    def draw_item(self, context, layout, data, item, icon, active_data, active_propname):
        ob = data

        obj = item
        if self.layout_type in {'DEFAULT', 'COMPACT'}:
            if obj:
                row = layout.row(align=True)
                row.label(text=obj.name)
                row.prop(obj,'path',text='')
        elif self.layout_type in {'GRID'}:
            row = layout.row()
            row.label(text=obj.name)
class TA_OT_Remove_Library(bpy.types.Operator):
    bl_idname = "ta.removelibrary"
    bl_label = "Remove"
    bl_description=""
    bl_options = {'REGISTER', 'UNDO'}
    
    def invoke(self, context,event):
        scene= context.scene
        try:
            scene.libraries_list.remove(scene.libraries_list_index)
        except Exception:
            pass
        
        return {'FINISHED'}
class TA_OT_Move_Library(bpy.types.Operator):
    bl_idname = "ta.movelibrary"
    bl_label = "Move"
    bl_description=""
    bl_options = {'REGISTER', 'UNDO'}
    direction: bpy.props.StringProperty(default="UP",options={'SKIP_SAVE','HIDDEN'})
    
    def invoke(self, context,event):
        scene= context.scene
        if self.direction=='UP':
            if scene.libraries_list_index>0:
                scene.libraries_list.move(scene.libraries_list_index,scene.libraries_list_index-1)
                scene.libraries_list_index=max(scene.libraries_list_index-1,0)
        else:
            if scene.libraries_list_index<len(scene.libraries_list)-1:
                scene.libraries_list.move(scene.libraries_list_index,scene.libraries_list_index+1)
                scene.libraries_list_index=min(len(scene.libraries_list),scene.libraries_list_index+1)
        return {'FINISHED'}
class TA_OT_Reorder_Libraries(bpy.types.Operator):
    bl_idname = 'ta.reorderlibraries'
    bl_label = 'Reorder Libraries'
    bl_description="Change the order of Libraries"
    def draw(self, context):
        layout=self.layout
        column = layout.column()
        column = column.split(factor=0.9)
        column.template_list("TA_UL_Library_Order_List", "", context.scene, "libraries_list",
                                    context.scene, "libraries_list_index" ,item_dyntip_propname='name',rows=len(context.scene.libraries_list))
        column = column.column(align=True)
        column.operator("ta.movelibrary", text="", icon="SORT_DESC").direction='UP'
        column.operator("ta.movelibrary", text="", icon="SORT_ASC").direction='DOWN'
        column.separator()
        column.operator("ta.removelibrary", text="", icon="REMOVE")
    def execute(self, context):
        for a in context.preferences.filepaths.asset_libraries:
            remove_library(a.path)
        for a in context.scene.libraries_list:
            ensure_library(a.name,a.path)
        return {'FINISHED'}
    def invoke(self, context, event):
        context.scene.libraries_list.clear()
        for a in context.preferences.filepaths.asset_libraries:
            t=context.scene.libraries_list.add()
            t.name=a.name
            t.path=a.path
            
        return context.window_manager.invoke_props_dialog(self,width=500)
class TA_OT_Batch_Delete_Catalogs(bpy.types.Operator):
    bl_idname = 'ta.batchdeletecatalogs'
    bl_label = 'Batch Remove Catalogs'
    bl_description="Delete multiple catalogs in one go"
    def draw(self, context):
        layout=self.layout
        for cat in context.scene.batch_delete_catalogs_list:
            row=layout.row()
            row.prop(cat,'checked',text='')
            row.label(text=cat.name)
    def execute(self, context):
        if context.space_data.params.asset_library_reference!='LOCAL':
            for cat in context.scene.batch_delete_catalogs_list:
                if cat.checked:
                    catalog_id=cat.id
                    catalog_name=cat.name.lstrip().lstrip('/')
                    cats_file=get_library_cat_file(context.space_data.params.asset_library_reference)
                    if cats_file:
                        file_data=[]
                        with open(cats_file, 'r') as filedata:
                            file_data=filedata.readlines()
                        new_data=[]
                        for line in file_data:
                            if ':' not in line:
                                new_data.append(line)
                            else:
                                catalog_path=line.split(':')
                                if len(catalog_path)==3:
                                    catalog_path=catalog_path[1].lstrip().lstrip('/')+"/"
                                    if not catalog_path.startswith(catalog_name):
                                        new_data.append(line)
                        if file_data:
                            with open(cats_file,mode='w+',newline='\n') as file:
                                for line in new_data:
                                        file.write(line)
        try:
            bpy.ops.asset.library_refresh() 
        except:
            pass
        return {'FINISHED'}
    def invoke(self, context, event):
        context.scene.batch_delete_catalogs_list.clear()
        self.dir_path=[a.path for a in context.preferences.filepaths.asset_libraries if a.name==context.space_data.params.asset_library_reference][0]
        catalogs=get_catalog_from_directory(self.dir_path,True)
        
        for a in catalogs:
            t=context.scene.batch_delete_catalogs_list.add()
            t.name=a[1]
            t.id=a[0]
        return context.window_manager.invoke_props_dialog(self,width=300)
class TA_OT_Open_Asset_Browser(bpy.types.Operator):
    bl_idname = "ta.openassetbrowser"
    bl_label = "Open Asset Browser"
    bl_description = ""
    bl_options = {"REGISTER", "UNDO"}
    def execute(self,context):
        create_ab_area(context)
        return {'FINISHED'}
class TA_OT_Open_TA_Library(bpy.types.Operator):
    bl_idname = "ta.opentalibrary"
    bl_label = "Open this Library"
    bl_description = ""
    bl_options = {"REGISTER"}
    name:bpy.props.StringProperty(name='Name')
    @classmethod
    def description(self,context,properties):
        return f"Open {properties.name} Library"
    def execute(self,context):
        if context.area.ui_type=='ASSETS':
            set_ab_library(context,self.name)
        else:
            create_ab_area(context,self.name)
        return {'FINISHED'}
def update_node_toggle(self,context):
    index=1
    node=[a for a in self.id_data.nodes if a.is_ta_node]
    if node:
        node=node[0]
        for toggle in [a for a in node.inputs if '*t' in a.name]:
            if getattr(self,f"toggle_{index}"):
                toggle.default_value=1
            else:
                toggle.default_value=0
            index+=1
class TA_NODE_TOGGLES(bpy.types.PropertyGroup):
    toggle_1:bpy.props.BoolProperty(default=False,update=update_node_toggle)
    toggle_2:bpy.props.BoolProperty(default=False,update=update_node_toggle)
    toggle_3:bpy.props.BoolProperty(default=False,update=update_node_toggle)
    toggle_4:bpy.props.BoolProperty(default=False,update=update_node_toggle)
    toggle_5:bpy.props.BoolProperty(default=False,update=update_node_toggle)
    toggle_6:bpy.props.BoolProperty(default=False,update=update_node_toggle)
    toggle_7:bpy.props.BoolProperty(default=False,update=update_node_toggle)
    toggle_8:bpy.props.BoolProperty(default=False,update=update_node_toggle)
    toggle_9:bpy.props.BoolProperty(default=False,update=update_node_toggle)
    toggle_10:bpy.props.BoolProperty(default=False,update=update_node_toggle)
    toggle_11:bpy.props.BoolProperty(default=False,update=update_node_toggle)
    toggle_12:bpy.props.BoolProperty(default=False,update=update_node_toggle)
class Library_List(bpy.types.PropertyGroup):
    name:bpy.props.StringProperty()
    path:bpy.props.StringProperty()
class Catalog_List(bpy.types.PropertyGroup):
    name:bpy.props.StringProperty()
    id:bpy.props.StringProperty()
    checked:bpy.props.BoolProperty()
class TA_OT_Install_Module(bpy.types.Operator,ImportHelper):
    bl_idname = 'ta.installmodule'
    bl_label = 'Install Module'
    # filename_ext = '.txt'
    
    filter_glob: bpy.props.StringProperty(
        default='*.zip',
        options={'HIDDEN'}
    )
    def execute(self, context):
        path=self.filepath
        if os.path.isfile(path):
            try:
                if not os.path.isdir(os.path.join(os.path.dirname(__file__),"modules")):
                    os.mkdir(os.path.join(os.path.dirname(__file__),"modules"))
                with zipfile.ZipFile(path, 'r') as zip_ref:
                                zip_ref.extractall(os.path.join(os.path.dirname(__file__),"modules"))
                                self.report({'INFO'},"Module Installed Successfully! Please restart Blender!")
            except:
                self.report({'WARNING'},"An Error Occured!")
        return {'FINISHED'}
TEMP_STUDIO_NAMES=[]
_item_map = dict()

def make_enum_item(id, name, descr, preview_id, uid):
    lookup = f"{str(id)}\0{str(name)}\0{str(descr)}\0{str(preview_id)}\0{str(uid)}"
    if not lookup in _item_map:
        _item_map[lookup] = (id, name, descr, preview_id, uid)
    return _item_map[lookup]
def enum_previews_from_directory(self, context):
    enum_items = []
    for i, name in enumerate(TA_State.studios):
        
        enum_items.append(make_enum_item(name,name, "", TA_State.get_icon(name), i))
    TA_State.studios_enum = enum_items
    return TA_State.studios_enum
def get_materials_enum(self,context):
    enum_items = [("None","None","None")]
    for i, name in enumerate(TA_State.saved_materials):
        enum_items.append(
            (name[1], name[0], ""))
    TA_State.materials_enum = enum_items
    return TA_State.materials_enum
classes = (Create_Assets_From_This_File_Props_Base,Create_Assets_From_Directory_Props_Base,Create_Assets_FBX_Props_Base,Create_Assets_From_OBJ_Props_Base,Create_Assets_From_USD_Props_Base,Create_Materials_Props_Base,Create_Worlds_Props_Base
    ,TA_OT_Install_Module,TA_PT_ASSET_BROWSER
    ,TA_OT_Clear_Broken_Materials,TA_OT_Open_TA_Library,TA_OT_Open_Asset_Browser,Catalog_List,TA_OT_Batch_Delete_Catalogs,TA_OT_Remove_Library,TA_OT_Move_Library,Library_List,TA_UL_Library_Order_List,TA_OT_Reorder_Libraries,TA_OT_Show_Downloaded,TA_NODE_TOGGLES,TA_OT_Switch_Shader_Type,TA_OT_Delete_Key,TA_OT_Add_Key,TA_OT_Replace_In_Scene,TA_OT_Fetch_Online_Asset,TA_OT_Cancel_Download,TA_OT_USD_Assets_From_Directory,
TA_OT_Rename_Catalog,TA_OT_Delete_Catalog,TA_OT_Clear_K3D_Asset,TA_Sub_Section,TA_OT_Update_Asset_Data,TA_OT_OBJ_Assets_From_Directory,TA_OT_FBX_Assets_From_Directory,TA_MT_Header_Menu,TA_Props,TA_Prefs,TA_OT_Reset_Progress,Asset_Names,TA_OT_Rename_Assets,TA_OT_Rerender_Thumbnail,TA_OT_Edit_Asset,TA_OT_Move_To_Catalog,TA_PT_TA,TA_OT_Create_Asset,TA_OT_Clear_Asset,TA_OT_Clear_Asset_By_Type,TA_OT_Assets_From_Directory,TA_OT_Materials_From_Directory,TA_OT_Save_Assets_From_Current_File,TA_OT_Create_Node_Group_Asset,TA_OT_WORLDS_From_Directory,TA_PT_Help
)


@persistent
def handle_undo(self,context):
    for mat in bpy.data.materials:
        if getattr(mat,'is_ta_dummy_mat',False):
            
            bpy.data.materials.remove(mat)
            # bpy.ops.ed.undo()
    for world in bpy.data.worlds:
        if getattr(world,'is_ta_dummy_world',False):
            
            bpy.data.worlds.remove(world)
            # bpy.ops.ed.undo()
@persistent
def check_for_material_drop(scene,despgraph):
    for obj in bpy.data.objects:
        for mat in getattr(obj.data,'materials',[]):
    # for mat in bpy.data.materials:
            if getattr(mat,'is_ta_dummy_mat',False):
                mat.is_ta_dummy_mat=False
                override = bpy.context.copy()
                for win in [a for a in bpy.context.window_manager.windows if a.screen.name!='temp']:
                    for area in win.screen.areas:
                        if area.ui_type == "ASSETS":
                            override['area']=area
                            override['window']=win
                            override['screen']=win.screen
                            # override['region_data']=area.regions[-1].region_data
                            break
                # override["region"] = 
                # with bpy.context.temp_override(**override):
                if getattr(mat,'st_slug',''):
                    bpy.ops.ta.importsharetexturesmaterial('INVOKE_DEFAULT',mat=mat.name)
                else:
                    bpy.ops.ta.importmaterial('INVOKE_DEFAULT',mat=mat.name)
                return
    for world in bpy.data.worlds:
            if getattr(world,'is_ta_dummy_world',False):
                world.is_ta_dummy_world=False
                override = bpy.context.copy()
                for win in [a for a in bpy.context.window_manager.windows if a.screen.name!='temp']:
                    for area in win.screen.areas:
                        if area.ui_type == "ASSETS":
                            override['area']=area
                            override['window']=win
                            override['screen']=win.screen
                # override["region"] = 
                # with bpy.context.temp_override(**override):
                bpy.ops.ta.importworld('INVOKE_DEFAULT',world=world.name)
@persistent
def set_progress(scene):
    ensure_all_libraries()
    for scene in bpy.data.scenes:
        scene.ta_progress=""
LAST_QUALITY=None
LAST_ASSET_TAGS=[]
LAST_ASSET_TYPE='MATERIAL'
def get_asset_quality(self):
    global LAST_QUALITY,LAST_ASSET_TYPE,LAST_ASSET_TAGS
    if LAST_ASSET_TAGS:
        if LAST_ASSET_TYPE=='MATERIAL':
            quality_indices=['1K-PNG','2K-PNG','4K-PNG','6K-PNG','8K-PNG','10K-PNG','12K-PNG','16K-PNG','1K-JPG','2K-JPG','4K-JPG','6K-JPG','8K-JPG','10K-JPG','12K-JPG','16K-JPG']
        else:
            quality_indices=['1K-HDR','2K-HDR','4K-HDR','6K-HDR','8K-HDR','12K-HDR','16K-HDR','1K-TONEMAPPED','2K-TONEMAPPED','4K-TONEMAPPED','6K-TONEMAPPED','8K-TONEMAPPED','12K-TONEMAPPED','16K-TONEMAPPED']
        tags=LAST_ASSET_TAGS
        qualities=[a[:a.find('(')] for a in tags if '(' in a and 'K' in a]
        if LAST_QUALITY is not None:
            last_quality_string=quality_indices[LAST_QUALITY]
            if last_quality_string not in qualities:
                if last_quality_string.replace("PNG","JPG") in qualities:
                    return quality_indices.index(last_quality_string.replace("PNG","JPG"))
                else:
                    if not qualities:
                        return 0
                    return quality_indices.index(qualities.pop())
        if LAST_QUALITY is not None and LAST_QUALITY<len(quality_indices):
            return LAST_QUALITY
        return 0
    return 0
def set_asset_quality(self,value):
    global LAST_QUALITY
    LAST_QUALITY=value

def get_material_qualities(self,context):
    tags=[]
    global LAST_ASSET_TAGS,LAST_ASSET_TYPE
    if context.asset:
        LAST_ASSET_TAGS=[a.name for a in context.asset.metadata.tags]
        LAST_ASSET_TYPE=context.asset.id_type
    # if context.asset:
    if LAST_ASSET_TAGS:
        if LAST_ASSET_TYPE=='MATERIAL':
            quality_indices=['1K-PNG','2K-PNG','4K-PNG','6K-PNG','8K-PNG','10K-PNG','12K-PNG','16K-PNG','1K-JPG','2K-JPG','4K-JPG','6K-JPG','8K-JPG','10K-JPG','12K-JPG','16K-JPG']
            tags=LAST_ASSET_TAGS
            qualities=[a[:a.find('(')] for a in tags if '(' in a and 'K' in a]
        else:
            quality_indices=['1K-HDR','2K-HDR','4K-HDR','6K-HDR','8K-HDR','12K-HDR','16K-HDR','1K-TONEMAPPED','2K-TONEMAPPED','4K-TONEMAPPED','6K-TONEMAPPED','8K-TONEMAPPED','12K-TONEMAPPED','16K-TONEMAPPED']
            tags=LAST_ASSET_TAGS
            qualities=[a[:a.find('(')] for a in tags if '(' in a and 'K' in a]
                
    else:
        quality_indices=['1K-JPG','2K-JPG','4K-JPG','8K-JPG','12K-JPG','1K-PNG','2K-PNG','4K-PNG','8K-PNG','12K-PNG']
        qualities=['1K-JPG','2K-JPG','4K-JPG','8K-JPG','12K-JPG','1K-PNG','2K-PNG','4K-PNG','8K-PNG','12K-PNG']
    qualities=sorted(qualities, key=lambda q: quality_indices.index(q))
    tags=','.join(tags)
    TA_State.temp_qualities=[(a,a+"(Downloaded)" if f"Downloaded-{a[:5]}" in tags else a,a,quality_indices.index(a)) for a in qualities]
    return TA_State.temp_qualities

last_library=''
def check_for_library_change():
    if preferences().auto_clear_search:
        global last_library
        for win in bpy.context.window_manager.windows:
            for area in win.screen.areas:
                if area.ui_type == "ASSETS":
                    if last_library!=area.spaces[0].params.asset_library_reference:
                        last_library=area.spaces[0].params.asset_library_reference
                        area.spaces[0].params.filter_search=''
        return 1
    else:
        return 10
icon_collection={}
def load_icons():
    pcoll = bpy.utils.previews.new()
    my_icons_dir = os.path.join(os.path.dirname(__file__), "icons")
    pcoll.load("youtube", os.path.join(my_icons_dir, "Youtube.png"), 'IMAGE')
    pcoll.load("discord", os.path.join(my_icons_dir, "Discord.png"), 'IMAGE')
    pcoll.load("ambientcg", os.path.join(my_icons_dir, "ambientcg.png"), 'IMAGE')
    pcoll.load("sharetextures", os.path.join(my_icons_dir, "ShareTextures.png"), 'IMAGE')
    icon_collection["icons"] = pcoll
def unload_icons():
    for pcoll in icon_collection.values():
        bpy.utils.previews.remove(pcoll)
    icon_collection.clear()
def ta_manual_map():
    url_manual_prefix = "https://www.true-vfx.xyz/true-assets-documentation/"
    url_manual_mapping = (
        ("bpy.ops.ta.createassetsfromthisfile", "creating-assets-from-current-file"),
        ("bpy.ops.ta.createassetsfromdirectory", "creating-assets-from-directory"),
        ("bpy.ops.ta.createfbxassetsfromdirectory", "create-assets-from-fbx"),
        ("bpy.ops.ta.createobjassetsfromdirectory", "create-assets-from-obj"),
        ("bpy.ops.ta.createusdassetsfromdirectory", "create-assets-from-usd"),
        ("bpy.ops.ta.creatematerialsfromdirectory", "create-materials"),
        ("bpy.ops.ta.createworldsfromdirectory", "create-worlds"),
        ("bpy.ops.ta.switchshader", "asset-browser-utilities#toc-switch-shaders"),
        ("bpy.ops.ta.reorderlibraries", "asset-browser-utilities#toc-reorder-libraries"),
        ("bpy.ops.ta.renamecatalog", "asset-browser-utilities#toc-rename-catalog"),
        ("bpy.ops.ta.batchdeletecatalogs", "asset-browser-utilities#toc-batch-delete-catalogs"),
        ("bpy.ops.ta.deletecatalog", "asset-browser-utilities#toc-delete-catalog"),
        ("bpy.ops.ta.rerenderthumbnail", "asset-browser-utilities#toc-re-render-thumbnail"),
        ("bpy.ops.ta.movetocatalog", "asset-browser-utilities#toc-move-to-catalog"),
        ("bpy.ops.ta.fixkitbashassets", "asset-browser-utilities#toc-fix-kitbash-assets"),
        ("bpy.ops.ta.clearassetsbytype", "asset-browser-utilities#toc-clear-assets-trueassets"),
        ("bpy.ops.ta.clearasset", "asset-browser-utilities#toc-clear-assets-trueassets"),
        ("bpy.ops.ta.renameassets", "asset-browser-utilities#toc-rename-assets"),
        ("bpy.ops.ta.updateassetdata", "asset-browser-utilities#toc-update-asset-data"),
        ("bpy.ops.ta.editasset", "asset-browser-utilities#toc-update-asset-data"),
        ("bpy.ops.ta.fetchonlineassets", "first-time-use#toc-getting-the-library-to-load"),
        ("bpy.ops.ta.showdownloaded", "first-time-use#toc-downloaded-assets"),


    )
    return url_manual_prefix, url_manual_mapping
def register():
    from bpy.utils import register_class
    for cls in classes:
        register_class(cls)
    register_extra_modules()
    load_icons()
    wm = bpy.context.window_manager
    kc = wm.keyconfigs.addon

    km = kc.keymaps.new(name="3D View", space_type="VIEW_3D")
    if kc:
        kmi = km.keymap_items.new(
                "ta.createassetsfromthisfile",
                type="W",
                value="PRESS",
                ctrl=True,
            )
        kmi.properties.quit=True
        addon_keymaps.append((km, kmi))
    bpy.types.World.ta_asset_id=bpy.props.StringProperty(name="",default="")
    bpy.types.World.ta_tags=bpy.props.StringProperty(name="Tags",default="")
    bpy.types.World.ta_qualities=bpy.props.StringProperty(name="Qualities",default="")
    bpy.types.World.ta_quality=bpy.props.StringProperty()
    bpy.types.World.is_ta_dummy_world=bpy.props.BoolProperty(name="Is Ta Dummy",default=False)
    bpy.types.Material.ta_asset_id=bpy.props.StringProperty(name="",default="")
    bpy.types.Material.ta_tags=bpy.props.StringProperty(name="Tags",default="")
    bpy.types.Material.ta_qualities=bpy.props.StringProperty(name="Qualities",default="")
    bpy.types.Material.is_ta_dummy_mat=bpy.props.BoolProperty(name="Is Ta Dummy",default=False)
    bpy.types.Material.ta_quality=bpy.props.StringProperty()
    bpy.types.Material.ta_shader_type=bpy.props.StringProperty()
    bpy.types.Material.st_slug=bpy.props.StringProperty()
    bpy.types.World.ta_shader_type=bpy.props.StringProperty()
    bpy.types.Scene.ta_shader_type=bpy.props.EnumProperty(items=(("Ambient CG (Uber)","AmbientCG-Uber","Ambient CG (Uber)"),("Ambient CG (Standard)","AmbientCG Standard","Ambient CG (Standard)"),("True-Uber","Uber","True-Uber"),("True-Plastic","Plastic","True-Plastic"),("True-Organic","Organic","True-Organic"),("True-BSDF","BSDF","True-BSDF"),("Standard","Standard","Standard")),name="Shader Type")
    bpy.types.Scene.ta_world_shader_type=bpy.props.EnumProperty(items=(("True-HDRI","True-HDRI","True-HdRi"),("True-HDRI (Standard)","Standard","True-HDRI Standard")),name="Shader Type")
    bpy.types.Scene.ta_material_quality=bpy.props.EnumProperty(items=get_material_qualities,name="Material Quality",get=get_asset_quality,set=set_asset_quality)
    bpy.types.NODE_MT_context_menu.prepend(drawIntoNodeGroupMenu)
    #bpy.types.OUTLINER_MT_asset.prepend(drawIntoOutliner)
    bpy.types.VIEW3D_MT_object_context_menu.append(drawIntoOutliner)
    bpy.types.ASSETBROWSER_MT_editor_menus.append(drawIntoAssetHeader)
    bpy.types.FILEBROWSER_HT_header.append(draw_into_header_right)
    
    bpy.types.Scene.ta_props=bpy.props.PointerProperty(type=TA_Props)
    bpy.types.ASSETBROWSER_MT_context_menu.append(drawIntoAssetContextMenu)
    bpy.types.Scene.ta_progress=bpy.props.StringProperty(default="")
    bpy.types.Scene.ta_progress_bar=bpy.props.FloatProperty(default=0.0,min=0.0,max=100.0,subtype='PERCENTAGE',name='Fetching Materials')
    bpy.types.Scene.paths_in_use=bpy.props.StringProperty(default="")
    bpy.types.Scene.show_material_settings=bpy.props.BoolProperty(default=False,name="Material Settings")
    
    bpy.types.Scene.show_hdri_settings=bpy.props.BoolProperty(default=False,name="World Settings")
    bpy.types.Scene.sub_sections=bpy.props.PointerProperty(type=TA_Sub_Section)
    bpy.types.Node.ta_toggles=bpy.props.PointerProperty(type=TA_NODE_TOGGLES)
    bpy.types.Node.is_ta_node=bpy.props.BoolProperty(default=False)
    bpy.types.Scene.libraries_list=bpy.props.CollectionProperty(type=Library_List)
    bpy.types.Scene.libraries_list_index= bpy.props.IntProperty(default=0,name="Selected LIbrary")
    bpy.types.Scene.batch_delete_catalogs_list=bpy.props.CollectionProperty(type=Catalog_List)
    bpy.types.Scene.ta_studio=bpy.props.EnumProperty(items=enum_previews_from_directory,name="Studio")
    bpy.types.Scene.ta_material=bpy.props.EnumProperty(items=get_materials_enum,name="Material")
    bpy.types.Collection.og_instance_offset=bpy.props.FloatVectorProperty(default=(0,0,0),size=3)
    bpy.types.Scene.use_as_tt_layer=bpy.props.BoolProperty(default=False,name="Drop as TT Layers")
    loadPreferences()
    bpy.app.handlers.load_post.append(set_progress)
    bpy.app.handlers.depsgraph_update_post.append(check_for_material_drop)
    bpy.app.handlers.undo_post.append(handle_undo)
    bpy.utils.register_manual_map(ta_manual_map)
    # TA_State.load_studio_icons()
    # bpy.app.timers.register(check_for_library_change)
    addon_update_checker.register('e1bef0c93fee907842789d259ba4f57e')
    try:
        bpy.app.translations.register(__package__, GetTranslationDict())
    except Exception as e: print(e)


import pkgutil
def register_extra_modules():
    global EXTRA_MODULES
    modules_path = "True-Assets.modules"
    modules_folder=os.path.join(os.path.dirname(__file__),'modules')
    if os.path.isdir(modules_folder):
        submodules = [name for name in os.listdir(modules_folder) if os.path.isdir(os.path.join(modules_folder, name)) and name!='__pycache__']

        for submodule in submodules:
            print("Registering TA module:",submodule)
            try:
                module = importlib.import_module(f"{modules_path}.{submodule}")
                
                # Check if the 'register' function exists in the submodule
                if hasattr(module, 'register') and callable(module.register):
                    module.register()
                    EXTRA_MODULES.append(module)
                else:
                    print(f"Submodule {submodule} does not have a 'register' function.")
            except ImportError as e:
                print(f"Error importing submodule {submodule}: {e}")
def unregister():
    global EXTRA_MODULES
    bpy.context.scene.ta_progress=""
    savePreferences()
    
    # TA_State.unload_studio_icons()
    from bpy.utils import unregister_class
    extra_modules=set(EXTRA_MODULES)
    for module in extra_modules:
        try:
            module.unregister()
        except:
            print("Could not unregister",module)
    
    for cls in reversed(classes):
        unregister_class(cls)
    unload_icons()
    for (km, kmi) in addon_keymaps:
        km.keymap_items.remove(kmi)
    addon_keymaps.clear()
    del bpy.types.Scene.ta_progress
    #bpy.types.OUTLINER_MT_asset.remove(drawIntoOutliner)
    bpy.types.VIEW3D_MT_object_context_menu.remove(drawIntoOutliner)
    bpy.types.NODE_MT_context_menu.remove(drawIntoNodeGroupMenu)
    bpy.types.ASSETBROWSER_MT_context_menu.remove(drawIntoAssetContextMenu)
    bpy.types.ASSETBROWSER_MT_editor_menus.remove(drawIntoAssetHeader)
    bpy.types.FILEBROWSER_HT_header.remove(draw_into_header_right)
    try:
        bpy.app.handlers.load_post.remove(set_progress)
    except:
        pass
    try:
        bpy.app.handlers.depsgraph_update_post.remove(check_for_material_drop)
    except Exception:
        pass
    try:
        bpy.app.handlers.undo_post.remove(handle_undo)
    except Exception:
        pass
    try:
        bpy.app.timers.unregister(check_for_library_change)
    except Exception:
        pass
    bpy.utils.unregister_manual_map(ta_manual_map)
    addon_update_checker.unregister()
    try:
        bpy.app.translations.unregister(__package__)
    except Exception as e: print(e)
if __name__ == "__main__":
    register()

